@@ -7,9 +7,6 @@
 /* */
 #undef HAVE_OPENMP
 
-/* */
-#undef HAVE_TYPEOF
-
 /* */
 #undef HAVE_LONG_LONG
 
@@ -28,6 +25,12 @@
 /* */
 #undef HAVE_VOLATILE
 
+/* */
+#undef HAVE_RESTRICT
+
+/* */
+#undef HAVE___RESTRICT
+
 /* */
 #undef HAVE_ERRNO
 
@@ -58,12 +61,6 @@
 /* */
 #undef HAVE_STRNCMP
 
-/* */
-#undef HAVE_STRCASECMP
-
-/* */
-#undef HAVE_STRNCASECMP
-
 /* */
 #undef HAVE_STRCSPN
 
@@ -88,12 +85,24 @@
 /* */
 #undef HAVE_MEMSET
 
+/* */
+#undef HAVE_MEMMEM
+
 /* */
 #undef HAVE_STRERROR
 
+/* */
+#undef HAVE_SIGFILLSET
+
 /* */
 #undef HAVE_SIGSETJMP
 
+/* */
+#undef HAVE___BUILTIN_ADD_OVERFLOW
+
+/* */
+#undef HAVE___BUILTIN_CLZ
+
 /* */
 #undef DEBUGLEVEL
 
@@ -154,15 +163,27 @@
 /* */
 #undef HAVE_NSS
 
+/* */
+#undef HAVE_CRYPTO_SET_MEM_FUNCTIONS_1
+
+/* */
+#undef HAVE_CRYPTO_SET_MEM_FUNCTIONS_2
+
 /* */
 #undef HAVE_ZLIB
 
+/* */
+#undef HAVE_BROTLI
+
 /* */
 #undef HAVE_BZIP2
 
 /* */
 #undef HAVE_LZMA
 
+/* */
+#undef HAVE_LZIP
+
 /* */
 #undef G
 
@@ -190,9 +211,6 @@
 /* */
 #undef GRDRV_GRX
 
-/* */
-#undef DONT_INCLUDE_SETJMP
-
 /* Jpeg by Clock */
 #undef HAVE_JPEG
 
