@@ -295,7 +295,7 @@ static void bookmark_edit_item(struct dialog_data *dlg, struct list *data, void
 			break;
 
 		default:
-			internal_error("Unsupported dialog title.\n");
+			internal("Unsupported dialog title.\n");
 	}
 	d->fn = bookmark_edit_item_fn;
 	d->udata = item;  /* item */
@@ -572,7 +572,7 @@ static void load_bookmarks(struct session *ses)
 		while (p < end && *p != '<') p++;  /* find start of html tag */
 		if (p >= end) break;   /* parse end */
 		s = p;
-		if (end - p >= 2 && (p[1] == '!' || p[1]== '?')) { p = skip_comment(p, end); continue; }
+		if (p + 2 <= end && (p[1] == '!' || p[1]== '?')) { p = skip_comment(p,end); continue; }
 		if (parse_element(p, end, &name, &namelen, &attr, &p)) { p++; continue; }
 
 		switch (status) {
@@ -782,7 +782,7 @@ static void save_bookmarks(struct session *ses)
 	} else {
 		if (ses) {
 			unsigned char *f = stracpy(bookmarks_file);
-			msg_box(ses->term, getml(f, NULL), TEXT_(T_BOOKMARK_ERROR), AL_CENTER, TEXT_(T_UNABLE_TO_WRITE_TO_BOOKMARK_FILE), cast_uchar " ", f, cast_uchar ": ", get_err_msg(err, ses->term), MSG_BOX_END, NULL, 1, TEXT_(T_CANCEL), msg_box_null, B_ENTER | B_ESC);
+			msg_box(ses->term, getml(f, NULL), TEXT_(T_BOOKMARK_ERROR), AL_CENTER | AL_EXTD_TEXT, TEXT_(T_UNABLE_TO_WRITE_TO_BOOKMARK_FILE), cast_uchar " ", f, cast_uchar ": ", get_err_msg(err), NULL, NULL, 1, TEXT_(T_CANCEL), NULL, B_ENTER | B_ESC);
 		}
 	}
 
@@ -791,9 +791,8 @@ static void save_bookmarks(struct session *ses)
 		memset(&bookmarks_st, -1, sizeof bookmarks_st);
 }
 
-void menu_bookmark_manager(struct terminal *term, void *fcp, void *ses_)
+void menu_bookmark_manager(struct terminal *term,void *fcp,struct session *ses)
 {
-	struct session *ses = (struct session *)ses_;
 	struct stat st;
 	int rs;
 	EINTRLOOP(rs, stat(cast_const_char bookmarks_file, &st));
