@@ -63,6 +63,8 @@ int set_cookie(struct terminal *term, unsigned char *url, unsigned char *str)
 	date = parse_header_param(str, cast_uchar "expires", 0);
 	if (date) {
 		cookie->expires = parse_http_date(date);
+		/* kdo tohle napsal a proc ?? */
+		/*if (! cookie->expires) cookie->expires++;*/ /* no harm and we can use zero then */
 		mem_free(date);
 	} else {
 		cookie->expires = 0;
@@ -119,7 +121,6 @@ static void accept_cookie(struct cookie *c)
 	struct list_head *lcd;
 	struct cookie *d;
 	struct list_head *ld;
-	size_t sl;
 	foreach(struct cookie, d, ld, all_cookies) if (!casestrcmp(d->name, c->name) && !casestrcmp(d->domain, c->domain)) {
 		ld = ld->prev;
 		del_from_list(d);
@@ -131,9 +132,7 @@ static void accept_cookie(struct cookie *c)
 	}
 	add_to_list(all_cookies, c);
 	foreach(struct c_domain, cd, lcd, c_domains) if (!casestrcmp(cd->domain, c->domain)) return;
-	sl = strlen(cast_const_char c->domain);
-	if (sl > MAXINT - sizeof(struct c_domain)) overalloc();
-	cd = mem_alloc(sizeof(struct c_domain) + sl);
+	cd = mem_alloc(sizeof(struct c_domain) + strlen(cast_const_char c->domain));
 	strcpy(cast_char cd->domain, cast_const_char c->domain);
 	add_to_list(c_domains, cd);
 }
@@ -158,12 +157,12 @@ int is_path_prefix(unsigned char *d, unsigned char *s)
 	return d[dl - 1] == '/' || !s[dl] || s[dl] == '/' || s[dl] == POST_CHAR || s[dl] == '?' || s[dl] == '&';
 }
 
-int cookie_expired(struct cookie *c)
+int cookie_expired(struct cookie *c)	/* parse_http_date is broken */
 {
 	time_t t;
 	errno = 0;
 	EINTRLOOPX(t, time(NULL), (time_t)-1);
-	return c->expires && c->expires < t;
+	return 0 && (c->expires && c->expires < t);
 }
 
 void add_cookies(unsigned char **s, int *l, unsigned char *url)
@@ -192,7 +191,7 @@ void add_cookies(unsigned char **s, int *l, unsigned char *url)
 		else add_to_str(s, l, cast_uchar "; ");
 		add_to_str(s, l, c->name);
 		if (c->value) {
-			add_chr_to_str(s, l, '=');
+			add_to_str(s, l, cast_uchar "=");
 			add_to_str(s, l, c->value);
 		}
 	}
