@@ -540,7 +540,7 @@ static void parse_config_file(unsigned char *name, unsigned char *file, struct o
 		f:
 		if (tok) mem_free(tok);
 	}
-	if (err) fprintf(stderr, "\007"), sleep(1);
+	if (err) fprintf(stderr, "\007"), portable_sleep(1000);
 }
 
 static unsigned char *create_config_string(struct option *options)
@@ -549,7 +549,7 @@ static unsigned char *create_config_string(struct option *options)
 	int l = 0;
 	int i;
 	add_to_str(&s, &l, cast_uchar "# This file is automatically generated by Links -- please do not edit.");
-#ifdef __DECC_VER
+#if defined(__DECC_VER) && !defined(OPENVMS)
 	do_not_optimize_here(&options);
 #endif
 	for (i = 0; options[i].p; i++) {
@@ -583,7 +583,7 @@ unsigned char *read_config_file(unsigned char *name)
 	return s;
 }
 
-int write_to_config_file(unsigned char *name, unsigned char *c)
+int write_to_config_file(unsigned char *name, unsigned char *c, int do_sync)
 {
 	int rr;
 	int h, w;
@@ -620,8 +620,14 @@ try_new_count:
 	rr = (int)strlen(cast_const_char c);
 	if (hard_write(h, c, rr) != rr) {
 		err = errno;
-		EINTRLOOP(rs, close(h));
-		goto unlink_err;
+		goto close_unlink_err;
+	}
+	if (do_sync) {
+		EINTRLOOP(rs, fsync(h));
+		if (rs) {
+			err = errno;
+			goto close_unlink_err;
+		}
 	}
 	EINTRLOOP(rs, close(h));
 	if (rs) {
@@ -643,8 +649,37 @@ try_new_count:
 		goto unlink_err;
 	}
 	mem_free(tmp_name);
+#if !defined(OS2) && !defined(OPENVMS)
+	if (do_sync) {
+		unsigned char *e, *le;
+		tmp_name = stracpy(name);
+		le = tmp_name;
+		for (e = tmp_name; *e; e++) if (dir_sep(*e)) le = e;
+		while (le > tmp_name && dir_sep(le[-1])) le--;
+		if (le == tmp_name && dir_sep(*le)) le++;
+#ifdef DOS_FS
+		if (le - tmp_name <= 2 && upcase(tmp_name[0]) >= 'A' && upcase(tmp_name[0]) <= 'Z' && tmp_name[1] == ':') {
+			if (dir_sep(tmp_name[2]))
+				le = tmp_name + 3;
+			else
+				le = tmp_name + 2;
+		}
+#endif
+		*le = 0;
+
+		h = c_open(*tmp_name ? tmp_name : cast_uchar ".", O_RDONLY | O_NOCTTY);
+		if (h != -1) {
+			EINTRLOOP(rs, fsync(h));
+			EINTRLOOP(rs, close(h));
+		}
+		mem_free(tmp_name);
+	}
+#endif
+
 	return 0;
 
+	close_unlink_err:
+	EINTRLOOP(rs, close(h));
 	unlink_err:
 	EINTRLOOP(rs, unlink(cast_const_char tmp_name));
 	free_err:
@@ -752,21 +787,29 @@ skip_path_conv:;
 			add_to_strn(&home_links, cast_uchar "/links");
 		} else {
 			fprintf(stderr, "CONFIG_DIR set to %s. But directory %s doesn't exist.\n\007", config_dir, home_links);
-			sleep(3);
+			portable_sleep(3000);
 			mem_free(home_links);
 			home_links = stracpy(home);
-			add_to_strn(&home_links, cast_uchar ".links");
+			goto add_dot_links;
 		}
 		mem_free(config_dir);
-	} else add_to_strn(&home_links, cast_uchar ".links");
+	} else {
+		add_dot_links:
+#ifdef OPENVMS
+		add_to_strn(&home_links, cast_uchar "links");
+#else
+		add_to_strn(&home_links, cast_uchar ".links");
+#endif
+	}
 	EINTRLOOP(rs, stat(cast_const_char home_links, &st));
 	if (rs) {
-#ifdef HAVE_MKDIR
 		EINTRLOOP(rs, mkdir(cast_const_char home_links, 0700));
 		if (!rs) goto home_creat;
 #ifdef OPENVMS
-		if (errno == EEXIST) goto home_ok;
-#endif
+		if (errno == EEXIST) {
+			memset(&st, 0, sizeof st);
+			goto home_ok;
+		}
 #endif
 		if (config_dir) goto failed;
 		goto first_failed;
@@ -775,10 +818,8 @@ skip_path_conv:;
 	/* This is a Cygwin hack! Cygwin reports stat for "links" if no
 	   "links" exists and only "links.exe" does. So try to create directory
 	   anyway. */
-#ifdef HAVE_MKDIR
 	EINTRLOOP(rs, mkdir(cast_const_char home_links, 0700));
 	if (!rs) goto home_creat;
-#endif
 	first_failed:
 	mem_free(home_links);
 	home_links = stracpy(home);
@@ -789,24 +830,19 @@ skip_path_conv:;
 #endif
 	EINTRLOOP(rs, stat(cast_const_char home_links, &st));
 	if (rs) {
-#ifdef HAVE_MKDIR
 		EINTRLOOP(rs, mkdir(cast_const_char home_links, 0700));
 		if (!rs) goto home_creat;
 #ifdef OPENVMS
-		if (errno == EEXIST) goto home_ok;
-#endif
-#else
-		mem_free(home_links);
-		home_links = stracpy(home);
-		goto home_ok;
+		if (errno == EEXIST) {
+			memset(&st, 0, sizeof st);
+			goto home_ok;
+		}
 #endif
 		goto failed;
 	}
 	if (S_ISDIR(st.st_mode)) goto home_ok;
-#ifdef HAVE_MKDIR
 	EINTRLOOP(rs, mkdir(cast_const_char home_links, 0700));
 	if (!rs) goto home_creat;
-#endif
 	failed:
 	mem_free(home_links);
 	mem_free(home);
@@ -814,10 +850,12 @@ skip_path_conv:;
 
 	home_ok:
 	if (n) *n = 0;
-	home_creat:
+	if ((st.st_mode & 07777) != 0700) {
+		home_creat:
 #ifdef HAVE_CHMOD
-	EINTRLOOP(rs, chmod(cast_const_char home_links, 0700));
+		EINTRLOOP(rs, chmod(cast_const_char home_links, 0700));
 #endif
+	}
 	add_to_strn(&home_links, cast_uchar "/");
 	mem_free(home);
 	return home_links;
@@ -830,7 +868,7 @@ void init_home(void)
 	links_home = get_home(&first_use);
 	if (!links_home) {
 		fprintf(stderr, "Unable to find or create links config directory. Please check, that you have $HOME variable set correctly and that you have write permission to your home directory.\n\007");
-		sleep(3);
+		portable_sleep(3000);
 		return;
 	}
 }
@@ -846,10 +884,11 @@ static int write_config_data(unsigned char *prefix, unsigned char *name, struct
 	config_file = stracpy(prefix);
 	if (!config_file) {
 		mem_free(c);
+		if (term) msg_box(term, NULL, TEXT_(T_CONFIG_ERROR), AL_CENTER | AL_EXTD_TEXT, TEXT_(T_UNABLE_TO_WRITE_TO_CONFIG_FILE), cast_uchar ": ", TEXT_(T_HOME_DIRECTORY_INACCESSIBLE), NULL, NULL, 1, TEXT_(T_CANCEL), NULL, B_ENTER | B_ESC);
 		return -1;
 	}
 	add_to_strn(&config_file, name);
-	if ((err = write_to_config_file(config_file, c))) {
+	if ((err = write_to_config_file(config_file, c, 1))) {
 		if (term) msg_box(term, NULL, TEXT_(T_CONFIG_ERROR), AL_CENTER | AL_EXTD_TEXT, TEXT_(T_UNABLE_TO_WRITE_TO_CONFIG_FILE), cast_uchar ": ", get_err_msg(err), NULL, NULL, 1, TEXT_(T_CANCEL), NULL, B_ENTER | B_ESC);
 		mem_free(c);
 		mem_free(config_file);
@@ -1055,7 +1094,7 @@ static unsigned char *block_rd(struct option *o, unsigned char *c)
 	unsigned char *err = cast_uchar "Error reading image block specification";
 	unsigned char* url;
 
-	if(!(url = get_token(&c)))
+	if (!(url = get_token(&c)))
 		return err;
 
 	block_url_add(NULL, url);
@@ -1067,27 +1106,31 @@ static unsigned char *block_rd(struct option *o, unsigned char *c)
 
 static void block_wr(struct option *o, unsigned char **s, int *l)
 {
-	struct block *a;
-	foreachback(a, blocks) {
+	struct list *a;
+	struct list_head *la;
+	foreachback(struct list, a, la, blocks.list_entry) {
+		struct block *b = get_struct(a, struct block, head);
 		add_nm(o, s, l);
-		add_quoted_to_str(s, l, a->url);
+		add_quoted_to_str(s, l, b->url);
 	}
 }
 
 static void type_wr(struct option *o, unsigned char **s, int *l)
 {
-	struct assoc *a;
-	foreachback(a, assoc) {
+	struct list *a;
+	struct list_head *la;
+	foreachback(struct list, a, la, assoc.list_entry) {
+		struct assoc *as = get_struct(a, struct assoc, head);
 		add_nm(o, s, l);
-		add_quoted_to_str(s, l, a->label);
+		add_quoted_to_str(s, l, as->label);
 		add_to_str(s, l, cast_uchar " ");
-		add_quoted_to_str(s, l, a->ct);
+		add_quoted_to_str(s, l, as->ct);
 		add_to_str(s, l, cast_uchar " ");
-		add_quoted_to_str(s, l, a->prog);
+		add_quoted_to_str(s, l, as->prog);
 		add_to_str(s, l, cast_uchar " ");
-		add_num_to_str(s, l, (!!a->cons) + (!!a->xwin) * 2 + (!!a->ask) * 4 + (!a->block) * 8 + (!!a->block) * 16 + (!!a->accept_http) * 32 + (!!a->accept_ftp) * 64);
+		add_num_to_str(s, l, (!!as->cons) + (!!as->xwin) * 2 + (!!as->ask) * 4 + (!as->block) * 8 + (!!as->block) * 16 + (!!as->accept_http) * 32 + (!!as->accept_ftp) * 64);
 		add_to_str(s, l, cast_uchar " ");
-		add_num_to_str(s, l, a->system);
+		add_num_to_str(s, l, as->system);
 	}
 }
 
@@ -1108,12 +1151,14 @@ static unsigned char *ext_rd(struct option *o, unsigned char *c)
 
 static void ext_wr(struct option *o, unsigned char **s, int *l)
 {
-	struct extension *a;
-	foreachback(a, extensions) {
+	struct list *a;
+	struct list_head *la;
+	foreachback(struct list, a, la, extensions.list_entry) {
+		struct extension *e = get_struct(a, struct extension, head);
 		add_nm(o, s, l);
-		add_quoted_to_str(s, l, a->ext);
+		add_quoted_to_str(s, l, e->ext);
 		add_to_str(s, l, cast_uchar " ");
-		add_quoted_to_str(s, l, a->ct);
+		add_quoted_to_str(s, l, e->ct);
 	}
 }
 
@@ -1138,7 +1183,8 @@ static unsigned char *prog_rd(struct option *o, unsigned char *c)
 static void prog_wr(struct option *o, unsigned char **s, int *l)
 {
 	struct protocol_program *a;
-	foreachback(a, *(struct list_head *)o->ptr) {
+	struct list_head *la;
+	foreachback(struct protocol_program, a, la, *(struct list_head *)o->ptr) {
 		if (!*a->prog) continue;
 		add_nm(o, s, l);
 		add_quoted_to_str(s, l, a->prog);
@@ -1254,7 +1300,8 @@ static unsigned char *term2_rd(struct option *o, unsigned char *c)
 static void term_wr(struct option *o, unsigned char **s, int *l)
 {
 	struct term_spec *ts;
-	foreachback(ts, term_specs) {
+	struct list_head *lts;
+	foreachback(struct term_spec, ts, lts, term_specs) {
 		add_nm(o, s, l);
 		add_quoted_to_str(s, l, ts->term);
 		add_to_str(s, l, cast_uchar " ");
@@ -1284,8 +1331,9 @@ static struct list_head driver_params = { &driver_params, &driver_params };
 struct driver_param *get_driver_param(unsigned char *n)
 {
 	struct driver_param *dp;
-	foreach(dp, driver_params) if (!casestrcmp(dp->name, n)) return dp;
-	dp = mem_calloc(sizeof(struct driver_param) + strlen(cast_const_char n) + 1);
+	struct list_head *ldp;
+	foreach(struct driver_param, dp, ldp, driver_params) if (!casestrcmp(dp->name, n)) return dp;
+	dp = mem_calloc(sizeof(struct driver_param) + strlen(cast_const_char n));
 	dp->kbd_codepage = -1;
 	strcpy(cast_char dp->name, cast_const_char n);
 	dp->shell = mem_calloc(1);
@@ -1341,7 +1389,8 @@ static unsigned char *dp_rd(struct option *o, unsigned char *c)
 static void dp_wr(struct option *o, unsigned char **s, int *l)
 {
 	struct driver_param *dp;
-	foreachback(dp, driver_params) {
+	struct list_head *ldp;
+	foreachback(struct driver_param, dp, ldp, driver_params) {
 		if ((!dp->param || !*dp->param) && dp->kbd_codepage == -1 && !*dp->shell) continue;
 		if (dp->nosave) continue;
 		add_nm(o, s, l);
@@ -1392,14 +1441,49 @@ static unsigned char *gen_cmd(struct option *o, unsigned char ***argv, int *argc
 	return NULL;
 }
 
+static unsigned char *x_proxy_cmd(struct option *o, unsigned char ***argv, int *argc, int (*save)(int, unsigned char *, unsigned char *), unsigned char *err)
+{
+	unsigned char **pass_argv;
+	unsigned char *result, *ret;
+	if (!*argc) return cast_uchar "Parameter expected";
+	result = mem_alloc(MAX_STR_LEN);
+	if (save(get_commandline_charset(), result, **argv)) {
+		mem_free(result);
+		return err;
+	}
+	pass_argv = &result;
+	ret = gen_cmd(o, &pass_argv, argc);
+	mem_free(result);
+	if (ret) return ret;
+	(*argv)++;
+	return NULL;
+}
+
+static unsigned char *proxy_cmd(struct option *o, unsigned char ***argv, int *argc)
+{
+	return x_proxy_cmd(o, argv, argc, save_proxy, cast_uchar "Invalid proxy");
+}
+
+static unsigned char *noproxy_cmd(struct option *o, unsigned char ***argv, int *argc)
+{
+	return x_proxy_cmd(o, argv, argc, save_noproxy_list, cast_uchar "Invalid list of domains");
+}
+
 static unsigned char *lookup_cmd(struct option *o, unsigned char ***argv, int *argc)
 {
 	int i;
 	struct lookup_result addr;
+	unsigned char *h, *h2, *h3;
 	if (!*argc) return cast_uchar "Parameter expected";
 	if (*argc >= 2) return cast_uchar "Too many parameters";
-	(*argv)++; (*argc)--;
-	do_real_lookup(*(*argv - 1), ipv6_options.addr_preference, &addr);
+	h = *(*argv)++;
+	(*argc)--;
+	h2 = convert(get_commandline_charset(), utf8_table, h, NULL);
+	h3 = idn_encode_host(h2, (int)strlen(cast_const_char h2), cast_uchar ".", 0);
+	if (!h3) h3 = stracpy(h2);
+	mem_free(h2);
+	do_real_lookup(h3, ipv6_options.addr_preference, &addr);
+	mem_free(h3);
 	if (!addr.n) {
 #if !defined(USE_GETADDRINFO) && defined(HAVE_GETHOSTBYNAME) && defined(HAVE_HERROR)
 		herror("error");
@@ -1410,30 +1494,9 @@ static unsigned char *lookup_cmd(struct option *o, unsigned char ***argv, int *a
 		return NULL;
 	}
 	for (i = 0; i < addr.n; i++) {
-		switch (addr.a[i].af) {
-			case AF_INET: {
-				printf("%d.%d.%d.%d", (int)addr.a[i].addr[0], (int)addr.a[i].addr[1], (int)addr.a[i].addr[2], (int)addr.a[i].addr[3]);
-				break;
-			}
-#ifdef SUPPORT_IPV6
-			case AF_INET6: {
-				int j;
-#ifdef HAVE_INET_NTOP
-				unsigned char buffer[INET6_ADDRSTRLEN];
-				struct in6_addr i6a;
-				memcpy(&i6a, addr.a[i].addr, 16);
-				if (inet_ntop(AF_INET6, &i6a, cast_char buffer, sizeof buffer))
-					printf("%s", cast_const_char buffer);
-				else
-#endif
-				for (j = 0; j < 16; j += 2)
-					printf("%x%s", (addr.a[i].addr[j] << 8) + addr.a[i].addr[j + 1], j == 14 ? "" : ":");
-			}
-#endif
-		}
-		if (addr.a[i].scope_id)
-			printf("%%%u", addr.a[i].scope_id);
-		printf("\n");
+		unsigned char *a;
+		if ((a = print_address(&addr.a[i])))
+			printf("%s\n", a);
 	}
 	fflush(stdout);
 	do_exit(RET_OK);
@@ -1577,6 +1640,10 @@ fprintf(stdout, "%s%s%s%s%s%s\n",
 "  Timeout on non restartable connections.\n"
 "    (default: 600)\n"
 "\n"
+" -timeout-when-trying-multiple-addresses <sec>\n"
+"  Timeout for connection when trying multiple addresses.\n"
+"    (default: 3)\n"
+"\n"
 " -bind-address <ip address>\n"
 "  Use a specific local IP address.\n"
 "\n"
@@ -1646,6 +1713,10 @@ fprintf(stdout, "%s%s%s%s%s%s\n",
 "    tor exit node.\n"
 "    (default: blank)\n"
 "\n"
+" -no-proxy-domains <domain,domain...>\n"
+"  No proxy for specified domains.\n"
+"    (default: blank)\n"
+"\n"
 " -only-proxies <0>/<1>\n"
 "    (default 0)\n"
 "  \"1\" causes that Links won't initiate any non-proxy connection.\n"
@@ -1657,6 +1728,14 @@ fprintf(stdout, "%s%s%s%s%s%s\n",
 "  1 - warn on invalid certificate\n"
 "  2 - reject invalid certificate\n"
 "\n"
+" -ssl.builtin-certificates <0>/<1>\n"
+#if defined(DOS) || defined(OPENVMS)
+"    (default 1)\n"
+#else
+"    (default 0)\n"
+#endif
+"  Use built-in certificates instead of system certificates.\n"
+"\n"
 " -ssl.client-cert-key <filename>\n"
 "  Name of the PEM encoded file with the user private key\n"
 "  for client certificate authentication.\n"
@@ -1972,12 +2051,12 @@ fprintf(stdout, "%s%s%s%s%s%s\n",
 void end_config(void)
 {
 	struct driver_param *dp;
-	foreach(dp,driver_params)
-	{
-		if (dp->param)mem_free(dp->param);
-		if (dp->shell)mem_free(dp->shell);
+	struct list_head *ldp;
+	foreach(struct driver_param, dp, ldp, driver_params) {
+		if (dp->param) mem_free(dp->param);
+		if (dp->shell) mem_free(dp->shell);
 	}
-	free_list(driver_params);
+	free_list(struct driver_param, driver_params);
 	if (links_home) mem_free(links_home);
 }
 
@@ -2014,6 +2093,7 @@ int max_connections_to_host = 8;
 int max_tries = 3;
 int receive_timeout = 120;
 int unrestartable_receive_timeout = 600;
+int timeout_multiple_addresses = 3;
 unsigned char bind_ip_address[16] = "";
 unsigned char bind_ipv6_address[INET6_ADDRSTRLEN] = "";
 int async_lookup = 1;
@@ -2026,8 +2106,14 @@ int font_cache_size = 2097152;
 int aggressive_cache = 1;
 
 struct ipv6_options ipv6_options = { ADDR_PREFERENCE_DEFAULT };
-struct proxies proxies = { "", "", "", "", "", 0 };
-struct ssl_options ssl_options = { SSL_WARN_ON_INVALID_CERTIFICATE, "", "", "" };
+struct proxies proxies = { "", "", "", "", "", "", 0 };
+struct ssl_options ssl_options = { SSL_WARN_ON_INVALID_CERTIFICATE,
+#if defined(DOS) || defined(OPENVMS)
+	1,
+#else
+	0,
+#endif
+	"", "", "" };
 struct http_options http_options = { 0, 1, 1, 0, 0, 0, 0, { 0, 0, REFERER_REAL_SAME_SERVER, "", "", "" } };
 struct ftp_options ftp_options = { "somebody@host.domain", 0, 0, 0, 1 };
 struct smb_options smb_options = { 0 };
@@ -2124,6 +2210,7 @@ static struct option links_options[] = {
 	{1, gen_cmd, num_rd, num_wr, 0, 16, &max_tries, "retries", "retries"},
 	{1, gen_cmd, num_rd, num_wr, 1, 9999, &receive_timeout, "receive_timeout", "receive-timeout"},
 	{1, gen_cmd, num_rd, num_wr, 1, 9999, &unrestartable_receive_timeout, "unrestartable_receive_timeout", "unrestartable-receive-timeout"},
+	{1, gen_cmd, num_rd, num_wr, 1, 999, &timeout_multiple_addresses, "timeout_when_trying_multiple_addresses", "timeout-when-trying-multiple-addresses"},
 	{1, gen_cmd, ip_rd, str_wr, 0, 16, bind_ip_address, "bind_address", "bind-address"},
 	{1, gen_cmd, ipv6_rd, str_wr, 0, INET6_ADDRSTRLEN, bind_ipv6_address, "bind_address_ipv6", "bind-address-ipv6"},
 	{1, set_cmd, NULL, NULL, 0, 0, &disable_libevent, NULL, "no-libevent"},
@@ -2136,14 +2223,16 @@ static struct option links_options[] = {
 	{1, gen_cmd, num_rd, num_wr, 0, MAXINT, &font_cache_size, "font_cache_size", "font-cache-size"},
 	{1, gen_cmd, num_rd, num_wr, 0, 1, &aggressive_cache, "http_bugs.aggressive_cache", "aggressive-cache"},
 	{1, gen_cmd, num_rd, num_wr, 0, 4, &ipv6_options.addr_preference, "ipv6.address_preference", "address-preference"},
-	{1, gen_cmd, str_rd, str_wr, 0, MAX_STR_LEN, proxies.http_proxy, "http_proxy", "http-proxy"},
-	{1, gen_cmd, str_rd, str_wr, 0, MAX_STR_LEN, proxies.ftp_proxy, "ftp_proxy", "ftp-proxy"},
-	{1, gen_cmd, str_rd, str_wr, 0, MAX_STR_LEN, proxies.https_proxy, "https_proxy", "https-proxy"},
-	{1, gen_cmd, str_rd, str_wr, 0, MAX_STR_LEN, proxies.socks_proxy, "socks_proxy", "socks-proxy"},
+	{1, proxy_cmd, str_rd, str_wr, 0, MAX_STR_LEN, proxies.http_proxy, "http_proxy", "http-proxy"},
+	{1, proxy_cmd, str_rd, str_wr, 0, MAX_STR_LEN, proxies.ftp_proxy, "ftp_proxy", "ftp-proxy"},
+	{1, proxy_cmd, str_rd, str_wr, 0, MAX_STR_LEN, proxies.https_proxy, "https_proxy", "https-proxy"},
+	{1, proxy_cmd, str_rd, str_wr, 0, MAX_STR_LEN, proxies.socks_proxy, "socks_proxy", "socks-proxy"},
 	{1, gen_cmd, str_rd, NULL, 0, MAX_STR_LEN, proxies.dns_append, "-append_text_to_dns_lookups", NULL}, /* old version incorrectly saved it with '-' */
+	{1, noproxy_cmd, str_rd, str_wr, 0, MAX_STR_LEN, proxies.no_proxy, "no_proxy_domains", "no-proxy-domains"},
 	{1, gen_cmd, str_rd, str_wr, 0, MAX_STR_LEN, proxies.dns_append, "append_text_to_dns_lookups", "append-text-to-dns-lookups"},
 	{1, gen_cmd, num_rd, num_wr, 0, 1, &proxies.only_proxies, "only_proxies", "only-proxies"},
 	{1, gen_cmd, num_rd, num_wr, 0, 2, &ssl_options.certificates, "ssl.certificates", "ssl.certificates"},
+	{1, gen_cmd, num_rd, num_wr, 0, 1, &ssl_options.built_in_certificates, "ssl.builtin_certificates", "ssl.builtin-certificates"},
 	{1, gen_cmd, str_rd, str_wr, 0, MAX_STR_LEN, &ssl_options.client_cert_key, "ssl.client_cert_key", "ssl.client-cert-key"},
 	{1, gen_cmd, str_rd, str_wr, 0, MAX_STR_LEN, &ssl_options.client_cert_crt, "ssl.client_cert_crt", "ssl.client-cert-crt"},
 	{1, gen_cmd, str_rd, NULL, 0, MAX_STR_LEN, &ssl_options.client_cert_password, NULL, "ssl.client-cert-password"},
@@ -2308,7 +2397,7 @@ void load_url_history(void)
 		for (hsl = hsp; *hsl && *hsl != 10 && *hsl != 13; hsl++)
 			;
 		hsc = memacpy(hsp, hsl - hsp);
-		add_to_history(&goto_url_history, hsc, 0);
+		add_to_history(NULL, &goto_url_history, hsc);
 		mem_free(hsc);
 		hsp = hsl;
 		while (*hsp == 10 || *hsp == 13) hsp++;
@@ -2319,6 +2408,7 @@ void load_url_history(void)
 void save_url_history(void)
 {
 	struct history_item *hi;
+	struct list_head *lhi;
 	unsigned char *history_file;
 	unsigned char *hs;
 	int hsl = 0;
@@ -2330,14 +2420,14 @@ void save_url_history(void)
 	add_to_strn(&history_file, cast_uchar "links.his");
 	hs = init_str();
 	hsl = 0;
-	foreachback(hi, goto_url_history.items) {
-		if (!*hi->d || hi->d[0] == ' ' || strchr(cast_const_char hi->d, 10) || strchr(cast_const_char hi->d, 13)) continue;
-		if (!url_not_saveable(hi->d)) {
-			add_to_str(&hs, &hsl, hi->d);
+	foreachback(struct history_item, hi, lhi, goto_url_history.items) {
+		if (!*hi->str || hi->str[0] == ' ' || strchr(cast_const_char hi->str, 10) || strchr(cast_const_char hi->str, 13)) continue;
+		if (!url_not_saveable(hi->str)) {
+			add_to_str(&hs, &hsl, hi->str);
 			add_to_str(&hs, &hsl, cast_uchar NEWLINE);
 		}
 	}
-	write_to_config_file(history_file, hs);
+	write_to_config_file(history_file, hs, 0);
 	mem_free(history_file);
 	mem_free(hs);
 	return;
