@@ -5,6 +5,8 @@
 #include <pc.h>
 #include <dpmi.h>
 #include <bios.h>
+#include <go32.h>
+#include <libc/dosio.h>
 #include <sys/exceptn.h>
 #include <sys/movedata.h>
 
@@ -75,7 +77,7 @@ static unsigned char dos_mouse_buttons;
 static int dos_mouse_last_x;
 static int dos_mouse_last_y;
 static int dos_mouse_last_button;
-static ttime dos_mouse_time;
+static uttime dos_mouse_time;
 
 static struct links_event *dos_mouse_queue = DUMMY;
 static int dos_mouse_queue_n;
@@ -176,7 +178,7 @@ static void dos_mouse_poll(void)
 	__dpmi_regs r;
 	dos_poll_break();
 	if (dos_mouse_initialized) {
-		if (dos_mouse_initialized == 1 && (uttime)get_time() - (uttime)dos_mouse_time > DOS_MOUSE_FLUSH_TIME)
+		if (dos_mouse_initialized == 1 && get_time() - dos_mouse_time > DOS_MOUSE_FLUSH_TIME)
 			dos_mouse_initialized = 2;
 		for (i = 0; i < dos_mouse_buttons; i++) {
 			if (dos_mouse_initialized == 1 && i > 0) {
@@ -365,6 +367,7 @@ process_ansi:
 				continue;
 			case 10:
 				cursor_y++;
+				/*-fallthrough*/
 			case 13:
 				cursor_x = 0;
 				break;
@@ -374,6 +377,7 @@ process_ansi:
 					case ')':
 						if (!size--) goto ret;
 						str++;
+						continue;
 					case '7':
 					default:
 						continue;
@@ -754,7 +758,7 @@ void setcooked(int ctl)
 
 #define RANDOM_POOL_SIZE	65536
 
-void os_seed_random(unsigned char **pool, unsigned *pool_size)
+void os_seed_random(unsigned char **pool, int *pool_size)
 {
 	unsigned *random_pool, *tmp_pool;
 	int a, i;
@@ -798,6 +802,43 @@ void terminate_osdep(void)
 		mem_free(screen_backbuffer);
 }
 
+int os_default_language(void)
+{
+	__dpmi_regs r;
+	memset(&r, 0, sizeof r);
+	r.x.ax = 0x3800;
+	r.x.dx = __tb_offset;
+	r.x.ds = __tb_segment;
+	__dpmi_int(0x21, &r);
+	if (!(r.x.flags & 1)) {
+		return get_country_language(r.x.bx);
+	}
+	return -1;
+}
+
+int os_default_charset(void)
+{
+	__dpmi_regs r;
+	memset(&r, 0, sizeof r);
+	r.x.ax = 0x6601;
+	__dpmi_int(0x21, &r);
+	if (!(r.x.flags & 1)) {
+		unsigned char a[8];
+		int cp;
+		snprintf(cast_char a, sizeof a, "%d", r.x.bx);
+		if ((cp = get_cp_index(a)) >= 0 && cp != utf8_table)
+			return cp;
+	}
+	return 0;
+}
+
+int dos_is_bw(void)
+{
+	unsigned char cfg;
+	dosmemget(0x410, sizeof cfg, &cfg);
+	return (cfg & 0x30) == 0x30;
+}
+
 #else
 
 typedef int dos_c_no_empty_unit;
