@@ -110,13 +110,6 @@ char *tempnam(const char *dir, const char *pfx)
 	return cast_char a;
 }
 #endif
-#ifndef HAVE_SIGFILLSET
-int sigfillset(sigset_t *set)
-{
-	memset(set, -1, sizeof(sigset_t));
-	return 0;
-}
-#endif
 #ifndef HAVE_STRDUP
 char *strdup(const char *s)
 {
@@ -271,8 +264,9 @@ char *strstr(const char *haystack, const char *needle)
 {
 	size_t hs = strlen(haystack);
 	size_t ns = strlen(needle);
+	if (!ns) return (char *)haystack;
 	while (hs >= ns) {
-		if (!memcmp(haystack, needle, ns)) return (char *)haystack;
+		if (*haystack == *needle && !memcmp(haystack, needle, ns)) return (char *)haystack;
 		haystack++, hs--;
 	}
 	return NULL;
@@ -350,6 +344,17 @@ void *memset(void *s, int c, size_t n)
 	return s;
 }
 #endif
+#ifndef HAVE_MEMMEM
+void *memmem(const void *haystack, size_t hs, const void *needle, size_t ns)
+{
+	if (!ns) return (void *)haystack;
+	while (hs >= ns) {
+		if (*(const char *)haystack == *(const char *)needle && !memcmp(haystack, needle, ns)) return (void *)haystack;
+		haystack = (const void *)((const char *)haystack + 1), hs--;
+	}
+	return NULL;
+}
+#endif
 #ifndef HAVE_STRERROR
 extern char *sys_errlist[];
 extern int sys_nerr;
