@@ -215,11 +215,29 @@ static unsigned char *get_url_val(unsigned char *e, unsigned char *name)
 
 static unsigned char *get_url_val(unsigned char *e, unsigned char *name)
 {
-	unsigned char *a;
+	unsigned char *a, *p, *c;
+	int l;
 	get_attr_val_nl = 1;
 	a = get_attr_val(e, name);
 	get_attr_val_nl = 0;
+	if (!a)
+		return NULL;
+	if (d_opt->real_cp != utf8_table) {
+		if (url_non_ascii(a))
+			goto need_convert;
+	}
 	return a;
+
+	need_convert:
+	c = init_str();
+	l = 0;
+	for (p = a; *p; p++) {
+		int u = cp2u(*p, d_opt->real_cp);
+		unsigned char *us = encode_utf_8(u);
+		add_to_str(&c, &l, us);
+	}
+	mem_free(a);
+	return c;
 }
 
 static unsigned char *get_exact_attr_val(unsigned char *e, unsigned char *name)
@@ -305,7 +323,7 @@ struct color_spec {
 	int rgb;
 };
 
-static const struct color_spec color_specs[] = {
+static_const struct color_spec color_specs[] = {
 	{"aliceblue",		0xF0F8FF},
 	{"antiquewhite",	0xFAEBD7},
 	{"aqua",		0x00FFFF},
@@ -457,7 +475,7 @@ int decode_color(unsigned char *str, struct rgb *col)
 	if (*str != '#') {
 		const struct color_spec *cs;
 		for (cs = color_specs; cs < endof(color_specs); cs++)
-			if (!strcasecmp(cast_const_char cs->name, cast_const_char str)) {
+			if (!casestrcmp(cast_uchar cs->name, str)) {
 				ch = cs->rgb;
 				goto found;
 			}
@@ -531,17 +549,15 @@ void kill_html_stack_item(struct html_element *e)
 	free_js_event_spec(e->attr.js_event);
 	del_from_list(e);
 	mem_free(e);
-	/*if ((void *)(html_stack.next) == &html_stack || !html_stack.next) {
-		debug("killing last element");
-	}*/
 }
 
 #if defined(DEBUG) && 0
 void debug_stack(void)
 {
 	struct html_element *e;
+	struct list_head *le;
 	printf("HTML stack debug: \n");
-	foreachback(e, html_stack) {
+	foreachback(struct html_element, e, le, html_stack) {
 		int i;
 		printf("\"");
 		for (i = 0; i < e->namelen; i++) printf("%c", e->name[i]);
@@ -549,7 +565,7 @@ void debug_stack(void)
 	}
 	printf("%c", 7);
 	fflush(stdout);
-	sleep(1);
+	portable_sleep(1000);
 }
 #endif
 
@@ -558,10 +574,7 @@ void html_stack_dup(void)
 	struct html_element *e;
 	struct html_element *ep;
 	html_format_changed = 1;
-	if ((void *)(ep = html_stack.next) == &html_stack || !html_stack.next) {
-		internal("html stack empty");
-		return;
-	}
+	ep = &html_top;
 	e = mem_alloc(sizeof(struct html_element));
 	memcpy(e, ep, sizeof(struct html_element));
 	e->attr.fontface = stracpy(ep->attr.fontface);
@@ -768,9 +781,10 @@ static void put_chrs(unsigned char *start, int len)
 static void kill_until(int ls, ...)
 {
 	int l;
-	struct html_element *e = &html_top;
+	struct list_head *e = &html_top.list_entry;
 	if (ls) e = e->next;
-	while ((void *)e != &html_stack) {
+	while (e != &html_stack) {
+		struct html_element *he = list_struct(e, struct html_element);
 		int sk = 0;
 		va_list arg;
 		va_start(arg, ls);
@@ -778,9 +792,9 @@ static void kill_until(int ls, ...)
 			unsigned char *s = va_arg(arg, unsigned char *);
 			if (!s) break;
 			if (!*s) sk++;
-			else if ((size_t)e->namelen == strlen(cast_const_char s) && !casecmp(e->name, s, strlen(cast_const_char s))) {
+			else if ((size_t)he->namelen == strlen(cast_const_char s) && !casecmp(he->name, s, strlen(cast_const_char s))) {
 				if (!sk) {
-					if (e->dontkill) break;
+					if (he->dontkill) break;
 					va_end(arg);
 					goto killll;
 				}
@@ -791,8 +805,8 @@ static void kill_until(int ls, ...)
 			}
 		}
 		va_end(arg);
-		if (e->dontkill || (e->namelen == 5 && !casecmp(e->name, cast_uchar "TABLE", 5))) break;
-		if (e->namelen == 2 && upcase(e->name[0]) == 'T' && (upcase(e->name[1]) == 'D' || upcase(e->name[1]) == 'H' || upcase(e->name[1]) == 'R')) break;
+		if (he->dontkill || (he->namelen == 5 && !casecmp(he->name, cast_uchar "TABLE", 5))) break;
+		if (he->namelen == 2 && upcase(he->name[0]) == 'T' && (upcase(he->name[1]) == 'D' || upcase(he->name[1]) == 'H' || upcase(he->name[1]) == 'R')) break;
 		e = e->next;
 	}
 	return;
@@ -800,18 +814,19 @@ static void kill_until(int ls, ...)
 	e = e->prev;
 	killll:
 	l = 0;
-	while ((void *)e != &html_stack) {
+	while (e != &html_stack) {
+		struct html_element *he = list_struct(e, struct html_element);
 		if (ls && e == html_stack.next) break;
-		if (e->linebreak > l) l = e->linebreak;
+		if (he->linebreak > l) l = he->linebreak;
 		e = e->prev;
-		kill_html_stack_item(e->next);
+		kill_html_stack_item(he);
 	}
 	ln_break(l);
 }
 
 static inline unsigned char *top_href_base(void)
 {
-	return ((struct html_element *)html_stack.prev)->attr.href_base;
+	return list_struct(html_stack.prev, struct html_element)->attr.href_base;
 }
 
 int get_num(unsigned char *a, unsigned char *n)
@@ -876,6 +891,18 @@ int get_width(unsigned char *a, unsigned char *n, int trunc)
 	return r;
 }
 
+static unsigned char *find_element_end(unsigned char *a)
+{
+	unsigned char *p;
+	for (p = a - 1; *p != '<'; p--)
+		;
+	if (parse_element(p, eoff, NULL, NULL, NULL, &p)) {
+		internal("parse element failed");
+		return a;
+	}
+	return p;
+}
+
 struct form form = { NULL, NULL, NULL, NULL, 0, 0 };
 
 unsigned char *last_form_tag;
@@ -889,6 +916,8 @@ static inline void set_link_attr(void)
 
 static void put_link_line(unsigned char *prefix, unsigned char *linkname, unsigned char *link, unsigned char *target)
 {
+	if (!casecmp(link, cast_uchar "android-app:", 12))
+		return;
 	html_stack_dup();
 	ln_break(1);
 	if (format_.link) mem_free(format_.link), format_.link = NULL;
@@ -910,6 +939,11 @@ static void html_span(unsigned char *a)
 	if ((al = get_attr_val(a, cast_uchar "class"))) {
 		if (!strcmp(cast_const_char al, "line-number"))
 			ln_break(1);
+		if (!strcmp(cast_const_char al, "blob-code-inner")) {	/* github hack */
+			ln_break(1);
+			format_.attr |= AT_FIXED;
+			par_format.align = AL_NO;
+		}
 
 		mem_free(al);
 	}
@@ -1039,7 +1073,7 @@ static void html_img(unsigned char *a)
 	int ismap, usemap = 0;
 	/*put_chrs(cast_uchar " ", 1);*/
 	get_js_events(a);
-	if ((!F || !d_opt->display_images) && ((al = get_attr_val(a, cast_uchar "usemap")))) {
+	if ((!F || !d_opt->display_images) && ((al = get_url_val(a, cast_uchar "usemap")))) {
 		unsigned char *u;
 		usemap = 1;
 		html_stack_dup();
@@ -1056,11 +1090,18 @@ static void html_img(unsigned char *a)
 	ismap = format_.link && (F || !has_attr(a, cast_uchar "usemap")) && has_attr(a, cast_uchar "ismap");
 	if (format_.image) mem_free(format_.image), format_.image = NULL;
 	if (
+		(s = get_url_val(a, cast_uchar "data-full")) ||
+		(s = get_url_val(a, cast_uchar "data-normal")) ||
+		(s = get_url_val(a, cast_uchar "data-src")) ||
 		(s = get_url_val(a, cast_uchar "data-defer-src")) ||
+		(s = get_url_val(a, cast_uchar "data-li-src")) ||
 		(s = get_url_val(a, cast_uchar "data-original")) ||
+		(s = get_url_val(a, cast_uchar "data-small")) ||
+		(s = get_url_val(a, cast_uchar "data-lazy")) ||
 		(s = get_url_val(a, cast_uchar "src")) ||
-		(s = get_attr_val(a, cast_uchar "dynsrc")) ||
-		(s = get_attr_val(a, cast_uchar "data"))
+		(s = get_url_val(a, cast_uchar "dynsrc")) ||
+		(s = get_url_val(a, cast_uchar "data")) ||
+		(s = get_url_val(a, cast_uchar "content"))
 	    ) {
 		 if (!format_.link && d_opt->braille) goto skip_img;
 		 format_.image = join_urls(format_.href_base, s);
@@ -1106,12 +1147,12 @@ static void html_img(unsigned char *a)
 		int aa = -1;
 
 		if ((al = get_attr_val(a, cast_uchar "align"))) {
-			if (!strcasecmp(cast_const_char al, "left")) aa = AL_LEFT;
-			if (!strcasecmp(cast_const_char al, "right")) aa = AL_RIGHT;
-			if (!strcasecmp(cast_const_char al, "center")) aa = AL_CENTER;
-			if (!strcasecmp(cast_const_char al, "bottom")) aa = AL_BOTTOM;
-			if (!strcasecmp(cast_const_char al, "middle")) aa = AL_MIDDLE;
-			if (!strcasecmp(cast_const_char al, "top")) aa = AL_TOP;
+			if (!casestrcmp(al, cast_uchar "left")) aa = AL_LEFT;
+			if (!casestrcmp(al, cast_uchar "right")) aa = AL_RIGHT;
+			if (!casestrcmp(al, cast_uchar "center")) aa = AL_CENTER;
+			if (!casestrcmp(al, cast_uchar "bottom")) aa = AL_BOTTOM;
+			if (!casestrcmp(al, cast_uchar "middle")) aa = AL_MIDDLE;
+			if (!casestrcmp(al, cast_uchar "top")) aa = AL_TOP;
 			mem_free(al);
 		}
 
@@ -1160,7 +1201,7 @@ static void html_img(unsigned char *a)
 		if (!i.alt) i.alt = get_attr_val(a, cast_uchar "alt");
 		i.insert_flag = !(format_.form);
 		i.ismap = ismap;
-		if ((u = get_attr_val(a, cast_uchar "usemap"))) {
+		if ((u = get_url_val(a, cast_uchar "usemap"))) {
 			i.usemap = join_urls(*u == '#' ? top_href_base() : format_.href_base, u);
 			mem_free(u);
 		}
@@ -1193,19 +1234,24 @@ static void html_obj(unsigned char *a, int obj)
 	unsigned char *url;
 	unsigned char *type = get_attr_val(a, cast_uchar "type");
 	unsigned char *base;
-	if ((base = get_attr_val(a, cast_uchar "codebase"))) format_.href_base = join_urls(format_.href_base, base);
+	if ((base = get_url_val(a, cast_uchar "codebase"))) format_.href_base = join_urls(format_.href_base, base);
 	if (!type) {
-		url = get_attr_val(a, cast_uchar "src");
-		if (!url) url = get_attr_val(a, cast_uchar "data");
-		if (url) type = get_content_type(NULL, url), mem_free(url);
+		url = get_url_val(a, cast_uchar "src");
+		if (!url) url = get_url_val(a, cast_uchar "data");
+		if (url) {
+			unsigned char *ju = join_urls(format_.href_base, url);
+			type = get_content_type(NULL, ju);
+			mem_free(url);
+			mem_free(ju);
+		}
 	}
 	if (type && known_image_type(type)) {
 		html_img(a);
-		if (obj == 1) html_top.invisible = 1;
+		if (obj == 1) html_top.invisible = INVISIBLE;
 		goto ret;
 	}
-	url = get_attr_val(a, cast_uchar "src");
-	if (!url) url = get_attr_val(a, cast_uchar "data");
+	url = get_url_val(a, cast_uchar "src");
+	if (!url) url = get_url_val(a, cast_uchar "data");
 	if (url) put_link_line(cast_uchar "", !obj ? cast_uchar "[EMBED]" : cast_uchar "[OBJ]", url, cast_uchar ""), mem_free(url);
 	ret:
 	if (base) mem_free(format_.href_base), format_.href_base = old_base, mem_free(base);
@@ -1235,22 +1281,38 @@ static void html_body(unsigned char *a)
 
 static void html_skip(unsigned char *a)
 {
-	html_top.invisible = html_top.dontkill = 1;
+	html_top.dontkill = 1;
+	html_top.invisible = INVISIBLE;
 }
 
 static void html_title(unsigned char *a)
 {
 	if (a[0] == '>' && a[-1] == '/') return;
-	html_top.invisible = html_top.dontkill = 1;
+	html_top.dontkill = 1;
+	html_top.invisible = INVISIBLE;
+}
+
+int should_skip_script(unsigned char *a)
+{
+	return !has_attr(a, cast_uchar "/");
 }
 
 static void html_script(unsigned char *a)
 {
 	unsigned char *s;
-	s = get_attr_val(a, cast_uchar "src");
+	s = get_url_val(a, cast_uchar "src");
 	special_f(ff, SP_SCRIPT, s);
 	if (s) mem_free(s);
-	html_skip(a);
+	if (should_skip_script(a)) {
+		html_top.dontkill = 1;
+		html_top.invisible = INVISIBLE_SCRIPT;
+	}
+}
+
+static void html_style(unsigned char *a)
+{
+	html_top.dontkill = 1;
+	html_top.invisible = INVISIBLE_STYLE;
 }
 
 static void html_noscript(unsigned char *a)
@@ -1268,13 +1330,13 @@ static void html_linebrk(unsigned char *a)
 {
 	unsigned char *al;
 	if ((al = get_attr_val(a, cast_uchar "align"))) {
-		if (!strcasecmp(cast_const_char al, "left")) par_format.align = AL_LEFT;
-		if (!strcasecmp(cast_const_char al, "right")) par_format.align = AL_RIGHT;
-		if (!strcasecmp(cast_const_char al, "center")) {
+		if (!casestrcmp(al, cast_uchar "left")) par_format.align = AL_LEFT;
+		if (!casestrcmp(al, cast_uchar "right")) par_format.align = AL_RIGHT;
+		if (!casestrcmp(al, cast_uchar "center")) {
 			par_format.align = AL_CENTER;
 			if (!table_level && !F) par_format.leftmargin = par_format.rightmargin = 0;
 		}
-		if (!strcasecmp(cast_const_char al, "justify")) par_format.align = AL_BLOCK;
+		if (!casestrcmp(al, cast_uchar "justify")) par_format.align = AL_BLOCK;
 		mem_free(al);
 	}
 }
@@ -1315,7 +1377,9 @@ static void html_blockquote(unsigned char *a)
 
 static void html_h(int h, unsigned char *a)
 {
+#if defined(__GNUC__) && defined(__arm__)
 	do_not_optimize_here(&h);
+#endif
 #ifdef G
 	if (F) {
 		html_linebrk(a);
@@ -1374,9 +1438,14 @@ static void html_div(unsigned char *a)
 	unsigned char *al;
 	if ((al = get_attr_val(a, cast_uchar "class"))) {
 		if (!strcmp(cast_const_char al, "commit-msg") ||
-		    !strncmp(cast_const_char al, "diff", 4) /* gitweb hack */) {
+		    !strcmp(cast_const_char al, "pre") /* sourceware hack */ ||
+		    (!strncmp(cast_const_char al, "diff", 4) && casestrcmp(al, cast_uchar "diff-view") && strncmp(cast_const_char al, "diffbar", 7)) /* gitweb hack, github counter-hacks */ ||
+		    0) {
 			format_.attr |= AT_FIXED;
 			par_format.align = AL_NO;
+		} else if (strstr(cast_const_char al, "plain-text-white-space")) {
+			format_.attr |= AT_FIXED;
+			par_format.align = AL_NO_BREAKABLE;
 		}
 		mem_free(al);
 	}
@@ -1472,9 +1541,9 @@ static void html_ul(unsigned char *a)
 	par_format.list_number = 0;
 	par_format.flags = P_STAR;
 	if ((al = get_attr_val(a, cast_uchar "type"))) {
-		if (!strcasecmp(cast_const_char al, "disc") ||
-		    !strcasecmp(cast_const_char al, "circle")) par_format.flags = P_O;
-		if (!strcasecmp(cast_const_char al, "square")) par_format.flags = P_PLUS;
+		if (!casestrcmp(al, cast_uchar "disc") ||
+		    !casestrcmp(al, cast_uchar "circle")) par_format.flags = P_O;
+		if (!casestrcmp(al, cast_uchar "square")) par_format.flags = P_PLUS;
 		mem_free(al);
 	}
 	if ((par_format.leftmargin += 2 + (par_format.list_level > 1)) > par_format.width * 2 / 3 && !table_level)
@@ -1546,7 +1615,7 @@ static void html_li(unsigned char *a)
 		put_chrs(cast_uchar ".&nbsp;", 7);
 		if (!F) par_format.leftmargin += (int)strlen(cast_const_char n) + c + 2;
 		par_format.align = AL_LEFT;
-		html_top.next->parattr.list_number = par_format.list_number + 1;
+		list_struct(html_top.list_entry.next, struct html_element)->parattr.list_number = par_format.list_number + 1;
 		par_format.list_number = 0;
 		putsp = -1;
 	}
@@ -1592,11 +1661,11 @@ static void get_html_form(unsigned char *a, struct form *form)
 	unsigned char *ch;
 	form->method = FM_GET;
 	if ((al = get_attr_val(a, cast_uchar "method"))) {
-		if (!strcasecmp(cast_const_char al, "post")) {
+		if (!casestrcmp(al, cast_uchar "post")) {
 			unsigned char *ax;
 			form->method = FM_POST;
 			if ((ax = get_attr_val(a, cast_uchar "enctype"))) {
-				if (!strcasecmp(cast_const_char ax, "multipart/form-data"))
+				if (!casestrcmp(ax, cast_uchar "multipart/form-data"))
 					form->method = FM_POST_MP;
 				mem_free(ax);
 			}
@@ -1687,9 +1756,9 @@ static void html_button(unsigned char *a)
 		fc->type = FC_SUBMIT;
 		goto xxx;
 	}
-	if (!strcasecmp(cast_const_char al, "submit")) fc->type = FC_SUBMIT;
-	else if (!strcasecmp(cast_const_char al, "reset")) fc->type = FC_RESET;
-	else if (!strcasecmp(cast_const_char al, "button")) fc->type = FC_BUTTON;
+	if (!casestrcmp(al, cast_uchar "submit")) fc->type = FC_SUBMIT;
+	else if (!casestrcmp(al, cast_uchar "reset")) fc->type = FC_RESET;
+	else if (!casestrcmp(al, cast_uchar "button")) fc->type = FC_BUTTON;
 	else {
 		mem_free(al);
 		mem_free(fc);
@@ -1708,7 +1777,6 @@ static void html_button(unsigned char *a)
 	fc->name = get_attr_val(a, cast_uchar "name");
 	fc->default_value = get_exact_attr_val(a, cast_uchar "value");
 	fc->ro = has_attr(a, cast_uchar "disabled") ? 2 : has_attr(a, cast_uchar "readonly") ? 1 : 0;
-	if (fc->type == FC_IMAGE) fc->alt = get_attr_val(a, cast_uchar "alt");
 	if (fc->type == FC_SUBMIT && !fc->default_value) fc->default_value = stracpy(cast_uchar "Submit");
 	if (fc->type == FC_RESET && !fc->default_value) fc->default_value = stracpy(cast_uchar "Reset");
 	if (fc->type == FC_BUTTON && !fc->default_value) fc->default_value = stracpy(cast_uchar "BUTTON");
@@ -1716,10 +1784,27 @@ static void html_button(unsigned char *a)
 	special_f(ff, SP_CONTROL, fc);
 	format_.form = fc;
 	format_.attr |= AT_BOLD | AT_FIXED;
-	/*put_chrs(cast_uchar "[&nbsp;", 7);
-	if (fc->default_value) put_chrs(fc->default_value, strlen(cast_const_char fc->default_value));
-	put_chrs(cast_uchar "&nbsp;]", 7);
-	put_chrs(cast_uchar " ", 1);*/
+
+	if (fc->type != FC_BUTTON) {
+		unsigned char *p, *name;
+		int namelen;
+		p = find_element_end(a);
+p1:
+		while (p < eoff && WHITECHAR(*p)) p2:p++;
+		if (p == eoff) goto put_text;
+		if (*p != '<') return;
+		if (parse_element(p, eoff, &name, &namelen, NULL, &p)) goto p2;
+		if (namelen == 6 && !casecmp(name, cast_uchar "BUTTON", 6)) goto put_text;
+		if (namelen == 7 && !casecmp(name, cast_uchar "/BUTTON", 7)) goto put_text;
+		if (namelen == 3 && !casecmp(name, cast_uchar "IMG", 3)) return;
+		goto p1;
+
+put_text:
+		put_chrs(cast_uchar "[&nbsp;", 7);
+		put_chrs(fc->default_value, (int)strlen(cast_const_char fc->default_value));
+		put_chrs(cast_uchar "&nbsp;]", 7);
+		putsp = -1;
+	}
 }
 
 static void set_max_textarea_width(int *w)
@@ -1761,16 +1846,16 @@ static void html_input(unsigned char *a)
 		else fc->type = FC_TEXT;
 		goto xxx;
 	}
-	if (!strcasecmp(cast_const_char al, "text")) fc->type = FC_TEXT;
-	else if (!strcasecmp(cast_const_char al, "password")) fc->type = FC_PASSWORD;
-	else if (!strcasecmp(cast_const_char al, "checkbox")) fc->type = FC_CHECKBOX;
-	else if (!strcasecmp(cast_const_char al, "radio")) fc->type = FC_RADIO;
-	else if (!strcasecmp(cast_const_char al, "submit")) fc->type = FC_SUBMIT;
-	else if (!strcasecmp(cast_const_char al, "reset")) fc->type = FC_RESET;
-	else if (!strcasecmp(cast_const_char al, "file")) fc->type = FC_FILE;
-	else if (!strcasecmp(cast_const_char al, "hidden")) fc->type = FC_HIDDEN;
-	else if (!strcasecmp(cast_const_char al, "image")) fc->type = FC_IMAGE;
-	else if (!strcasecmp(cast_const_char al, "button")) fc->type = FC_BUTTON;
+	if (!casestrcmp(al, cast_uchar "text")) fc->type = FC_TEXT;
+	else if (!casestrcmp(al, cast_uchar "password")) fc->type = FC_PASSWORD;
+	else if (!casestrcmp(al, cast_uchar "checkbox")) fc->type = FC_CHECKBOX;
+	else if (!casestrcmp(al, cast_uchar "radio")) fc->type = FC_RADIO;
+	else if (!casestrcmp(al, cast_uchar "submit")) fc->type = FC_SUBMIT;
+	else if (!casestrcmp(al, cast_uchar "reset")) fc->type = FC_RESET;
+	else if (!casestrcmp(al, cast_uchar "file")) fc->type = FC_FILE;
+	else if (!casestrcmp(al, cast_uchar "hidden")) fc->type = FC_HIDDEN;
+	else if (!casestrcmp(al, cast_uchar "image")) fc->type = FC_IMAGE;
+	else if (!casestrcmp(al, cast_uchar "button")) fc->type = FC_BUTTON;
 	else fc->type = FC_TEXT;
 	mem_free(al);
 	xxx:
@@ -1888,13 +1973,8 @@ static void html_option(unsigned char *a)
 		unsigned char *name;
 		int namelen;
 		int l = 0;
-		for (p = a - 1; *p != '<'; p--)
-			;
 		val = init_str();
-		if (parse_element(p, eoff, NULL, NULL, NULL, &p)) {
-			internal("parse element failed");
-			goto x;
-		}
+		p = find_element_end(a);
 		rrrr:
 		while (p < eoff && WHITECHAR(*p)) p++;
 		while (p < eoff && !WHITECHAR(*p) && *p != '<') {
@@ -2296,8 +2376,8 @@ static void do_html_textarea(unsigned char *attr, unsigned char *html, unsigned
 	fc->rows = rows;
 	fc->wrap = 1;
 	if ((w = get_attr_val(attr, cast_uchar "wrap"))) {
-		if (!strcasecmp(cast_const_char w, "hard") || !strcasecmp(cast_const_char w, "physical")) fc->wrap = 2;
-		else if (!strcasecmp(cast_const_char w, "off")) fc->wrap = 0;
+		if (!casestrcmp(w, cast_uchar "hard") || !casestrcmp(w, cast_uchar "physical")) fc->wrap = 2;
+		else if (!casestrcmp(w, cast_uchar "off")) fc->wrap = 0;
 		mem_free(w);
 	}
 	if ((fc->maxlength = get_num(attr, cast_uchar "maxlength")) == -1) fc->maxlength = MAXINT / 4;
@@ -2349,13 +2429,13 @@ static void html_frame(unsigned char *a)
 		url = join_urls(format_.href_base, u2);
 		mem_free(u2);
 	}
-	if (!url) return;
-	name = get_attr_val (a, cast_uchar "name");
-	if (!name)
-		name = stracpy(url);
-	else if (!name[0]) { /* When name doesn't have a value */
-		mem_free(name);
-		name = stracpy(url);
+	name = get_attr_val(a, cast_uchar "name");
+	if (name && !name[0])
+		mem_free(name), name = NULL;
+	if (!name) {
+		name = get_attr_val(a, cast_uchar "src");
+		if (!name)
+			name = stracpy(cast_uchar "Frame");
 	}
 	if (!d_opt->frames || !html_top.frameset) put_link_line(cast_uchar "Frame: ", name, url, cast_uchar "");
 	else {
@@ -2368,9 +2448,9 @@ static void html_frame(unsigned char *a)
 		fp.marginheight = get_num(a, cast_uchar "marginheight");
 		fp.scrolling = SCROLLING_AUTO;
 		if (scroll) {
-			if (!strcasecmp(cast_const_char scroll, "no"))
+			if (!casestrcmp(scroll, cast_uchar "no"))
 				fp.scrolling = SCROLLING_NO;
-			else if (!strcasecmp(cast_const_char scroll, "yes"))
+			else if (!casestrcmp(scroll, cast_uchar "yes"))
 				fp.scrolling = SCROLLING_YES;
 			mem_free(scroll);
 		}
@@ -2417,7 +2497,11 @@ static void parse_frame_widths(unsigned char *a, int ww, int www, int **op, int
 		qq = q;
 		for (i = 0; i < ol; i++) {
 			q -= o[i] - o[i] * (d - qq) / (d ? d : 1);
+#ifdef __GNUC__
+#if __GNUC__ == 2
 			do_not_optimize_here(&d);
+#endif
+#endif
 				/* SIGH! gcc 2.7.2.* has an optimizer bug! */
 			o[i] = o[i] * (d - qq) / (d ? d : 1);
 		}
@@ -2539,11 +2623,27 @@ static void html_frameset(unsigned char *a)
 	mem_free(c);
 }*/
 
+static void html_meta(unsigned char *a)
+{
+	unsigned char *prop;
+	if ((prop = get_attr_val(a, cast_uchar "property"))) {
+		if (!strcmp(cast_const_char prop, "og:image")) {
+			unsigned char *host = get_host_name(format_.href_base);
+			if (host) {
+				if (strstr(cast_const_char host, "instagram."))
+					html_img(a);
+				mem_free(host);
+			}
+		}
+		mem_free(prop);
+	}
+}
+
 static void html_link(unsigned char *a)
 {
 	unsigned char *name, *url, *title;
 	if ((name = get_attr_val(a, cast_uchar "type"))) {
-		if (strcasecmp(cast_const_char name, "text/html")) {
+		if (casestrcmp(name, cast_uchar "text/html")) {
 			mem_free(name);
 			return;
 		}
@@ -2561,34 +2661,50 @@ static void html_link(unsigned char *a)
 			mem_free(lang);
 		}
 	}
-	if (!name)
-		name = stracpy(url);
+	if (!name) {
+		if (has_attr(a, cast_uchar "title"))
+			name = stracpy(cast_uchar "");
+		else {
+			name = get_attr_val(a, cast_uchar "href");
+			if (!name)
+				name = stracpy(cast_uchar "Link");
+		}
+	}
 	if (
-	    !casecmp(name, cast_uchar "apple-touch-icon", 16) ||
 	    !casecmp(name, cast_uchar "schema", 6) ||
-	    !strcasecmp(cast_const_char name, "Edit-Time-Data") ||
-	    !strcasecmp(cast_const_char name, "File-List") ||
-	    !strcasecmp(cast_const_char name, "alternate stylesheet") ||
-	    !strcasecmp(cast_const_char name, "generator-home") ||
-	    !strcasecmp(cast_const_char name, "https://github.com/WP-API/WP-API") ||
-	    !strcasecmp(cast_const_char name, "icon") ||
-	    !strcasecmp(cast_const_char name, "made") ||
-	    !strcasecmp(cast_const_char name, "meta") ||
-	    !strcasecmp(cast_const_char name, "pingback") ||
-	    !strcasecmp(cast_const_char name, "preconnect") ||
-	    !strcasecmp(cast_const_char name, "shortcut icon") ||
-	    !strcasecmp(cast_const_char name, "stylesheet") ||
+	    !casestrcmp(name, cast_uchar "Edit-Time-Data") ||
+	    !casestrcmp(name, cast_uchar "File-List") ||
+	    !casestrcmp(name, cast_uchar "alternate stylesheet") ||
+	    !casestrcmp(name, cast_uchar "generator-home") ||
+	    !casestrcmp(name, cast_uchar "https://api.w.org/") ||
+	    !casestrcmp(name, cast_uchar "https://github.com/WP-API/WP-API") ||
+	    !casestrcmp(name, cast_uchar "made") ||
+	    !casestrcmp(name, cast_uchar "manifest") ||
+	    !casestrcmp(name, cast_uchar "meta") ||
+	    !casestrcmp(name, cast_uchar "pingback") ||
+	    !casestrcmp(name, cast_uchar "preconnect") ||
+	    !casestrcmp(name, cast_uchar "stylesheet") ||
+	    casestrstr(name, cast_uchar "icon") ||
 	    0) goto skip;
-	if (!strcasecmp(cast_const_char name, "prefetch") ||
-	    !strcasecmp(cast_const_char name, "dns-prefetch") ||
-	    !strcasecmp(cast_const_char name, "prerender")) {
+	if (!casestrcmp(name, cast_uchar "prefetch") ||
+	    !casestrcmp(name, cast_uchar "prerender") ||
+	    !casestrcmp(name, cast_uchar "preload")) {
 		unsigned char *pre_url = join_urls(format_.href_base, url);
 		if (!dmp) load_url(pre_url, format_.href_base, NULL, PRI_PRELOAD, NC_ALWAYS_CACHE, 0, 0, 0);
 		mem_free(pre_url);
 		goto skip;
 	}
+	if (!casestrcmp(name, cast_uchar "dns-prefetch")) {
+		unsigned char *pre_url, *host;
+		pre_url = join_urls(format_.href_base, url);
+		host = get_host_name(pre_url);
+		mem_free(pre_url);
+		if (!dmp) dns_prefetch(host);
+		mem_free(host);
+		goto skip;
+	}
 	if ((title = get_attr_val(a, cast_uchar "title"))) {
-		add_to_strn(&name, cast_uchar ": ");
+		if (*name) add_to_strn(&name, cast_uchar ": ");
 		add_to_strn(&name, title);
 		mem_free(title);
 	}
@@ -2621,6 +2737,7 @@ static struct element_info elements[] = {
 	{"FIXED",	html_fixed,	0, 0},
 	{"CODE",	html_fixed,	0, 0},
 	{"TT",		html_fixed,	0, 0},
+	{"SAMP",	html_fixed,	0, 0},
 	{"SUB",		html_sub,	0, 0},
 	{"SUP",		html_sup,	0, 0},
 	{"FONT",	html_font,	0, 0},
@@ -2640,7 +2757,7 @@ static struct element_info elements[] = {
 	{"TITLE",	html_title,	0, 0},
 	{"SCRIPT",	html_script,	0, 0},
 	{"NOSCRIPT",	html_noscript,	0, 0},
-	{"STYLE",	html_skip,	0, 0},
+	{"STYLE",	html_style,	0, 0},
 	{"NOEMBED",	html_skip,	0, 0},
 
 	{"BR",		html_br,	1, 1},
@@ -2681,6 +2798,7 @@ static struct element_info elements[] = {
 	{"OPTION",	html_option,	1, 1},
 	{"BUTTON",	html_button,	0, 0},
 
+	{"META",	html_meta,	0, 1},
 	{"LINK",	html_link,	1, 1},
 	{"IFRAME",	html_iframe,	1, 1},
 	{"FRAME",	html_frame,	1, 1},
@@ -2696,7 +2814,7 @@ unsigned char *skip_comment(unsigned char *html, unsigned char *eof)
 		if (!comm && html[0] == '>') return html + 1;
 		if (comm && html + 2 <= eof && html[0] == '-' && html[1] == '-') {
 			html += 2;
-			while (html < eof && *html == '-') html++;
+			while (html < eof && (*html == '-' || *html == '!')) html++;
 			while (html < eof && WHITECHAR(*html)) html++;
 			if (html >= eof) return eof;
 			if (*html == '>') return html + 1;
@@ -2776,7 +2894,7 @@ do {					\
 
 	set_lt:
 
-	set_globals;	/* !!! FIXME: toto odstranit */
+	/*set_globals;*/
 
 	lt = html;
 	while (html < eof) {
@@ -2853,7 +2971,7 @@ do {					\
 			html += xl;
 			goto set_lt;
 		}
-		if (html + 2 <= eof && html[0] == '<' && (html[1] == '!' || html[1] == '?') && !(d_opt->plain & 1)) {
+		if (html + 2 <= eof && html[0] == '<' && (html[1] == '!' || html[1] == '?') && !(d_opt->plain & 1) && html_top.invisible != INVISIBLE_STYLE) {
 			/*if (putsp == 1) goto put_sp;
 			putsp = 0;*/
 			put_chrs(lt, (int)(html - lt));
@@ -2869,6 +2987,14 @@ do {					\
 		element:
 		html_format_changed = 1;
 		inv = *name == '/'; name += inv; namelen -= inv;
+		if (html_top.invisible == INVISIBLE_SCRIPT && !(inv && namelen == 6 && !casecmp(name, cast_uchar "SCRIPT", 6))) {
+			html++;
+			continue;
+		}
+		if (html_top.invisible == INVISIBLE_STYLE && !(inv && namelen == 5 && !casecmp(name, cast_uchar "STYLE", 5))) {
+			html++;
+			continue;
+		}
 		if (!inv && putsp == 1 && !html_top.invisible) goto put_sp;
 		put_chrs(lt, (int)(html - lt));
 		if (par_format.align != AL_NO && par_format.align != AL_NO_BREAKABLE) if (!inv && !putsp) {
@@ -2896,7 +3022,7 @@ do {					\
 				int display_none = 0;
 				int noskip = 0;
 	/* treat <br> literally in <pre> (fixes source code viewer on github) */
-				if ((par_format.align == AL_NO || par_format.align == AL_NO_BREAKABLE) && !strcasecmp(cast_const_char ei->name, "BR"))
+				if ((par_format.align == AL_NO || par_format.align == AL_NO_BREAKABLE) && !casestrcmp(cast_uchar ei->name, cast_uchar "BR"))
 					line_breax = 0;
 				ln_break(ei->linebreak);
 				if ((a = get_attr_val(attr, cast_uchar "id"))) {
@@ -2906,10 +3032,10 @@ do {					\
 				if ((a = get_attr_val(attr, cast_uchar "style"))) {
 					unsigned char *d, *s;
 
-					if (!strcasecmp(cast_const_char ei->name, "INPUT")) {
+					if (!casestrcmp(cast_uchar ei->name, cast_uchar "INPUT")) {
 						unsigned char *aa = get_attr_val(attr, cast_uchar "type");
 						if (aa) {
-							if (!strcasecmp(cast_const_char aa, "hidden"))
+							if (!casestrcmp(aa, cast_uchar "hidden"))
 								noskip = 1;
 							mem_free(aa);
 						}
@@ -2927,7 +3053,7 @@ do {					\
 					html_top.namelen = namelen;
 					html_top.options = attr;
 					html_top.linebreak = 0;
-					html_top.invisible = 1;
+					html_top.invisible = INVISIBLE;
 					goto set_lt;
 				}
 				if (!html_top.invisible) {
@@ -2948,19 +3074,20 @@ do {					\
 						goto set_lt;
 					}
 					if (ei->nopair == 2 || ei->nopair == 3) {
-						struct html_element *e;
+						struct html_element *e = NULL;	/* against warning */
+						struct list_head *le;
 						if (ei->nopair == 2) {
-							foreach(e, html_stack) {
+							foreach(struct html_element, e, le, html_stack) {
 								if (e->dontkill) break;
 								if (e->linebreak || !ei->linebreak) break;
 							}
-						} else foreach(e, html_stack) {
+						} else foreach(struct html_element, e, le, html_stack) {
 							if (e->linebreak && !ei->linebreak) break;
 							if (e->dontkill) break;
 							if (e->namelen == namelen && !casecmp(e->name, name, e->namelen)) break;
 						}
 						if (e->namelen == namelen && !casecmp(e->name, name, e->namelen)) {
-							while (e->prev != (void *)&html_stack) kill_html_stack_item(e->prev);
+							while (e->list_entry.prev != &html_stack) kill_html_stack_item(list_struct(e->list_entry.prev, struct html_element));
 							if (e->dontkill != 2) kill_html_stack_item(e);
 						}
 					}
@@ -2975,15 +3102,25 @@ do {					\
 					if (ei->func == html_a) html_a_special(attr, html, eof);
 					if (ei->func != html_br) was_br = 0;
 					if (a) par_format = pa;
+				} else {
+					if (!ei->nopair) {
+						html_stack_dup();
+						html_top.name = name;
+						html_top.namelen = namelen;
+						html_top.options = attr;
+						html_top.linebreak = 0;
+						html_top.invisible = INVISIBLE;
+					}
 				}
 			} else {
-				struct html_element *e, *ff;
+				struct html_element *e, *fx;
+				struct list_head *le, *lfx;
 				int lnb = 0;
 				int xxx = 0;
 				was_br = 0;
 				if (ei->nopair == 1 || ei->nopair == 3) break;
 				/*debug_stack();*/
-				foreach(e, html_stack) {
+				foreach(struct html_element, e, le, html_stack) {
 					if (e->linebreak && !ei->linebreak) xxx = 1;
 					if (e->namelen != namelen || casecmp(e->name, name, e->namelen)) {
 						if (e->dontkill) break;
@@ -2993,11 +3130,11 @@ do {					\
 						kill_html_stack_item(e);
 						break;
 					}
-					for (ff = e; ff != (void *)&html_stack; ff = ff->prev)
-						if (ff->linebreak > lnb) lnb = ff->linebreak;
-					format_.fontsize = e->next->attr.fontsize;
+					foreachbackfrom(struct html_element, fx, lfx, html_stack, le)
+						if (fx->linebreak > lnb) lnb = fx->linebreak;
+					format_.fontsize = list_struct(e->list_entry.next, struct html_element)->attr.fontsize;
 					ln_break(lnb);
-					while (e->prev != (void *)&html_stack) kill_html_stack_item(e->prev);
+					while (e->list_entry.prev != &html_stack) kill_html_stack_item(list_struct(e->list_entry.prev, struct html_element));
 					kill_html_stack_item(e);
 					break;
 				}
@@ -3065,7 +3202,7 @@ int get_image_map(unsigned char *head, unsigned char *s, unsigned char *eof, uns
 	if (namelen != 3 || casecmp(name, cast_uchar "MAP", 3)) goto se;
 	if (tag && *tag) {
 		if (!(al = get_attr_val(attr, cast_uchar "name"))) goto se;
-		if (strcasecmp(cast_const_char al, cast_const_char tag)) {
+		if (casestrcmp(al, tag)) {
 			mem_free(al);
 			goto se;
 		}
@@ -3130,14 +3267,10 @@ int get_image_map(unsigned char *head, unsigned char *s, unsigned char *eof, uns
 	if (!(target = get_target(attr)) && !(target = stracpy(target_base)))
 		target = stracpy(cast_uchar "");
 	ld = mem_calloc(sizeof(struct link_def));
-	if (href) if (!(ld->link = join_urls(href_base, href))) {
+	if (href) {
+		ld->link = join_urls(href_base, href);
 		mem_free(href);
-		mem_free(target);
-		mem_free(ld);
-		if (label) mem_free(label);
-		goto se2;
 	}
-	if (href) mem_free(href);
 	ld->target = target;
 
 	add_to_ml(ml, ld, ld->target, NULL);
@@ -3217,8 +3350,10 @@ void scan_http_equiv(unsigned char *s, unsigned char *eof, unsigned char **head,
 	if (parse_element(s, eof, &name, &namelen, &attr, &s)) goto sp;
 	ps:
 	if (namelen == 6 && !casecmp(name, cast_uchar "SCRIPT", 6)) {
-		s = skip_element(s, eof, cast_uchar "SCRIPT", 0);
-		goto se;
+		if (should_skip_script(attr)) {
+			s = skip_element(s, eof, cast_uchar "SCRIPT", 0);
+			goto se;
+		}
 	}
 	if (namelen == 4 && !casecmp(name, cast_uchar "BODY", 4)) {
 		if (background) *background = get_attr_val(attr, cast_uchar "background"), background = NULL;
@@ -3257,6 +3392,7 @@ void scan_http_equiv(unsigned char *s, unsigned char *eof, unsigned char **head,
 	if ((he = get_attr_val(attr, cast_uchar "charset"))) {
 		add_to_str(head, hdl, cast_uchar "Charset: ");
 		add_to_str(head, hdl, he);
+		add_to_str(head, hdl, cast_uchar "\r\n");
 		mem_free(he);
 	}
 	if (!(he = get_attr_val(attr, cast_uchar "http-equiv"))) goto se;
