@@ -6,8 +6,6 @@
 
 #include "cfg.h"
 
-#define format format_
-
 #ifdef G
 
 #include "links.h"
@@ -17,7 +15,7 @@ static int g_nobreak;
 static int get_real_font_size(int size)
 {
 	int fs=d_opt->font_size;
-	
+
 	if (size < 1) size = 1;
 	if (size > 7) size = 7;
 	switch (size) {
@@ -37,9 +35,7 @@ struct background *get_background(unsigned char *bg, unsigned char *bgcolor)
 	struct background *b;
 	struct rgb r;
 	b = mem_alloc(sizeof(struct background));
-	/* !!! FIXME: background image */
 	{
-		b->img = 0;
 		if (bgcolor && !decode_color(bgcolor, &r)) {
 			b->u.sRGB=(r.r << 16) + (r.g << 8) + r.b;
 		} else {
@@ -61,30 +57,25 @@ static int gray (int r, int g, int b)
 /* 1=too near 0=not too near */
 static int too_near(int r1, int g1, int b1, int r2, int g2, int b2)
 {
-	int gray1,gray2;
+	int diff = abs(r1-r2) * 3 + abs(g1-g2) * 6 + abs(b1-b2);
 
-	gray1=gray(r1, g1, b1);
-	gray2=gray(r2, g2, b2);
-	if (gray1<=gray2)
-		return gray2-gray1<=400;
-	else
-		return gray1-gray2<=400;
+	return diff < 0x17f;
 }
 
 /* Fixes foreground based on background */
 static void separate_fg_bg(int *fgr, int *fgg, int *fgb
 	, int bgr, int bgg, int bgb)
 {
-	if (too_near(*fgr, *fgg, *fgb, bgr, bgg, bgb)){
-		*fgr=255-bgr;
-		*fgg=255-bgg;
-		*fgb=255-bgb;
-	}else return;
-	if (too_near(*fgr, *fgg, *fgb, bgr, bgg, bgb)){
-		if (gray(bgr, bgg, bgb)<=1275)
-			*fgr=*fgg=*fgb=255;
+	if (too_near(*fgr, *fgg, *fgb, bgr, bgg, bgb)) {
+		*fgr = 255 - bgr;
+		*fgg = 255 - bgg;
+		*fgb = 255 - bgb;
+	} else return;
+	if (too_near(*fgr, *fgg, *fgb, bgr, bgg, bgb)) {
+		if (gray(bgr, bgg, bgb) <= 1275)
+			*fgr = *fgg = *fgb = 255;
 		else
-			*fgr=*fgg=*fgb=0;
+			*fgr = *fgg = *fgb = 0;
 	}
 }
 
@@ -92,13 +83,12 @@ static unsigned char *make_html_font_name(int attr)
 {
 	unsigned char *str;
 	int len;
-	
+
 	str=init_str();len=0;
-	add_to_str(&str, &len, G_HTML_DEFAULT_FAMILY);
-	add_to_str(&str, &len, attr & AT_BOLD ? "-bold" : "-medium");
-	add_to_str(&str, &len, attr & AT_ITALIC ? "-italic-serif" :
-			"-roman-serif");
-	add_to_str(&str, &len, attr & AT_FIXED ? "-mono" : "-vari");
+	add_to_str(&str, &len, cast_uchar G_HTML_DEFAULT_FAMILY);
+	add_to_str(&str, &len, attr & AT_BOLD ? cast_uchar "-bold" : cast_uchar "-medium");
+	add_to_str(&str, &len, attr & AT_ITALIC ? cast_uchar "-italic-serif" : cast_uchar "-roman-serif");
+	add_to_str(&str, &len, attr & AT_FIXED ? cast_uchar "-mono" : cast_uchar "-vari");
 	return str;
 }
 
@@ -114,7 +104,7 @@ static struct style *get_style_by_ta(struct text_attrib *ta)
 	fg_b=ta->fg.b;
 	separate_fg_bg(&fg_r,&fg_g,&fg_b,ta->bg.r,ta->bg.g,ta->bg.b);
 	stl = g_get_style((fg_r << 16) + (fg_g << 8) + fg_b, (ta->bg.r << 16) +
-			(ta->bg.g << 8) + ta->bg.b, fs, 
+			(ta->bg.g << 8) + ta->bg.b, fs,
 			fontname=make_html_font_name(ta->attr),
 			ta->attr & AT_UNDERLINE ? FF_UNDERLINE : 0);
 	mem_free(fontname);
@@ -123,23 +113,6 @@ static struct style *get_style_by_ta(struct text_attrib *ta)
 
 #define rm(x) ((x).width - (x).rightmargin * G_HTML_MARGIN > 0 ? (x).width - (x).rightmargin * G_HTML_MARGIN : 0)
 
-#ifndef SPAD
-static inline int pw2(int a)
-{
-	int x = 1;
-	while (x < a && x) {
-		if ((unsigned)x > MAXINT / 2) overalloc();
-		x <<= 1;
-	}
-	return x;
-}
-#else
-static inline int pw2(int a)
-{
-	return a;
-}
-#endif
-
 void flush_pending_line_to_obj(struct g_part *p, int minheight)
 {
 	int i, pp, pos, w, lbl;
@@ -149,11 +122,12 @@ void flush_pending_line_to_obj(struct g_part *p, int minheight)
 		return;
 	}
 	for (i = l->n_entries - 1; i >= 0; i--) {
-		struct g_object_text *go = (struct g_object_text *)l->entries[i];
+		struct g_object *go = l->entries[i];
 		if (go->draw == g_text_draw) {
-			int l;
-			while ((l = strlen(go->text)) && go->text[l - 1] == ' ') go->text[l - 1] = 0, go->xw -= g_char_width(go->style, ' ');
-			if (go->xw < 0) internal("xw(%d) < 0", go->xw);
+			struct g_object_text *got = get_struct(go, struct g_object_text, goti.go);
+			int l = (int)strlen(cast_const_char got->text);
+			while (l && got->text[l - 1] == ' ') got->text[--l] = 0, got->goti.go.xw -= g_char_width(got->style, ' ');
+			if (got->goti.go.xw < 0) internal("xw(%d) < 0", got->goti.go.xw);
 		}
 		if (!go->xw) continue;
 		break;
@@ -165,44 +139,46 @@ void flush_pending_line_to_obj(struct g_part *p, int minheight)
 	for (i = 0; i < l->n_entries; i++) {
 		int yy = l->entries[i]->y;
 		if (yy >= G_OBJ_ALIGN_SPECIAL) yy = 0;
-		pp += l->entries[i]->xw;
-		if (l->entries[i]->xw && l->entries[i]->yw + yy > w) w = l->entries[i]->yw + yy;
+		pp = safe_add(pp, l->entries[i]->xw);
+		if (l->entries[i]->xw && safe_add(l->entries[i]->yw, yy) > w) w = safe_add(l->entries[i]->yw, yy);
 		if (yy < lbl) lbl = yy;
 	}
 	if (lbl < 0) {
 		for (i = 0; i < l->n_entries; i++) {
-			if (l->entries[i]->y < G_OBJ_ALIGN_SPECIAL) l->entries[i]->y -= lbl;
+			if (l->entries[i]->y < G_OBJ_ALIGN_SPECIAL) l->entries[i]->y = safe_add(l->entries[i]->y, -lbl);
 		}
 		goto scan_again;
 	}
-	if (par_format.align == AL_CENTER) pos = (rm(par_format) + par_format.leftmargin * G_HTML_MARGIN - pp) / 2;
+	if (par_format.align == AL_CENTER) pos = (safe_add(rm(par_format), par_format.leftmargin * G_HTML_MARGIN) - pp) / 2;
 	else if (par_format.align == AL_RIGHT) pos = rm(par_format) - pp;
 	else pos = par_format.leftmargin * G_HTML_MARGIN;
 	if (pos < par_format.leftmargin * G_HTML_MARGIN) pos = par_format.leftmargin * G_HTML_MARGIN;
 	pp = pos;
 	for (i = 0; i < l->n_entries; i++) {
 		l->entries[i]->x = pp;
-		pp += l->entries[i]->xw;
+		pp = safe_add(pp, l->entries[i]->xw);
 		if (l->entries[i]->y < G_OBJ_ALIGN_SPECIAL) {
 			l->entries[i]->y = w - l->entries[i]->yw - l->entries[i]->y;
 		} else if (l->entries[i]->y == G_OBJ_ALIGN_TOP) {
 			l->entries[i]->y = 0;
 		} else if (l->entries[i]->y == G_OBJ_ALIGN_MIDDLE) {
-			l->entries[i]->y = (w - l->entries[i]->yw + 1) / 2;
+			l->entries[i]->y = (w - safe_add(l->entries[i]->yw, 1)) / 2;
 		}
 	}
-	l->x = 0;
-	l->xw = par_format.width;
-	l->yw = w;
-	l->y = p->cy;
-	if (l->xw > p->root->xw) p->root->xw = l->xw;
-	p->root->yw = p->cy += w;
-	p->root->n_lines++;
-	if ((unsigned)p->root->n_lines > (MAXINT - sizeof(struct g_object_area)) / 2 / sizeof(struct g_object_text *) + 1) overalloc();
-	a = mem_realloc(p->root, sizeof(struct g_object_area) + sizeof(struct g_object_text *) * pw2(p->root->n_lines + 1));
-		/* note: +1 is for extend_area */
-	p->root = a;
-	p->root->lines[p->root->n_lines - 1] = l;
+	l->go.x = 0;
+	l->go.xw = par_format.width;
+	l->go.yw = w;
+	l->go.y = p->cy;
+	a = p->root;
+	if (l->go.xw > a->go.xw) a->go.xw = l->go.xw;
+	p->cy = safe_add(p->cy, w);
+	a->go.yw = p->cy;
+	if (!(a->n_lines & (a->n_lines + 1))) {
+		if ((unsigned)a->n_lines > ((MAXINT - sizeof(struct g_object_area)) / sizeof(struct g_object_text *) - 1) / 2) overalloc();
+		a = mem_realloc(a, sizeof(struct g_object_area) + sizeof(struct g_object_text *) * (a->n_lines * 2 + 1));
+		p->root = a;
+	}
+	a->lines[a->n_lines++] = l;
 	p->line = NULL;
 	p->w.pos = 0;
 	p->w.last_wrap = NULL;
@@ -218,14 +194,14 @@ void add_object_to_line(struct g_part *pp, struct g_object_line **lp, struct g_o
 	}
 	if (!*lp) {
 		l = mem_calloc(sizeof(struct g_object_line) + sizeof(struct g_object_text *));
-		l->mouse_event = g_line_mouse;
-		l->draw = g_line_draw;
-		l->destruct = g_line_destruct;
-		l->get_list = g_line_get_list;
-		/*l->x = 0;
-		l->y = 0;
-		l->xw = 0;
-		l->yw = 0;*/
+		l->go.mouse_event = g_line_mouse;
+		l->go.draw = g_line_draw;
+		l->go.destruct = g_line_destruct;
+		l->go.get_list = g_line_get_list;
+		/*l->go.x = 0;
+		l->go.y = 0;
+		l->go.xw = 0;
+		l->go.yw = 0;*/
 		l->bg = pp->root->bg;
 		if (!go) {
 			*lp = l;
@@ -243,9 +219,9 @@ void add_object_to_line(struct g_part *pp, struct g_object_line **lp, struct g_o
 	*lp = l;
 	if (pp->cx == -1) pp->cx = par_format.leftmargin * G_HTML_MARGIN;
 	if (go->xw) {
-		pp->cx += pp->cx_w;
+		pp->cx = safe_add(pp->cx, pp->cx_w);
 		pp->cx_w = 0;
-		pp->cx += go->xw;
+		pp->cx = safe_add(pp->cx, go->xw);
 	}
 }
 
@@ -253,7 +229,7 @@ void flush_pending_text_to_line(struct g_part *p)
 {
 	struct g_object_text *t = p->text;
 	if (!t) return;
-	add_object_to_line(p, &p->line, (struct g_object *)t);
+	add_object_to_line(p, &p->line, &t->goti.go);
 	p->text = NULL;
 }
 
@@ -263,86 +239,98 @@ static void split_line_object(struct g_part *p, struct g_object_text *text, unsi
 	struct g_object_line *l2;
 	int n;
 	if (!ptr) {
-		if (p->line && p->line->n_entries && (struct g_object *)text == p->line->entries[p->line->n_entries - 1]) {
+		if (p->line && p->line->n_entries && &text->goti.go == p->line->entries[p->line->n_entries - 1]) {
 			flush_pending_line_to_obj(p, 0);
 			goto wwww;
 		}
 		t2 = NULL;
 		goto nt2;
 	}
-	t2 = mem_calloc(sizeof(struct g_object_text) + strlen(ptr));
-	t2->mouse_event = g_text_mouse;
-	t2->draw = g_text_draw;
-	t2->destruct = g_text_destruct;
-	t2->get_list = NULL;
+	if (par_format.align == AL_NO_BREAKABLE && text == p->text && strspn(cast_const_char ptr, cast_const_char " ") == strlen(cast_const_char ptr)) {
+		return;
+	}
+#ifdef DEBUG
+	if (ptr < text->text || ptr >= text->text + strlen(cast_const_char text->text))
+		internal("split_line_object: split point (%p) pointing out of object (%p,%lx)", ptr, text->text, (unsigned long)strlen(cast_const_char text->text));
+#endif
+	t2 = mem_calloc(sizeof(struct g_object_text) + strlen(cast_const_char ptr));
+	t2->goti.go.mouse_event = g_text_mouse;
+	t2->goti.go.draw = g_text_draw;
+	t2->goti.go.destruct = g_text_destruct;
+	t2->goti.go.get_list = NULL;
 	if (*ptr == ' ') {
-		strcpy(t2->text, ptr + 1);
+		strcpy(cast_char t2->text, cast_const_char(ptr + 1));
 		*ptr = 0;
 		/*debug("split: (%s)(%s)", text->text, ptr + 1);*/
 	} else {
-		strcpy(t2->text, ptr);
+		strcpy(cast_char t2->text, cast_const_char ptr);
 		ptr[0] = '-';
 		ptr[1] = 0;
 	}
-	t2->y = text->y;
+	t2->goti.go.y = text->goti.go.y;
 	t2->style = g_clone_style(text->style);
-	t2->bg = text->bg;
-	t2->link_num = text->link_num;
-	t2->link_order = text->link_order;
-	text->xw = g_text_width(text->style, text->text);
+	t2->goti.link_num = text->goti.link_num;
+	t2->goti.link_order = text->goti.link_order;
+	text->goti.go.xw = g_text_width(text->style, text->text);
 	nt2:
-	if (p->line) for (n = 0; n < p->line->n_entries; n++) if (p->line->entries[n] == (struct g_object *)text) goto found;
+	if (p->line) for (n = 0; n < p->line->n_entries; n++) if (p->line->entries[n] == &text->goti.go) goto found;
 	if (text != p->text) {
 		internal("split_line_object: bad wrap");
-		t2->destruct(t2);
-		mem_free(t2);
 		return;
 	}
 	if (0) {
 		int nn;
+		int qw;
 		found:
-		n += !ptr;
+		n = safe_add(n, !ptr);
 		l2 = mem_calloc(sizeof(struct g_object_line) + (p->line->n_entries - n) * sizeof(struct g_object_text *));
-		l2->mouse_event = g_line_mouse;
-		l2->draw = g_line_draw;
-		l2->destruct = g_line_destruct;
-		l2->get_list = g_line_get_list;
+		l2->go.mouse_event = g_line_mouse;
+		l2->go.draw = g_line_draw;
+		l2->go.destruct = g_line_destruct;
+		l2->go.get_list = g_line_get_list;
 		l2->bg = p->root->bg;
 		l2->n_entries = p->line->n_entries - n;
-		l2->entries[0] = (struct g_object *)t2;
-		memcpy(&l2->entries[!!ptr], p->line->entries + n + !!ptr, (l2->n_entries - !!ptr) * sizeof(struct g_object_text *));
-		p->line->n_entries = n + !!ptr;
+		l2->entries[0] = &t2->goti.go;
+		memcpy(&l2->entries[!!ptr], p->line->entries + safe_add(n, !!ptr), (l2->n_entries - !!ptr) * sizeof(struct g_object_text *));
+		p->line->n_entries = safe_add(n, !!ptr);
 		flush_pending_line_to_obj(p, 0);
 		p->line = l2;
 		if (ptr) {
-			t2->xw = g_text_width(t2->style, t2->text);
-			t2->yw = text->yw;
+			t2->goti.go.xw = g_text_width(t2->style, t2->text);
+			t2->goti.go.yw = text->goti.go.yw;
 			p->w.pos = 0;
 		}
 		for (nn = 0; nn < l2->n_entries; nn++) {
-			p->w.pos += l2->entries[nn]->xw;	/* !!! FIXME: nastav last_wrap */
+			p->w.pos = safe_add(p->w.pos, l2->entries[nn]->xw);	/* !!! FIXME: nastav last_wrap */
 			/*debug("a1: %d (%s)", l2->entries[nn]->xw, tt->text);*/
 		}
 		wwww:
-		if (p->text) p->w.pos += g_text_width(p->text->style, p->text->text);
+		qw = g_text_width(p->text->style, p->text->text);
+		if (p->text) p->w.pos = safe_add(p->w.pos, qw);
 		/*debug("%d", p->w.pos);*/
 	} else {
 		flush_pending_text_to_line(p);
 		flush_pending_line_to_obj(p, 0);
 		p->line = NULL;
-		t2->xw = g_text_width(t2->style, t2->text);
-		t2->yw = text->yw;
+		t2->goti.go.xw = g_text_width(t2->style, t2->text);
+		t2->goti.go.yw = text->goti.go.yw;
 		p->text = t2;
 		p->pending_text_len = -1;
-		p->w.pos = t2->xw;
+		p->w.pos = t2->goti.go.xw;
 		p->cx_w = g_char_width(t2->style, ' ');
 	}
 	p->w.last_wrap = NULL;
 	p->w.last_wrap_obj = NULL;
 	t2 = p->text;
-	if (t2 && *t2->text && t2->text[strlen(t2->text) - 1] == ' ') {
-		p->w.last_wrap = &t2->text[strlen(t2->text) - 1];
-		p->w.last_wrap_obj = t2;
+	if (t2) {
+		int sl = (int)strlen(cast_const_char t2->text);
+		if (sl >= 1 && t2->text[sl - 1] == ' ') {
+			p->w.last_wrap = &t2->text[sl - 1];
+			p->w.last_wrap_obj = &t2->goti.go;
+		} else if (sl >= 2 && t2->text[sl - 2] == 0xc2 && t2->text[sl - 1] == 0xad) {
+			p->w.last_wrap = &t2->text[sl - 2];
+			p->w.last_wrap_obj = &t2->goti.go;
+		}
 	}
 }
 
@@ -351,16 +339,11 @@ void add_object(struct g_part *p, struct g_object *o)
 	g_nobreak = 0;
 	flush_pending_text_to_line(p);
 	p->w.width = rm(par_format) - par_format.leftmargin * G_HTML_MARGIN;
-	if (p->w.pos + o->xw > p->w.width) flush_pending_line_to_obj(p, 0);
+	if (safe_add(p->w.pos, o->xw) > p->w.width) flush_pending_line_to_obj(p, 0);
 	add_object_to_line(p, &p->line, o);
 	p->w.last_wrap = NULL;
 	p->w.last_wrap_obj = o;
-	p->w.pos += o->xw;
-	/*
-	if (p->w.pos > p->w.width && p->w.last_wrap_obj) {
-		split_line_object(p, p->w.last_wrap_obj, p->w.last_wrap);
-	}
-	*/
+	p->w.pos = safe_add(p->w.pos, o->xw);
 }
 
 static void g_line_break(void *p_)
@@ -371,10 +354,19 @@ static void g_line_break(void *p_)
 		return;
 	}
 	flush_pending_text_to_line(p);
-	if (!p->line) {
+	if (!p->line || par_format.align == AL_NO || par_format.align == AL_NO_BREAKABLE) {
 		add_object_to_line(p, &p->line, NULL);
-		flush_pending_line_to_obj(p, get_real_font_size(format.fontsize));
-	} else /*if (p->w.pos || par_format.align == AL_NO)*/ {
+		empty_line:
+		flush_pending_line_to_obj(p, get_real_font_size(format_.fontsize));
+	} else {
+		int i;
+		for (i = 0; i < p->line->n_entries; i++) {
+			struct g_object *go = p->line->entries[i];
+			if (go->destruct != g_tag_destruct)
+				goto flush;
+		}
+		goto empty_line;
+		flush:
 		flush_pending_line_to_obj(p, 0);
 	}
 	if (p->cx > p->xmax) p->xmax = p->cx;
@@ -391,15 +383,16 @@ static void g_html_form_control(struct g_part *p, struct form_control *fc)
 		add_to_list(p->uf, fc);
 		return;
 	}
-	fc->g_ctrl_num = g_ctrl_num++;
+	fc->g_ctrl_num = g_ctrl_num;
+	g_ctrl_num = safe_add(g_ctrl_num, 1);
 	if (fc->type == FC_TEXT || fc->type == FC_PASSWORD || fc->type == FC_TEXTAREA) {
-		unsigned char *dv = convert_string(convert_table, fc->default_value, strlen(fc->default_value), d_opt);
+		unsigned char *dv = convert_string(convert_table, fc->default_value, (int)strlen(cast_const_char fc->default_value), d_opt);
 		if (dv) {
 			mem_free(fc->default_value);
 			fc->default_value = dv;
 		}
 		/*
-		for (i = 0; i < fc->nvalues; i++) if ((dv = convert_string(convert_table, fc->values[i], strlen(fc->values[i]), d_opt))) {
+		for (i = 0; i < fc->nvalues; i++) if ((dv = convert_string(convert_table, fc->values[i], strlen(cast_const_char fc->values[i]), d_opt))) {
 			mem_free(fc->values[i]);
 			fc->values[i] = dv;
 		}
@@ -408,7 +401,7 @@ static void g_html_form_control(struct g_part *p, struct form_control *fc)
 	if (fc->type == FC_TEXTAREA) {
 		unsigned char *p;
 		for (p = fc->default_value; p[0]; p++) if (p[0] == '\r') {
-			if (p[1] == '\n') memmove(p, p + 1, strlen(p)), p--;
+			if (p[1] == '\n') memmove(p, p + 1, strlen(cast_const_char p)), p--;
 			else p[0] = '\n';
 		}
 	}
@@ -481,15 +474,16 @@ static void do_image(struct g_part *p, struct image_description *im)
 	putchars_link_ptr = NULL;
 	if (!link) im->link_num = -1;
 	else {
-		im->link_num = link - p->data->links;
-		im->link_order = link->obj_order++;
+		im->link_num = (int)(link - p->data->links);
+		im->link_order = link->obj_order;
+		link->obj_order = safe_add(link->obj_order, 1);
 		if (link->img_alt) mem_free(link->img_alt);
 		link->img_alt = stracpy(im->alt);
 	}
 	io = insert_image(p, im);
 	if (!io) goto ab;
-	io->ismap = im->ismap;
-	add_object(p, (struct g_object *)io);
+	io->goti.ismap = im->ismap;
+	add_object(p, &io->goti.go);
 	if (im->usemap && p->data) {
 		unsigned char *tag = extract_position(im->usemap);
 		struct additional_file *af = request_additional_file(current_f_data, im->usemap);
@@ -503,14 +497,20 @@ static void do_image(struct g_part *p, struct image_description *im)
 			struct image_map *map;
 			if (get_file(af->rq, &start, &end)) goto ft;
 			if (start == end) goto ft;
-			if (get_image_map(ce->head, start, end, tag, &menu, &ml, format.href_base, format.target_base, 0, 0, 0, 1)) goto ft;
+			if (get_image_map(ce->head, start, end, tag, &menu, &ml, format_.href_base, format_.target_base, 0, 0, 0, 1)) goto ft;
 			map = mem_alloc(sizeof(struct image_map));
 			map->n_areas = 0;
 			for (i = 0; menu[i].text; i++) {
 				struct link_def *ld = menu[i].data;
 				struct map_area *a;
 				struct link *link;
-				int shape = !ld->shape || !*ld->shape ? SHAPE_RECT : !strcasecmp(ld->shape, "default") ? SHAPE_DEFAULT : !strcasecmp(ld->shape, "rect") ? SHAPE_RECT : !strcasecmp(ld->shape, "circle") ? SHAPE_CIRCLE : !strcasecmp(ld->shape, "poly") || !strcasecmp(ld->shape, "polygon") ? SHAPE_POLY : -1;
+				int shape =
+		!ld->shape || !*ld->shape ? SHAPE_RECT :
+		!casestrcmp(ld->shape, cast_uchar "default") ? SHAPE_DEFAULT :
+		!casestrcmp(ld->shape, cast_uchar "rect") ? SHAPE_RECT :
+		!casestrcmp(ld->shape, cast_uchar "circle") ? SHAPE_CIRCLE :
+		!casestrcmp(ld->shape, cast_uchar "poly") ||
+		!casestrcmp(ld->shape, cast_uchar "polygon") ? SHAPE_POLY : -1;
 				if (shape == -1) continue;
 				if ((unsigned)map->n_areas > (MAXINT - sizeof(struct image_map)) / sizeof(struct map_area) - 1) overalloc();
 				map = mem_realloc(map, sizeof(struct image_map) + (map->n_areas + 1) * sizeof(struct map_area));
@@ -531,7 +531,7 @@ static void do_image(struct g_part *p, struct image_description *im)
 						while (*p >= '0' && *p <= '9') p++;
 					}
 					if (*p == '%' && num < 1000) {
-						int m = io->xw < io->yw ? io->xw : io->yw;
+						int m = io->goti.go.xw < io->goti.go.yw ? io->goti.go.xw : io->goti.go.yw;
 						num = num * m / 100;
 						p++;
 					} else num = num * d_opt->image_scale / 100;
@@ -561,11 +561,11 @@ static void do_image(struct g_part *p, struct image_description *im)
 						link->js_event->move_code = stracpy(ld->onmousemove);
 					}
 #endif
-					a->link_num = link - p->data->links;
+					a->link_num = (int)(link - p->data->links);
 				}
 				if (last_link) mem_free(last_link), last_link = NULL;
 			}
-			io->map = map;
+			io->goti.map = map;
 			freeml(ml);
 			ft:;
 		}
@@ -579,20 +579,19 @@ static void g_hr(struct g_part *gp, struct hr_param *hr)
 	unsigned char bgstr[8];
 	struct g_object_line *o;
 	o = mem_calloc(sizeof(struct g_object_line));
-	o->mouse_event = g_line_mouse;
-	o->draw = g_line_draw;
-	o->destruct = g_line_bg_destruct;
-	o->get_list = g_line_get_list;
-	/*o->x = 0;
-	o->y = 0;*/
-	o->xw = hr->width;
-	o->yw = hr->size;
-	table_bg(&format, bgstr);
+	o->go.mouse_event = g_line_mouse;
+	o->go.draw = g_line_draw;
+	o->go.destruct = g_line_bg_destruct;
+	o->go.get_list = g_line_get_list;
+	/*o->go.x = 0;
+	o->go.y = 0;*/
+	o->go.xw = hr->width;
+	o->go.yw = hr->size;
+	table_bg(&format_, bgstr);
 	o->bg = get_background(NULL, bgstr);
 	o->n_entries = 0;
 	flush_pending_text_to_line(gp);
-	/*flush_pending_line_to_obj(gp, get_real_font_size(format.fontsize));*/
-	add_object_to_line(gp, &gp->line, (struct g_object *)o);
+	add_object_to_line(gp, &gp->line, &o->go);
 	line_breax = 0;
 	gp->cx = -1;
 	gp->cx_w = 0;
@@ -617,13 +616,13 @@ static void *g_html_special(void *p_, int c, ...)
 			t = va_arg(l, unsigned char *);
 			va_end(l);
 			/* not needed to convert %AB here because html_tag will be called anyway */
-			tag = mem_calloc(sizeof(struct g_object_tag) + strlen(t) + 1);
-			tag->mouse_event = g_dummy_mouse;
-			tag->draw = g_dummy_draw;
-			tag->destruct = g_tag_destruct;
-			strcpy(tag->name, t);
+			tag = mem_calloc(sizeof(struct g_object_tag) + strlen(cast_const_char t));
+			tag->go.mouse_event = g_dummy_mouse;
+			tag->go.draw = g_dummy_draw;
+			tag->go.destruct = g_tag_destruct;
+			strcpy(cast_char tag->name, cast_const_char t);
 			flush_pending_text_to_line(p);
-			add_object_to_line(p, &p->line, (struct g_object *)tag);
+			add_object_to_line(p, &p->line, &tag->go);
 			break;
 		case SP_CONTROL:
 			fc = va_arg(l, struct form_control *);
@@ -680,7 +679,8 @@ static void *g_html_special(void *p_, int c, ...)
 	return NULL;
 }
 
-static unsigned char to_je_ale_prasarna[] = "";
+static const unsigned char to_je_ale_prasarna_[] = "";
+#define to_je_ale_prasarna	(cast_uchar to_je_ale_prasarna_)
 
 static unsigned char *cached_font_face = to_je_ale_prasarna;
 static struct text_attrib_beginning ta_cache = { -1, { 0, 0, 0, 0 }, { 0, 0, 0, 0 }, 0, 0 };
@@ -688,55 +688,75 @@ static struct text_attrib_beginning ta_cache = { -1, { 0, 0, 0, 0 }, { 0, 0, 0,
 static void g_put_chars(void *p_, unsigned char *s, int l)
 {
 	struct g_part *p = p_;
-	struct g_object_text *t = NULL;
+	struct g_object_text *t;
 	struct link *link;
 	int ptl;
+	unsigned char *sh;
+	int qw;
+
+	if (l < 0) overalloc();
+
+	while (l > 2 && (sh = memchr(s + 1, 0xad, l - 2)) && sh[-1] == 0xc2) {
+		sh++;
+		g_put_chars(p_, s, (int)(sh - s));
+		l -= (int)(sh - s);
+		s = sh;
+	}
+	while (par_format.align != AL_NO && l >= 2 && (sh = memchr(s, ' ', l - 1))) {
+		sh++;
+		g_put_chars(p_, s, (int)(sh - s));
+		l -= (int)(sh - s);
+		s = sh;
+	}
+
+	/*fprintf(stderr, "%d: '%.*s'\n", l, l, s);*/
+
+	t = NULL;
+
 	if (putchars_link_ptr) {
 		link = NULL;
 		goto check_link;
 	}
-	while (par_format.align != AL_NO && p->cx == -1 && l && *s == ' ') s++, l--;
+	while (par_format.align != AL_NO && par_format.align != AL_NO_BREAKABLE && p->cx == -1 && l && *s == ' ') s++, l--;
 	if (!l) return;
 	g_nobreak = 0;
 	if (p->cx < par_format.leftmargin * G_HTML_MARGIN) p->cx = par_format.leftmargin * G_HTML_MARGIN;
 	if (html_format_changed) {
-		if (memcmp(&ta_cache, &format, sizeof(struct text_attrib_beginning)) || xstrcmp(cached_font_face, format.fontface) || cached_font_face == to_je_ale_prasarna ||
-	xstrcmp(format.link, last_link) || xstrcmp(format.target, last_target) ||
-	    xstrcmp(format.image, last_image) || format.form != last_form
-	    || ((format.js_event || last_js_event) && compare_js_event_spec(format.js_event, last_js_event))	) {
+		if (memcmp(&ta_cache, &format_, sizeof(struct text_attrib_beginning)) || xstrcmp(cached_font_face, format_.fontface) || cached_font_face == to_je_ale_prasarna ||
+	xstrcmp(format_.link, last_link) || xstrcmp(format_.target, last_target) ||
+	    xstrcmp(format_.image, last_image) || format_.form != last_form
+	    || ((format_.js_event || last_js_event) && compare_js_event_spec(format_.js_event, last_js_event))	) {
 			/*if (!html_format_changed) internal("html_format_changed not set");*/
 			flush_pending_text_to_line(p);
-			if (xstrcmp(cached_font_face, format.fontface) || cached_font_face == to_je_ale_prasarna) {
+			if (xstrcmp(cached_font_face, format_.fontface) || cached_font_face == to_je_ale_prasarna) {
 				if (cached_font_face && cached_font_face != to_je_ale_prasarna) mem_free(cached_font_face);
-				cached_font_face = stracpy(format.fontface);
+				cached_font_face = stracpy(format_.fontface);
 			}
-			memcpy(&ta_cache, &format, sizeof(struct text_attrib_beginning));
+			memcpy(&ta_cache, &format_, sizeof(struct text_attrib_beginning));
 			if (p->current_style) g_free_style(p->current_style);
-			p->current_style = get_style_by_ta(&format);
+			p->current_style = get_style_by_ta(&format_);
 		}
 		html_format_changed = 0;
 	}
-	/*if (p->cx <= par_format.leftmargin * G_HTML_MARGIN && *s == ' ' && par_format.align != AL_NO) s++, l--;*/
+	/*if (p->cx <= par_format.leftmargin * G_HTML_MARGIN && *s == ' ' && par_format.align != AL_NO && par_format.align != AL_NO_BREAKABLE) s++, l--;*/
 	if (!p->text) {
 		link = NULL;
 		t = mem_calloc(sizeof(struct g_object_text) + ALLOC_GR);
-		t->mouse_event = g_text_mouse;
-		t->draw = g_text_draw;
-		t->destruct = g_text_destruct;
-		t->get_list = NULL;
-		/*t->style = get_style_by_ta(format);*/
+		t->goti.go.mouse_event = g_text_mouse;
+		t->goti.go.draw = g_text_draw;
+		t->goti.go.destruct = g_text_destruct;
+		t->goti.go.get_list = NULL;
 		t->style = g_clone_style(p->current_style);
-		t->bg = NULL; /* FIXME!!! */
-		t->yw = t->style->height;
-		/*t->xw = 0;
-		t->y = 0;*/
-		if (format.baseline) {
-			if (format.baseline < 0) t->y = -(t->style->height / 3);
-			if (format.baseline > 0) t->y = get_real_font_size(format.baseline) - (t->style->height / 2);
+		t->goti.go.yw = t->style->height;
+		/*t->goti.go.xw = 0;
+		t->goti.go.y = 0;*/
+		if (format_.baseline) {
+			if (format_.baseline < 0) t->goti.go.y = -(t->style->height / 3);
+			if (format_.baseline > 0) t->goti.go.y = get_real_font_size(format_.baseline) - (t->style->height / 2);
 		}
 		check_link:
-		if (last_link || last_image || last_form || format.link || format.image || format.form 
-		|| format.js_event || last_js_event
+		if (last_link || last_image || last_form || format_.link || format_.image || format_.form
+		|| format_.js_event || last_js_event
 		) goto process_link;
 		back_link:
 		if (putchars_link_ptr) {
@@ -744,10 +764,10 @@ static void g_put_chars(void *p_, unsigned char *s, int l)
 			return;
 		}
 
-		if (!link) t->link_num = -1;
+		if (!link) t->goti.link_num = -1;
 		else {
-			t->link_num = link - p->data->links;
-			t->link_order = link->obj_order++;
+			t->goti.link_num = (int)(link - p->data->links);
+			t->goti.link_order = link->obj_order++;
 		}
 
 		t->text[0] = 0;
@@ -755,32 +775,35 @@ static void g_put_chars(void *p_, unsigned char *s, int l)
 		p->text = t;
 	}
 	if (p->pending_text_len == -1) {
-		p->pending_text_len = strlen(p->text->text);
+		p->pending_text_len = (int)strlen(cast_const_char p->text->text);
 		ptl = p->pending_text_len;
 		if (!ptl) ptl = 1;
 		goto a1;
 	}
 	ptl = p->pending_text_len;
 	if (!ptl) ptl = 1;
+	safe_add(safe_add(ptl, l), ALLOC_GR);
 	if (((ptl + ALLOC_GR - 1) & ~(ALLOC_GR - 1)) != ((ptl + l + ALLOC_GR - 1) & ~(ALLOC_GR - 1))) a1: {
 		struct g_object_text *t;
 		if ((unsigned)l > MAXINT) overalloc();
 		if ((unsigned)ptl + (unsigned)l > MAXINT - ALLOC_GR) overalloc();
 		t = mem_realloc(p->text, sizeof(struct g_object_text) + ((ptl + l + ALLOC_GR - 1) & ~(ALLOC_GR - 1)));
-		if (p->w.last_wrap >= p->text->text && p->w.last_wrap < p->text->text + p->pending_text_len) p->w.last_wrap += (unsigned char *)t - (unsigned char *)p->text;
-		if (p->w.last_wrap_obj == p->text) p->w.last_wrap_obj = t;
+		if (p->w.last_wrap >= p->text->text && p->w.last_wrap < p->text->text + p->pending_text_len) p->w.last_wrap = p->w.last_wrap + ((unsigned char *)t - (unsigned char *)p->text);
+		if (p->w.last_wrap_obj == &p->text->goti.go) p->w.last_wrap_obj = &t->goti.go;
 		p->text = t;
 	}
-	memcpy(p->text->text + p->pending_text_len, s, l), p->text->text[p->pending_text_len += l] = 0;
-	p->text->xw += g_text_width(p->text->style, p->text->text + p->pending_text_len - l); /* !!! FIXME: move to g_wrap_text */
-	if (par_format.align != AL_NO) {
+	memcpy(p->text->text + p->pending_text_len, s, l), p->text->text[p->pending_text_len = safe_add(p->pending_text_len, l)] = 0;
+	qw = g_text_width(p->text->style, p->text->text + p->pending_text_len - l);
+	p->text->goti.go.xw = safe_add(p->text->goti.go.xw, qw); /* !!! FIXME: move to g_wrap_text */
+	if (par_format.align != AL_NO /*&& par_format.align != AL_NO_BREAKABLE*/) {
 		p->w.text = p->text->text + p->pending_text_len - l;
 		p->w.style = p->text->style;
-		p->w.obj = p->text;
+		p->w.obj = &p->text->goti.go;
 		p->w.width = rm(par_format) - par_format.leftmargin * G_HTML_MARGIN;
+		p->w.force_break = 0;
 		if (p->w.width < 0) p->w.width = 0;
 		if (!g_wrap_text(&p->w)) {
-			split_line_object(p, p->w.last_wrap_obj, p->w.last_wrap);
+			split_line_object(p, get_struct(p->w.last_wrap_obj, struct g_object_text, goti.go), p->w.last_wrap);
 		}
 	}
 	return;
@@ -790,9 +813,9 @@ static void g_put_chars(void *p_, unsigned char *s, int l)
 	process_link:
 	if ((last_link /*|| last_target*/ || last_image || last_form) &&
 	    !putchars_link_ptr &&
-	    !xstrcmp(format.link, last_link) && !xstrcmp(format.target, last_target) &&
-	    !xstrcmp(format.image, last_image) && format.form == last_form
-	    && ((!format.js_event && !last_js_event) || !compare_js_event_spec(format.js_event, last_js_event))) {
+	    !xstrcmp(format_.link, last_link) && !xstrcmp(format_.target, last_target) &&
+	    !xstrcmp(format_.image, last_image) && format_.form == last_form
+	    && ((!format_.js_event && !last_js_event) || !compare_js_event_spec(format_.js_event, last_js_event))) {
 		if (!p->data) goto back_link;
 		if (!p->data->nlinks) {
 			internal("no link");
@@ -808,36 +831,36 @@ static void g_put_chars(void *p_, unsigned char *s, int l)
 		last_link = last_target = last_image = NULL;
 		last_form = NULL;
 		last_js_event = NULL;
-		if (!(format.link || format.image || format.form || format.js_event)) goto back_link;
+		if (!(format_.link || format_.image || format_.form || format_.js_event)) goto back_link;
 		/*if (d_opt->num_links) {
 			unsigned char s[64];
-			unsigned char *fl = format.link, *ft = format.target, *fi = format.image;
-			struct form_control *ff = format.form;
-			struct js_event_spec *js = format.js_event;
-			format.link = format.target = format.image = NULL;
-			format.form = NULL;
-			format.js_event = NULL;
+			unsigned char *fl = format_.link, *ft = format_.target, *fi = format_.image;
+			struct form_control *ff = format_.form;
+			struct js_event_spec *js = format_.js_event;
+			format_.link = format_.target = format_.image = NULL;
+			format_.form = NULL;
+			format_.js_event = NULL;
 			s[0] = '[';
 			snzprint(s + 1, 62, p->link_num);
-			strcat(s, "]");
-			g_put_chars(p, s, strlen(s));
+			strcat(cast_char s, "]");
+			g_put_chars(p, s, strlen(cast_const_char s));
 			if (ff && ff->type == FC_TEXTAREA) g_line_break(p);
 			if (p->cx < par_format.leftmargin * G_HTML_MARGIN) p->cx = par_format.leftmargin * G_HTML_MARGIN;
-			format.link = fl, format.target = ft, format.image = fi;
-			format.form = ff;
-			format.js_event = js;
+			format_.link = fl, format_.target = ft, format_.image = fi;
+			format_.form = ff;
+			format_.js_event = js;
 		}*/
 		p->link_num++;
-		last_link = stracpy(format.link);
-		last_target = stracpy(format.target);
-		last_image = stracpy(format.image);
-		last_form = format.form;
-		copy_js_event_spec(&last_js_event, format.js_event);
+		last_link = stracpy(format_.link);
+		last_target = stracpy(format_.target);
+		last_image = stracpy(format_.image);
+		last_form = format_.form;
+		copy_js_event_spec(&last_js_event, format_.js_event);
 		if (!p->data) goto back_link;
 		if (!(link = new_link(p->data))) goto back_link;
 		link->num = p->link_num - 1;
 		link->pos = DUMMY;
-		copy_js_event_spec(&link->js_event, format.js_event);
+		copy_js_event_spec(&link->js_event, format_.js_event);
 		if (!last_form) {
 			link->type = L_LINK;
 			link->where = stracpy(last_link);
@@ -854,51 +877,35 @@ static void g_put_chars(void *p_, unsigned char *s, int l)
 	goto back_link;
 }
 
-static void g_do_format(char *start, char *end, struct g_part *part, unsigned char *head)
+static void g_do_format(unsigned char *start, unsigned char *end, struct g_part *part, unsigned char *head)
 {
 	pr(
-	parse_html(start, end, (void (*)(void *, unsigned char *, int))g_put_chars, g_line_break, (void *(*)(void *, int, ...))g_html_special, part, head);
-	/*if ((part->y -= line_breax) < 0) part->y = 0;*/
+	parse_html(start, end, g_put_chars, g_line_break, g_html_special, part, head);
 	flush_pending_text_to_line(part);
 	flush_pending_line_to_obj(part, 0);
 	) {};
 }
 
-struct g_table_cache_entry {
-	struct g_table_cache_entry *next;
-	struct g_table_cache_entry *prev;
-	unsigned char *start;
-	unsigned char *end;
-	int align;
-	int m;
-	int width;
-	int link_num;
-	struct g_part p;
-};
-
-static struct list_head g_table_cache = { &g_table_cache, &g_table_cache };
-
-void g_free_table_cache(void)
-{
-	free_list(g_table_cache);
-}
-
 struct g_part *g_format_html_part(unsigned char *start, unsigned char *end, int align, int m, int width, unsigned char *head, int link_num, unsigned char *bg, unsigned char *bgcolor, struct f_data *f_d)
 {
 	int wa;
 	struct g_part *p;
 	struct html_element *e;
 	struct form_control *fc;
+	struct list_head *lfc;
 	int lm = margin;
 
-	struct g_table_cache_entry *tce;
+	if (par_format.implicit_pre_wrap) {
+		int limit = d_opt->xw - G_SCROLL_BAR_WIDTH;
+		if (table_level) limit -= 2 * G_HTML_MARGIN * d_opt->margin;
+		if (limit < 0) limit = d_opt->xw;
+		if (width > limit)
+			width = limit;
+	}
 
-	if (!f_d) foreach(tce, g_table_cache) {
-		if (tce->start == start && tce->end == end && tce->align == align && tce->m == m && tce->width == width && tce->link_num == link_num) {
-			p = mem_alloc(sizeof(struct g_part));
-			memcpy(p, &tce->p, sizeof(struct part));
-			return p;
-		}
+	if (!f_d) {
+		p = find_table_cache_entry(start, end, align, m, width, 0, link_num);
+		if (p) return p;
 	}
 	margin = m;
 
@@ -916,17 +923,15 @@ struct g_part *g_format_html_part(unsigned char *start, unsigned char *end, int
 	p = mem_calloc(sizeof(struct g_part));
 	{
 		struct g_object_area *a;
-		a = mem_calloc(sizeof(struct g_object_area) + sizeof(struct g_object_line *));
+		a = mem_calloc(sizeof(struct g_object_area));
 		a->bg = get_background(bg, bgcolor);
-		if (bgcolor) decode_color(bgcolor, &format.bg);
+		if (bgcolor) decode_color(bgcolor, &format_.bg);
 		if (bgcolor) decode_color(bgcolor, &par_format.bgcolor);
-		a->mouse_event = g_area_mouse;
-		a->draw = g_area_draw;
-		a->destruct = g_area_destruct;
-		a->get_list = g_area_get_list;
-		a->lfo = DUMMY;
-		a->rfo = DUMMY;
-		a->n_lines = 0;
+		a->go.mouse_event = g_area_mouse;
+		a->go.draw = g_area_draw;
+		a->go.destruct = g_area_destruct;
+		a->go.get_list = g_area_get_list;
+		/*a->n_lines = 0;*/
 		p->root = a;
 		init_list(p->uf);
 	}
@@ -944,9 +949,11 @@ struct g_part *g_format_html_part(unsigned char *start, unsigned char *end, int
 	par_format.list_level = 0;
 	par_format.list_number = 0;
 	par_format.dd_margin = 0;
+	if (align == AL_NO || align == AL_NO_BREAKABLE)
+		format_.attr |= AT_FIXED;
 	p->cx = -1;
 	p->cx_w = 0;
-	g_nobreak = 1;
+	g_nobreak = align != AL_NO && par_format.align != AL_NO_BREAKABLE;
 	g_do_format(start, end, p, head);
 	g_nobreak = 0;
 	line_breax = 1;
@@ -961,9 +968,9 @@ struct g_part *g_format_html_part(unsigned char *start, unsigned char *end, int
 
 	wa = g_get_area_width(p->root);
 	if (wa > p->x) p->x = wa;
-	g_x_extend_area(p->root, p->x, 0);
+	g_x_extend_area(p->root, p->x, 0, align);
 	if (p->x > p->xmax) p->xmax = p->x;
-	p->y = p->root->yw;
+	p->y = p->root->go.yw;
 	/*debug("WIDTH: obj (%d, %d), p (%d %d)", p->root->xw, p->root->yw, p->x, p->y);*/
 
 	kill_html_stack_item(&html_top);
@@ -971,8 +978,8 @@ struct g_part *g_format_html_part(unsigned char *start, unsigned char *end, int
 	if (cached_font_face && cached_font_face != to_je_ale_prasarna) mem_free(cached_font_face);
 	cached_font_face = to_je_ale_prasarna;
 
-	foreach(fc, p->uf) destroy_fc(fc);
-	free_list(p->uf);
+	foreach(struct form_control, fc, lfc, p->uf) destroy_fc(fc);
+	free_list(struct form_control, p->uf);
 
 	margin = lm;
 
@@ -985,15 +992,7 @@ struct g_part *g_format_html_part(unsigned char *start, unsigned char *end, int
 	last_js_event = NULL;
 
 	if (table_level > 1 && !f_d) {
-		tce = mem_alloc(sizeof(struct g_table_cache_entry));
-		tce->start = start;
-		tce->end = end;
-		tce->align = align;
-		tce->m = m;
-		tce->width = width;
-		tce->link_num = link_num;
-		memcpy(&tce->p, p, sizeof(struct g_part));
-		add_to_list(g_table_cache, tce);
+		add_table_cache_entry(start, end, align, m, width, 0, link_num, p);
 	}
 
 	return p;
@@ -1001,26 +1000,19 @@ struct g_part *g_format_html_part(unsigned char *start, unsigned char *end, int
 
 void g_release_part(struct g_part *p)
 {
-	if (p->text) p->text->destruct(p->text);
-	if (p->line) p->line->destruct(p->line);
-	if (p->root) p->root->destruct(p->root);
+	if (p->text) p->text->goti.go.destruct(&p->text->goti.go);
+	if (p->line) p->line->go.destruct(&p->line->go);
+	if (p->root) p->root->go.destruct(&p->root->go);
 	if (p->current_style) g_free_style(p->current_style);
 }
 
-static void g_scan_width(struct g_object **o, int n, int *w)
-{
-	while (n--) {
-		if ((*o)->x + (*o)->xw > *w) *w = (*o)->x + (*o)->xw;
-		o++;
-	}
-}
-
 static void g_scan_lines(struct g_object_line **o, int n, int *w)
 {
 	while (n--) {
 		if ((*o)->n_entries) {
 			struct g_object *oo = (*o)->entries[(*o)->n_entries - 1];
-			if ((*o)->x + oo->x + oo->xw > *w) *w = (*o)->x + oo->x + oo->xw;
+			if (safe_add(safe_add((*o)->go.x, oo->x), oo->xw) > *w)
+				*w = (*o)->go.x + oo->x + oo->xw;
 		}
 		o++;
 	}
@@ -1029,46 +1021,42 @@ static void g_scan_lines(struct g_object_line **o, int n, int *w)
 int g_get_area_width(struct g_object_area *a)
 {
 	int w = 0;
-	g_scan_width(a->lfo, a->n_lfo, &w);
-	g_scan_width(a->rfo, a->n_rfo, &w);
 	g_scan_lines(a->lines, a->n_lines, &w);
 	return w;
 }
 
-void g_x_extend_area(struct g_object_area *a, int width, int height)
+void g_x_extend_area(struct g_object_area *a, int width, int height, int align)
 {
 	struct g_object_line *l;
 	int i;
-	a->xw = width;
+	a->go.xw = width;
 	for (i = 0; i < a->n_lines; i++) {
-		a->lines[i]->xw = width;
+		a->lines[i]->go.xw = width;
 	}
-	for (i = a->n_lines - 1; i >= 0; i--) {
+	if (align != AL_NO && par_format.align != AL_NO_BREAKABLE) for (i = a->n_lines - 1; i >= 0; i--) {
 		l = a->lines[i];
 		if (!l->n_entries) {
-			a->yw -= l->yw;
-			l->destruct(l);
+			a->go.yw -= l->go.yw;
+			l->go.destruct(&l->go);
 			a->n_lines--;
 			continue;
 		}
 		break;
 	}
-	if (a->yw >= height) return;
+	if (a->go.yw >= height) return;
 	l = mem_calloc(sizeof(struct g_object_line));
-	l->mouse_event = g_line_mouse;
-	l->draw = g_line_draw;
-	l->destruct = g_line_destruct;
-	l->get_list = g_line_get_list;
-	l->x = 0;
-	l->y = a->yw;
-	l->xw = width;
-	l->yw = height - a->yw;
+	l->go.mouse_event = g_line_mouse;
+	l->go.draw = g_line_draw;
+	l->go.destruct = g_line_destruct;
+	l->go.get_list = g_line_get_list;
+	l->go.x = 0;
+	l->go.y = a->go.yw;
+	l->go.xw = width;
+	l->go.yw = height - a->go.yw;
 	l->bg = a->bg;
 	l->n_entries = 0;
 	a->lines[a->n_lines] = l;
 	a->n_lines++;
 }
 
-
-
 #endif
