@@ -18,8 +18,9 @@ struct http_connection_info {
 };
 
 /* prototypes */
-static void http_send_header(struct connection *);
-static void http_get_header(struct connection *);
+static void http_send_header(struct connection *c);
+static void http_get_header(struct connection *c);
+static void test_restart(struct connection *c);
 static void add_user_agent(unsigned char **hdr, int *l);
 static void add_referer(unsigned char **hdr, int *l, unsigned char *url, unsigned char *prev_url);
 static void add_accept(unsigned char **hdr, int *l);
@@ -27,6 +28,7 @@ static void add_accept_language(unsigned char **hdr, int *l, struct http_connect
 static void add_accept_encoding(unsigned char **hdr, int *l, unsigned char *url, struct connection *c);
 static void add_accept_charset(unsigned char **hdr, int *l, struct http_connection_info *info);
 static void add_connection(unsigned char **hdr, int *l, int http10, int proxy, int post);
+static void add_upgrade(unsigned char **hdr, int *l);
 static void add_dnt(unsigned char **hdr, int *l);
 static void add_if_modified(unsigned char **hdr, int *l, struct connection *c);
 static void add_range(unsigned char **hdr, int *l, unsigned char *url, struct connection *c);
@@ -55,15 +57,13 @@ unsigned char *parse_http_header(unsigned char *head, unsigned char *item, unsig
 			for (g = ++f; *g >= ' '; g++)
 				;
 			while (g > f && g[-1] == ' ') g--;
-			h = mem_alloc(g - f + 1);
-			memcpy(h, f, g - f);
-			h[g - f] = 0;
+			h = memacpy(f, g - f);
 			if (ptr) {
 				*ptr = f;
 			}
 			return h;
 		}
-		cont:;
+		cont:
 		f--;
 	}
 	return NULL;
@@ -192,7 +192,7 @@ static void add_url_to_str(unsigned char **str, int *l, unsigned char *url)
 {
 	unsigned char *sp;
 	for (sp = url; *sp && *sp != POST_CHAR; sp++) {
-		if (*sp <= ' ') {
+		if (*sp <= ' ' || *sp >= 127) {
 			unsigned char esc[4];
 			sprintf(cast_char esc, "%%%02X", (int)*sp);
 			add_to_str(str, l, esc);
@@ -205,7 +205,7 @@ static void add_url_to_str(unsigned char **str, int *l, unsigned char *url)
 static void http_send_header(struct connection *c)
 {
 	struct http_connection_info *info;
-	int http10 = http_options.http10 && !SCRUB_HEADERS;
+	int http10 = http_options.http10;
 	int proxy;
 	unsigned char *hdr;
 	unsigned char *h, *u;
@@ -219,8 +219,8 @@ static void http_send_header(struct connection *c)
 			c->cache->refcount--;
 	}
 
-	proxy = upcase(c->url[0]) == 'P';
-	host = !proxy ? c->url : get_url_data(c->url);
+	proxy = is_proxy_url(c->url);
+	host = remove_proxy_prefix(c->url);
 	set_connection_timeout(c);
 	info = mem_calloc(sizeof(struct http_connection_info));
 	c->info = info;
@@ -228,11 +228,7 @@ static void http_send_header(struct connection *c)
 	info->https_forward = !c->ssl && proxy && host && !casecmp(host, cast_uchar "https://", 8);
 	if (c->ssl) proxy = 0;
 #endif
-	info->send_close = info->https_forward || http10 || (post && http_options.bug_post_no_keepalive)
-#ifdef HAVE_SSL
-		|| c->ssl
-#endif
-		;
+	info->send_close = info->https_forward || http10 || (post && http_options.bug_post_no_keepalive);
 	hdr = init_str();
 	if (!host) {
 		http_bad_url:
@@ -265,7 +261,7 @@ static void http_send_header(struct connection *c)
 		goto added_connect;
 	} else
 #endif
-	  if (!post) {
+	if (!post) {
 		add_to_str(&hdr, &l, cast_uchar "GET ");
 	} else {
 		add_to_str(&hdr, &l, cast_uchar "POST ");
@@ -300,6 +296,9 @@ static void http_send_header(struct connection *c)
 	else add_to_str(&hdr, &l, cast_uchar " HTTP/1.0\r\n");
 	if (!info->https_forward && (h = get_host_name(host))) {
 		add_to_str(&hdr, &l, cast_uchar "Host: ");
+		if (*h && h[strlen(cast_const_char h) - 1] == '.') {
+			h[strlen(cast_const_char h) - 1] = 0;
+		}
 		if (h[0] == '[' && h[strlen(cast_const_char h) - 1] == ']') {
 			unsigned char *pc = cast_uchar strchr(cast_const_char h, '%');
 			if (pc) {
@@ -319,6 +318,7 @@ static void http_send_header(struct connection *c)
 	add_user_agent(&hdr, &l);
 	if (proxy) add_proxy_auth_string(&hdr, &l, c->url);
 	if (!info->https_forward) {
+		test_restart(c);
 		add_referer(&hdr, &l, host, c->prev_url);
 		add_accept(&hdr, &l);
 		add_accept_language(&hdr, &l, info);
@@ -326,6 +326,7 @@ static void http_send_header(struct connection *c)
 		add_accept_charset(&hdr, &l, info);
 		add_dnt(&hdr, &l);
 		add_connection(&hdr, &l, http10, proxy, !info->send_close);
+		add_upgrade(&hdr, &l);
 		add_if_modified(&hdr, &l, c);
 		add_range(&hdr, &l, host, c);
 		add_pragma_no_cache(&hdr, &l, c->no_cache);
@@ -351,11 +352,33 @@ static void http_send_header(struct connection *c)
 	setcstate(c, S_SENT);
 }
 
+static void test_restart(struct connection *c)
+{
+/* If the cached entity is compressed, request the whole file and turn off compression */
+	if (c->cache && c->from) {
+		unsigned char *d;
+		if ((d = parse_http_header(c->cache->head, cast_uchar "Content-Encoding", NULL))) {
+			mem_free(d);
+			c->from = 0;
+			c->no_compress = 1;
+#ifdef HAVE_ANY_COMPRESSION
+			if (c->tries >= 1) {
+				unsigned char *h;
+				if ((h = get_host_name(c->url))) {
+					add_blacklist_entry(h, BL_NO_COMPRESSION);
+					mem_free(h);
+				}
+			}
+#endif
+		}
+	}
+}
+
 static void add_user_agent(unsigned char **hdr, int *l)
 {
 	add_to_str(hdr, l, cast_uchar "User-Agent: ");
 	if (SCRUB_HEADERS) {
-		add_to_str(hdr, l, cast_uchar "Mozilla/5.0 (Windows NT 6.1; rv:31.0) Gecko/20100101 Firefox/31.0\r\n");
+		add_to_str(hdr, l, cast_uchar "Mozilla/5.0 (Windows NT 6.1; rv:52.0) Gecko/20100101 Firefox/52.0\r\n");
 	} else if (!(*http_options.header.fake_useragent)) {
 		add_to_str(hdr, l, cast_uchar("Links (" VERSION_STRING "; "));
 		add_to_str(hdr, l, system_name);
@@ -364,8 +387,9 @@ static void add_user_agent(unsigned char **hdr, int *l)
 		add_to_str(hdr, l, cast_uchar "; ");
 		if (!F && !list_empty(terminals)) {
 			struct terminal *term;
+			struct list_head *lterm;
 			unsigned char *t = cast_uchar "text";
-			foreach(term, terminals) if (term->spec->braille) t = cast_uchar "braille";
+			foreach(struct terminal, term, lterm, terminals) if (term->spec->braille) t = cast_uchar "braille";
 			add_to_str(hdr, l, t);
 		}
 #ifdef G
@@ -385,38 +409,44 @@ static void add_user_agent(unsigned char **hdr, int *l)
 
 static void add_referer(unsigned char **hdr, int *l, unsigned char *url, unsigned char *prev_url)
 {
-	if (SCRUB_HEADERS)
-		return;
-	switch (http_options.header.referer)
-	{
-		case REFERER_FAKE:
-		add_to_str(hdr, l, cast_uchar "Referer: ");
-		add_to_str(hdr, l, http_options.header.fake_referer);
-		add_to_str(hdr, l, cast_uchar "\r\n");
-		break;
+	switch (http_options.header.referer) {
+		case REFERER_FAKE: {
+			add_to_str(hdr, l, cast_uchar "Referer: ");
+			add_to_str(hdr, l, http_options.header.fake_referer);
+			add_to_str(hdr, l, cast_uchar "\r\n");
+			break;
+		}
 
-		case REFERER_SAME_URL:
-		add_to_str(hdr, l, cast_uchar "Referer: ");
-		add_url_to_str(hdr, l, url);
-		add_to_str(hdr, l, cast_uchar "\r\n");
-		break;
+		case REFERER_SAME_URL: {
+			add_to_str(hdr, l, cast_uchar "Referer: ");
+			add_url_to_str(hdr, l, url);
+			add_to_str(hdr, l, cast_uchar "\r\n");
+			break;
+		}
 
-		case REFERER_REAL_SAME_SERVER:
-		{
+		case REFERER_REAL_SAME_SERVER: {
 			unsigned char *h, *j;
 			int brk = 1;
 			if ((h = get_host_name(url))) {
 				if ((j = get_host_name(prev_url))) {
-					if (!strcasecmp(cast_const_char h, cast_const_char j)) brk = 0;
+					if (!casestrcmp(h, j)) brk = 0;
+					else if (!casestrcmp(h, cast_uchar "imageproxy.jxs.cz")) {
+						int l = (int)strlen(cast_const_char j);
+						int q = (int)strlen(".blog.cz");
+						if (l > q && !casestrcmp((j + l - q), cast_uchar ".blog.cz")) brk = 0;
+						else if (!casestrcmp(j, cast_uchar "blog.cz")) brk = 0;
+					} else if (!casestrcmp(h, cast_uchar "www.google.com")) {
+						unsigned char *c = get_url_data(url);
+						if (c && !strncmp(cast_const_char c, "recaptcha/api/", 14)) brk = 0;
+					}
 					mem_free(j);
 				}
 				mem_free(h);
 			}
 			if (brk) break;
-			/* fall through */
 		}
-		case REFERER_REAL:
-		{
+			/*-fallthrough*/
+		case REFERER_REAL: {
 			unsigned char *ref;
 			unsigned char *user, *ins;
 			int ulen;
@@ -430,8 +460,8 @@ static void add_referer(unsigned char **hdr, int *l, unsigned char *url, unsigne
 			add_url_to_str(hdr, l, ref);
 			add_to_str(hdr, l, cast_uchar "\r\n");
 			mem_free(ref);
+			break;
 		}
-		break;
 	}
 }
 
@@ -449,11 +479,11 @@ static void add_accept_language(unsigned char **hdr, int *l, struct http_connect
 	if (!(info->bl_flags & BL_NO_ACCEPT_LANGUAGE)) {
 		add_to_str(hdr, l, cast_uchar "Accept-Language: ");
 		if (SCRUB_HEADERS) {
-			add_to_str(hdr, l, cast_uchar "en-us,en;q=0.5\r\n");
+			add_to_str(hdr, l, cast_uchar "en-US,en;q=0.5\r\n");
 		} else {
 			int la;
 			la = *l;
-			add_to_str(hdr, l, _(TEXT_(T__ACCEPT_LANGUAGE), NULL));
+			add_to_str(hdr, l, get_text_translation(TEXT_(T__ACCEPT_LANGUAGE), NULL));
 			add_to_str(hdr, l, cast_uchar ",");
 			if (!strstr(cast_const_char(*hdr + la), "en,") &&
 			    !strstr(cast_const_char(*hdr + la), "en;")) add_to_str(hdr, l, cast_uchar "en;q=0.2,");
@@ -484,7 +514,7 @@ static int advertise_compression(unsigned char *url, struct connection *c)
 
 static void add_accept_encoding(unsigned char **hdr, int *l, unsigned char *url, struct connection *c)
 {
-#if defined(HAVE_ZLIB) || defined(HAVE_BZIP2) || defined(HAVE_LZMA)
+#if defined(HAVE_ANY_COMPRESSION)
 #define info	((struct http_connection_info *)c->info)
 	if (advertise_compression(url, c)) {
 		int orig_l = *l;
@@ -495,6 +525,16 @@ static void add_accept_encoding(unsigned char **hdr, int *l, unsigned char *url,
 		if (*l != l1) add_to_str(hdr, l, cast_uchar ", ");
 		add_to_str(hdr, l, cast_uchar "gzip, deflate");
 #endif
+#if defined(HAVE_BROTLI)
+		if ((!SCRUB_HEADERS
+#ifdef HAVE_SSL
+		    || c->ssl
+#endif
+		    ) && !(info->bl_flags & BL_NO_BZIP2)) {
+			if (*l != l1) add_to_str(hdr, l, cast_uchar ", ");
+			add_to_str(hdr, l, cast_uchar "br");
+		}
+#endif
 #if defined(HAVE_BZIP2)
 		if (!SCRUB_HEADERS && !(info->bl_flags & BL_NO_BZIP2)) {
 			if (*l != l1) add_to_str(hdr, l, cast_uchar ", ");
@@ -505,7 +545,13 @@ static void add_accept_encoding(unsigned char **hdr, int *l, unsigned char *url,
 #if defined(HAVE_LZMA) && !defined(DOS)
 		if (!SCRUB_HEADERS && !(info->bl_flags & BL_NO_BZIP2)) {
 			if (*l != l1) add_to_str(hdr, l, cast_uchar ", ");
-			add_to_str(hdr, l, cast_uchar "lzma,lzma2");
+			add_to_str(hdr, l, cast_uchar "lzma, lzma2");
+		}
+#endif
+#if defined(HAVE_LZIP)
+		if (!SCRUB_HEADERS && !(info->bl_flags & BL_NO_BZIP2)) {
+			if (*l != l1) add_to_str(hdr, l, cast_uchar ", ");
+			add_to_str(hdr, l, cast_uchar "lzip");
 		}
 #endif
 		if (*l != l1) add_to_str(hdr, l, cast_uchar "\r\n");
@@ -518,8 +564,12 @@ static void add_accept_encoding(unsigned char **hdr, int *l, unsigned char *url,
 static void add_accept_charset(unsigned char **hdr, int *l, struct http_connection_info *info)
 {
 	static unsigned char *accept_charset = NULL;
-	if (SCRUB_HEADERS)
+
+	if (SCRUB_HEADERS ||
+	    info->bl_flags & BL_NO_CHARSET ||
+	    http_options.no_accept_charset)
 		return;
+
 	if (!accept_charset) {
 		int i;
 		unsigned char *cs, *ac;
@@ -531,17 +581,14 @@ static void add_accept_charset(unsigned char **hdr, int *l, struct http_connecti
 			add_to_str(&ac, &aclen, cs);
 		}
 		if (aclen) add_to_str(&ac, &aclen, cast_uchar "\r\n");
-		retry:
-		if (!(accept_charset = malloc(strlen(cast_const_char ac) + 1))) {
-			if (out_of_memory(0, NULL, 0))
-				goto retry;
+		if (!(accept_charset = cast_uchar strdup(cast_const_char ac))) {
+			add_to_str(hdr, l, ac);
 			mem_free(ac);
 			return;
 		}
-		strcpy(cast_char accept_charset, cast_const_char ac);
 		mem_free(ac);
 	}
-	if (!(info->bl_flags & BL_NO_CHARSET) && !http_options.no_accept_charset) add_to_str(hdr, l, accept_charset);
+	add_to_str(hdr, l, accept_charset);
 }
 
 static void add_dnt(unsigned char **hdr, int *l)
@@ -560,6 +607,13 @@ static void add_connection(unsigned char **hdr, int *l, int http10, int proxy, i
 	}
 }
 
+static void add_upgrade(unsigned char **hdr, int *l)
+{
+	if (proxies.only_proxies) {
+		add_to_str(hdr, l, cast_uchar "Upgrade-Insecure-Requests: 1\r\n");
+	}
+}
+
 static void add_if_modified(unsigned char **hdr, int *l, struct connection *c)
 {
 	struct cache_entry *e;
@@ -593,17 +647,6 @@ static void add_range(unsigned char **hdr, int *l, unsigned char *url, struct co
 			return;
 	}
 	if (c->from /*&& (c->est_length == -1 || c->from < c->est_length)*/ && c->no_cache < NC_IF_MOD && !(info->bl_flags & BL_NO_RANGE)) {
-/* If the cached entity is compressed and we turned off compression,
-   request the whole file */
-#ifdef HAVE_ANY_COMPRESSION
-		if (!advertise_compression(url, c) && e) {
-			unsigned char *d;
-			if ((d = parse_http_header(e->head, cast_uchar "Transfer-Encoding", NULL))) {
-				mem_free(d);
-				return;
-			}
-		}
-#endif
 		add_to_str(hdr, l, cast_uchar "Range: bytes=");
 		add_num_to_str(hdr, l, c->from);
 		add_to_str(hdr, l, cast_uchar "-\r\n");
@@ -788,7 +831,7 @@ static void read_http_data(struct connection *c, struct read_buffer *rb)
 			if (!info->chunk_remaining && rb->len >= 1) {
 				if (rb->data[0] == 10) kill_buffer_data(rb, 1);
 				else {
-					if (rb->data[0] != 13 || (rb->len >= 2 && rb->data[1] != 10)) {
+					if (rb->data[0] != 13 || (rb->len >= 2 && ((unsigned char *)rb->data)[1] != 10)) {
 						setcstate(c, S_HTTP_ERROR);
 						abort_connection(c);
 						return;
@@ -812,11 +855,11 @@ static int get_header(struct read_buffer *rb)
 	if (rb->len <= 0) return 0;
 	if (rb->data[0] != 'H') return -2;
 	if (rb->len <= 1) return 0;
-	if (rb->data[1] != 'T') return -2;
+	if (((unsigned char *)rb->data)[1] != 'T') return -2;
 	if (rb->len <= 2) return 0;
-	if (rb->data[2] != 'T') return -2;
+	if (((unsigned char *)rb->data)[2] != 'T') return -2;
 	if (rb->len <= 3) return 0;
-	if (rb->data[3] != 'P') return -2;
+	if (((unsigned char *)rb->data)[3] != 'P') return -2;
 	for (i = 0; i < rb->len; i++) {
 		unsigned char a = rb->data[i];
 		if (/*a < ' ' && a != 10 && a != 13*/ !a) return -1;
@@ -837,25 +880,24 @@ static void http_got_header(struct connection *c, struct read_buffer *rb)
 	off_t cf;
 	int state = c->state != S_PROC ? S_GETH : S_PROC;
 	unsigned char *head;
-	unsigned char *cookie, *ch;
 	int a, h = 0, version = 0;	/* against warning */
 	unsigned char *d;
 	struct cache_entry *e;
 	int previous_http_code;
 	struct http_connection_info *info;
-	unsigned char *host = upcase(c->url[0]) != 'P' ? c->url : get_url_data(c->url);
+	unsigned char *host = remove_proxy_prefix(c->url);
 	set_connection_timeout(c);
 	info = c->info;
 	if (rb->close == 2) {
-		unsigned char *h;
-		if (!c->tries && (h = get_host_name(host))) {
+		unsigned char *hs;
+		if (!c->tries && (hs = get_host_name(host))) {
 			if (info->bl_flags & BL_NO_CHARSET) {
-				del_blacklist_entry(h, BL_NO_CHARSET);
+				del_blacklist_entry(hs, BL_NO_CHARSET);
 			} else {
-				add_blacklist_entry(h, BL_NO_CHARSET);
+				add_blacklist_entry(hs, BL_NO_CHARSET);
 				c->tries = -1;
 			}
-			mem_free(h);
+			mem_free(hs);
 		}
 		setcstate(c, S_CANT_READ);
 		retry_connection(c);
@@ -892,12 +934,6 @@ static void http_got_header(struct connection *c, struct read_buffer *rb)
 		retry_connection(c);
 		return;
 	}
-	ch = head;
-	while ((cookie = parse_http_header(ch, cast_uchar "Set-Cookie", &ch))) {
-		unsigned char *host = upcase(c->url[0]) != 'P' ? c->url : get_url_data(c->url);
-		set_cookie(NULL, host, cookie);
-		mem_free(cookie);
-	}
 	if (h == 100) {
 		mem_free(head);
 		state = S_PROC;
@@ -925,6 +961,15 @@ static void http_got_header(struct connection *c, struct read_buffer *rb)
 		return;
 	}
 #endif
+	if (h != 401 && h != 407) {
+		unsigned char *cookie;
+		unsigned char *ch = head;
+		while ((cookie = parse_http_header(ch, cast_uchar "Set-Cookie", &ch))) {
+			unsigned char *host = remove_proxy_prefix(c->url);
+			set_cookie(NULL, host, cookie);
+			mem_free(cookie);
+		}
+	}
 	if (h == 204) {
 		mem_free(head);
 		http_end_request(c, 0, 0, S_HTTP_204);
@@ -940,6 +985,18 @@ static void http_got_header(struct connection *c, struct read_buffer *rb)
 		http_end_request(c, 0, 1, S__OK);
 		return;
 	}
+	if (h == 431) {
+		unsigned char *hs;
+		if (!(info->bl_flags & BL_NO_CHARSET) && (hs = get_host_name(host))) {
+			mem_free(head);
+			add_blacklist_entry(hs, BL_NO_CHARSET);
+			mem_free(hs);
+			c->tries = -1;
+			setcstate(c, S_RESTART);
+			retry_connection(c);
+			return;
+		}
+	}
 	if ((h == 500 || h == 502 || h == 503 || h == 504) && http_options.retry_internal_errors && is_connection_restartable(c)) {
 			/* !!! FIXME: wait some time ... */
 		if (is_last_try(c)) {
@@ -955,7 +1012,7 @@ static void http_got_header(struct connection *c, struct read_buffer *rb)
 		return;
 	}
 	if (!c->cache) {
-		if (get_cache_entry(c->url, &c->cache)) {
+		if (get_connection_cache_entry(c)) {
 			mem_free(head);
 			setcstate(c, S_OUT_OF_MEM);
 			abort_connection(c);
@@ -998,19 +1055,13 @@ static void http_got_header(struct connection *c, struct read_buffer *rb)
 	}
 #ifdef HAVE_SSL
 	if (c->ssl) {
-		int l = 0;
 		if (e->ssl_info) mem_free(e->ssl_info);
-		e->ssl_info = init_str();
-		add_num_to_str(&e->ssl_info, &l, SSL_get_cipher_bits(c->ssl, NULL));
-		add_to_str(&e->ssl_info, &l, cast_uchar "-bit ");
-		add_to_str(&e->ssl_info, &l, cast_uchar SSL_get_cipher_version(c->ssl));
-		add_to_str(&e->ssl_info, &l, cast_uchar " ");
-		add_to_str(&e->ssl_info, &l, cast_uchar SSL_get_cipher_name(c->ssl));
+		e->ssl_info = get_cipher_string(c->ssl);
 	}
 #endif
 	if (e->redirect) mem_free(e->redirect), e->redirect = NULL;
+	if ((h == 302 || h == 303 || h == 307 || h == 511) && !e->expire_time) e->expire_time = 1;
 	if (h == 301 || h == 302 || h == 303 || h == 307) {
-		if ((h == 302 || h == 303 || h == 307) && !e->expire_time) e->expire_time = 1;
 		if ((d = parse_http_header(e->head, cast_uchar "Location", NULL))) {
 			unsigned char *user, *ins;
 			unsigned char *newuser, *newpassword;
@@ -1040,7 +1091,7 @@ static void http_got_header(struct connection *c, struct read_buffer *rb)
 	info->length = -1;
 	info->version = version;
 	if ((d = parse_http_header(e->head, cast_uchar "Connection", NULL)) || (d = parse_http_header(e->head, cast_uchar "Proxy-Connection", NULL))) {
-		if (!strcasecmp(cast_const_char d, "close")) info->close = 1;
+		if (!casestrcmp(d, cast_uchar "close")) info->close = 1;
 		mem_free(d);
 	} else if (version < 11) info->close = 1;
 	cf = c->from;
@@ -1048,7 +1099,7 @@ static void http_got_header(struct connection *c, struct read_buffer *rb)
 	if ((d = parse_http_header(e->head, cast_uchar "Content-Range", NULL))) {
 		if (strlen(cast_const_char d) > 6) {
 			d[5] = 0;
-			if (!(strcasecmp(cast_const_char d, "bytes")) && d[6] >= '0' && d[6] <= '9') {
+			if (!(casestrcmp(d, cast_uchar "bytes")) && d[6] >= '0' && d[6] <= '9') {
 				my_strtoll_t f = my_strtoll(d + 6, NULL);
 				if (f >= 0 && (off_t)f >= 0 && (off_t)f == f) c->from = f;
 			}
@@ -1074,14 +1125,14 @@ static void http_got_header(struct connection *c, struct read_buffer *rb)
 		mem_free(d);
 	}
 	if ((d = parse_http_header(e->head, cast_uchar "Accept-Ranges", NULL))) {
-		if (!strcasecmp(cast_const_char d, "none") && !c->unrestartable) c->unrestartable = 1;
+		if (!casestrcmp(d, cast_uchar "none") && !c->unrestartable) c->unrestartable = 1;
 		mem_free(d);
 	} else {
 		if (!c->unrestartable && !c->from) c->unrestartable = 1;
 	}
 	if (info->bl_flags & BL_NO_RANGE && !c->unrestartable) c->unrestartable = 1;
 	if ((d = parse_http_header(e->head, cast_uchar "Transfer-Encoding", NULL))) {
-		if (!strcasecmp(cast_const_char d, "chunked")) {
+		if (!casestrcmp(d, cast_uchar "chunked")) {
 			info->length = -2;
 			info->chunk_remaining = -1;
 		}
@@ -1089,7 +1140,7 @@ static void http_got_header(struct connection *c, struct read_buffer *rb)
 	}
 	if (!info->close && info->length == -1) info->close = 1;
 	if ((d = parse_http_header(e->head, cast_uchar "Last-Modified", NULL))) {
-		if (e->last_modified && strcasecmp(cast_const_char e->last_modified, cast_const_char d)) {
+		if (e->last_modified && casestrcmp(e->last_modified, d)) {
 			delete_entry_content(e);
 			if (c->from) {
 				c->from = 0;
@@ -1121,6 +1172,11 @@ static void http_got_header(struct connection *c, struct read_buffer *rb)
 		truncate_entry(e, c->from, 0);
 	}
 
+	if (info->https_forward && h == 407) {
+		http_end_request(c, 0, 1, S__OK);
+		return;
+	}
+
 	read_http_data(c, rb);
 }
 
