@@ -12,16 +12,18 @@ static void unhandle_basic_signals(struct terminal *);
 static void poll_fg(void *);
 
 #ifdef WIN
-static void sig_terminate(struct terminal *t)
+static void sig_terminate(void *t_)
 {
+	struct terminal *t = (struct terminal *)t_;
 	unhandle_basic_signals(t);
 	terminate_loop = 1;
 	retval = RET_SIGNAL;
 }
 #endif
 
-static void sig_intr(struct terminal *t)
+static void sig_intr(void *t_)
 {
+	struct terminal *t = (struct terminal *)t_;
 	if (!t) {
 		unhandle_basic_signals(t);
 		terminate_loop = 1;
@@ -31,7 +33,7 @@ static void sig_intr(struct terminal *t)
 	}
 }
 
-static void sig_ctrl_c(struct terminal *t)
+static void sig_ctrl_c(void *t_)
 {
 	if (!is_blocked()) kbd_ctrl_c();
 }
@@ -44,8 +46,9 @@ static void sig_ign(void *x)
 
 static struct timer *fg_poll_timer = NULL;
 
-void sig_tstp(struct terminal *t)
+void sig_tstp(void *t_)
 {
+	struct terminal *t = (struct terminal *)t_;
 #if defined(SIGSTOP) && !defined(NO_CTRL_Z)
 #if defined(SIGCONT) && defined(SIGTTOU) && defined(HAVE_GETPID)
 	pid_t pid, newpid;
@@ -64,7 +67,7 @@ void sig_tstp(struct terminal *t)
 	if (!newpid) {
 		while (1) {
 			int rr;
-			sleep(1);
+			portable_sleep(1000);
 			EINTRLOOP(rr, kill(pid, SIGCONT));
 		}
 	}
@@ -84,8 +87,9 @@ void sig_tstp(struct terminal *t)
 	fg_poll_timer = install_timer(FG_POLL_TIME, poll_fg, t);
 }
 
-static void poll_fg(void *t)
+static void poll_fg(void *t_)
 {
+	struct terminal *t = (struct terminal *)t_;
 	int r;
 	fg_poll_timer = NULL;
 	if (!F) {
@@ -106,7 +110,7 @@ static void poll_fg(void *t)
 	}
 }
 
-void sig_cont(struct terminal *t)
+void sig_cont(void *t_)
 {
 	if (!F) {
 		unblock_itrm(1);
@@ -115,28 +119,27 @@ void sig_cont(struct terminal *t)
 		drv->unblock(NULL);
 #endif
 	}
-	/*else register_bottom_half(raise, SIGSTOP);*/
 }
 
 static void handle_basic_signals(struct terminal *term)
 {
-	install_signal_handler(SIGHUP, (void (*)(void *))sig_intr, term, 0);
-	if (!F) install_signal_handler(SIGINT, (void (*)(void *))sig_ctrl_c, term, 0);
-	/*install_signal_handler(SIGTERM, (void (*)(void *))sig_terminate, term, 0);*/
+	install_signal_handler(SIGHUP, sig_intr, term, 0);
+	if (!F) install_signal_handler(SIGINT, sig_ctrl_c, term, 0);
+	/*install_signal_handler(SIGTERM, sig_terminate, term, 0);*/
 #ifdef WIN
-	install_signal_handler(SIGQUIT, (void (*)(void *))sig_terminate, term, 0);
+	install_signal_handler(SIGQUIT, sig_terminate, term, 0);
 #endif
 #ifdef SIGTSTP
-	if (!F) install_signal_handler(SIGTSTP, (void (*)(void *))sig_tstp, term, 0);
+	if (!F) install_signal_handler(SIGTSTP, sig_tstp, term, 0);
 #endif
 #ifdef SIGTTIN
-	if (!F) install_signal_handler(SIGTTIN, (void (*)(void *))sig_tstp, term, 0);
+	if (!F) install_signal_handler(SIGTTIN, sig_tstp, term, 0);
 #endif
 #ifdef SIGTTOU
-	install_signal_handler(SIGTTOU, (void (*)(void *))sig_ign, term, 0);
+	install_signal_handler(SIGTTOU, sig_ign, term, 0);
 #endif
 #ifdef SIGCONT
-	if (!F) install_signal_handler(SIGCONT, (void (*)(void *))sig_cont, term, 0);
+	if (!F) install_signal_handler(SIGCONT, sig_cont, term, 0);
 #endif
 }
 
@@ -247,6 +250,8 @@ err_close:
 
 static void gfx_connection_terminate(void *p)
 {
+	int h = (int)(my_intptr_t)p;
+	set_handlers(h, NULL, NULL, NULL);
 	terminate_loop = 1;
 }
 
@@ -265,8 +270,9 @@ static void end_dump(struct object_request *r, void *p)
 	if (dmp == D_SOURCE) {
 		if (ce) {
 			struct fragment *frag;
+			struct list_head *lfrag;
 			nextfrag:
-			foreach(frag, ce->frag) if (frag->offset <= dump_pos && frag->offset + frag->length > dump_pos) {
+			foreach(struct fragment, frag, lfrag, ce->frag) if (frag->offset <= dump_pos && frag->offset + frag->length > dump_pos) {
 				off_t l;
 				int w;
 				l = frag->length - (dump_pos - frag->offset);
@@ -288,19 +294,22 @@ static void end_dump(struct object_request *r, void *p)
 	} else if (ce) {
 		struct document_options o;
 		struct f_data_c *fd;
-		if (!(fd = create_f_data_c(NULL, NULL))) goto terminate;
+		fd = create_f_data_c(NULL, NULL);
 		memset(&o, 0, sizeof(struct document_options));
 		o.xp = 0;
 		o.yp = 1;
 		o.xw = screen_width;
 		o.yw = 25;
 		o.col = 0;
-		o.cp = dump_codepage == -1 ? 0 : dump_codepage;
+		o.cp = get_commandline_charset();
 		ds2do(&dds, &o, 0);
 		o.plain = 0;
 		o.frames = 0;
 		o.js_enable = 0;
 		o.framename = cast_uchar "";
+		if (!casecmp(r->url, cast_uchar "file://", 7) && !o.hard_assume) {
+			o.assume_cp = get_commandline_charset();
+		}
 		if (!(fd->f_data = cached_format_html(fd, r, r->url, &o, NULL))) goto term_1;
 		dump_to_file(fd->f_data, oh);
 		term_1:
@@ -369,11 +378,8 @@ static void init(void)
 	}
 	if ((dds.assume_cp = get_cp_index(cast_uchar "ISO-8859-1")) == -1) dds.assume_cp = 0;
 	load_config();
-	init_b = 1;
-	init_bookmarks();
-	create_initial_extensions();
-	load_url_history();
-	init_cookies();
+	if (proxies.only_proxies)
+		reset_settings_for_tor();
 	u = parse_options(g_argc - 1, g_argv + 1);
 	fixup_g();
 	if (!u) {
@@ -383,6 +389,7 @@ static void init(void)
 		terminate_loop = 1;
 		return;
 	}
+	init_cookies();
 	if (!dmp) {
 		if (ggr) {
 			close_socket(&terminal_pipe[0]);
@@ -450,14 +457,13 @@ static void init(void)
 							goto ttt;
 						}
 						mem_free(info);
-						set_handlers(uh, gfx_connection_terminate, NULL, NULL);
+						set_handlers(uh, gfx_connection_terminate, NULL, (void *)(my_intptr_t)uh);
 						initialize_all_subsystems_2();
-#if defined(HAVE_MALLOC_TRIM)
-						malloc_trim(0);
-#endif
+						heap_trim();
 						return;
 					}
 				}
+				spawn_font_thread();
 				init_dither(drv->depth);
 			}
 #else
@@ -466,6 +472,10 @@ static void init(void)
 			goto ttt;
 #endif
 		}
+		init_b = 1;
+		init_bookmarks();
+		create_initial_extensions();
+		load_url_history();
 		initialize_all_subsystems_2();
 		info = create_session_info(base_session, u, default_target, &len);
 		if (!F) {
@@ -483,7 +493,7 @@ static void init(void)
 		}
 #endif
 	} else {
-		unsigned char *uu, *wd;
+		unsigned char *uu, *uuu, *wd;
 		initialize_all_subsystems_2();
 		close_socket(&terminal_pipe[0]);
 		close_socket(&terminal_pipe[1]);
@@ -492,9 +502,11 @@ static void init(void)
 			retval = RET_SYNTAX;
 			goto tttt;
 		}
-		if (!(uu = translate_url(u, wd = get_cwd()))) uu = stracpy(u);
-		request_object(NULL, uu, NULL, PRI_MAIN, NC_RELOAD, ALLOW_ALL, end_dump, NULL, &dump_obj);
+		uu = convert(get_commandline_charset(), utf8_table, u, NULL);
+		if (!(uuu = translate_url(uu, wd = get_cwd()))) uuu = stracpy(uu);
 		mem_free(uu);
+		request_object(NULL, uuu, NULL, PRI_MAIN, NC_RELOAD, ALLOW_ALL, end_dump, NULL, &dump_obj);
+		mem_free(uuu);
 		if (wd) mem_free(wd);
 	}
 }
@@ -507,8 +519,9 @@ static void initialize_all_subsystems(void)
 	set_sigcld();
 	init_home();
 	init_dns();
+	init_session_cache();
 	init_cache();
-	iinit_bfu();
+	init_blocks();
 	memset(&dd_opt, 0, sizeof dd_opt);
 }
 
@@ -526,9 +539,6 @@ static void terminate_all_subsystems(void)
 {
 	check_bottom_halves();
 	abort_all_downloads();
-#ifdef HAVE_SSL
-	ssl_finish();
-#endif
 	check_bottom_halves();
 	destroy_all_terminals();
 	check_bottom_halves();
@@ -539,16 +549,19 @@ static void terminate_all_subsystems(void)
 
 	free_all_caches();
 	free_format_text_cache();
+#ifdef HAVE_SSL
+	ssl_finish();
+#endif
 	if (init_b) save_url_history();
 	free_history_lists();
 	free_term_specs();
 	free_types();
 	free_blocks();
-	if (init_b) finalize_bookmarks();
+	finalize_bookmarks();
 	free_conv_table();
 	free_blacklist();
-	if (init_b) cleanup_cookies();
-	cleanup_auth();
+	free_cookies();
+	free_auth();
 	check_bottom_halves();
 	end_config();
 	free_strerror_buf();
@@ -556,7 +569,7 @@ static void terminate_all_subsystems(void)
 	GF(free_dither());
 	GF(shutdown_graphics());
 	af_unix_close();
-	if (clipboard) mem_free(clipboard), clipboard = NULL;
+	os_free_clipboard();
 	if (fg_poll_timer != NULL) kill_timer(fg_poll_timer), fg_poll_timer = NULL;
 	terminate_select();
 	terminate_osdep();
@@ -567,6 +580,13 @@ int main(int argc, char *argv[])
 	g_argc = argc;
 	g_argv = (unsigned char **)argv;
 
+#if 0
+	if (argc != 3)
+		fprintf(stderr, "two args expected\n"), exit(1);
+	printf("cookie %saccepted\n", allow_cookie_domain(cast_uchar argv[1], cast_uchar argv[2]) ? "" : "not ");
+	return 0;
+#endif
+
 	init_heap();
 
 	init_os();
@@ -595,13 +615,124 @@ int main(int argc, char *argv[])
 			gettimeofday(&tv2, NULL);
 			if (dst) mem_free(dst);
 			if (i)
-				tm += (tv2.tv_sec * 1000000 + tv2.tv_usec) - (tv1.tv_sec * 1000000 + tv1.tv_usec);
+				tm += ((ulonglong)tv2.tv_sec * 1000000 + tv2.tv_usec) - ((ulonglong)tv1.tv_sec * 1000000 + tv1.tv_usec);
 		}
 		fprintf(stderr, "time: %f\n", (double)tm / 1000 / rep);
 		check_memory_leaks();
 		return 0;
 	}
 #endif
+#if 0
+	{
+		int i;
+		for (i = 0; i < 100; i++) {
+			unsigned char *a = mem_calloc(i);
+			unsigned char *b = base64_encode(a, i, cast_uchar "", cast_uchar "", 5);
+			fprintf(stderr, "X:\n%.*s\n", (int)strlen(cast_const_char b), b);
+			mem_free(a);
+			mem_free(b);
+		}
+		check_memory_leaks();
+		return 0;
+	}
+#endif
+#if 0
+	{
+		unsigned char *puny_encode(unsigned char *s);
+		unsigned char *puny_decode(unsigned char *s);
+		int i;
+		for (i = 1; i < g_argc; i++) {
+			unsigned char *str = puny_encode(g_argv[i]);
+			if (str) {
+				fprintf(stderr, "'%s'\n", str);
+				unsigned char *d = puny_decode(str);
+				if (!d || strcmp(cast_const_char d, cast_const_char g_argv[i])) {
+					internal("mismatch - %s", d);
+				}
+				mem_free(str);
+				mem_free(d);
+			}
+		}
+		check_memory_leaks();
+		return 0;
+	}
+#endif
+#if 0
+	while (1) {
+#define maxn	12
+		static unsigned long long count = 0;
+		unsigned char *puny_encode(unsigned char *s, int len);
+		unsigned char *puny_decode(unsigned char *s, int len);
+		int punycode_encode(size_t input_length, const uint32_t input[], const unsigned char case_flags[], size_t *output_length, char output[]);
+		int punycode_decode(size_t input_length, const char input[], size_t * output_length, uint32_t output[], unsigned char case_flags[]);
+		uint32_t unistr[maxn];
+		unsigned i, n;
+		unsigned char *utfstr, *puny, *dec;
+		int utfstr_l;
+		char pce_o[60];
+		size_t pce_ol;
+		int pce_s;
+		n = random() % (maxn + 1);
+		for (i = 0; i < n; i++) {
+			uint32_t uni;
+			uni = random() % (0x10FFFF + 1);
+			if (uni >= 0xd800 && uni <= 0xdfff)
+				uni = random() % 128;
+			while (uni < 128 && !(
+				uni == '-' ||
+				(uni >= '0' && uni <= '9') ||
+				(uni >= 'A' && uni <= 'Z') ||
+				(uni >= 'a' && uni <= 'z'))) {
+				uni = random() % 128;
+			}
+			unistr[i] = uni;
+		}
+		utfstr = init_str();
+		utfstr_l = 0;
+		for (i = 0; i < n; i++) {
+			unsigned char *us = encode_utf_8(unistr[i]);
+			add_to_str(&utfstr, &utfstr_l, us);
+		}
+		puny = puny_encode(utfstr, utfstr_l);
+		if (!puny) {
+			fprintf(stderr, "failed:");
+			goto err;
+		}
+		dec = puny_decode(puny, (int)strlen(cast_const_char puny));
+		if (!dec)
+			dec = stracpy(puny);
+		if (strcmp(cast_const_char utfstr, cast_const_char dec)) {
+			fprintf(stderr, "mismatch(%s,%s):", utfstr, dec);
+			goto err;
+		}
+		pce_ol = 59;
+		pce_s = punycode_encode(n, unistr, NULL, &pce_ol, pce_o);
+		if (pce_s) {
+			fprintf(stderr, "punycode_encode(%d):", pce_s);
+			goto err;
+		}
+		pce_o[pce_ol] = 0;
+		if (!strncmp(cast_const_char puny, "xn--", 4) && strcmp(cast_const_char puny + 4, pce_o)) {
+			fprintf(stderr, "punycode_encode differs(%s,%s):", puny, pce_o);
+			goto err;
+		}
+		mem_free(dec);
+		mem_free(puny);
+		mem_free(utfstr);
+		if (0) {
+err:
+			for (i = 0; i < n; i++) {
+				fprintf(stderr, " %u", unistr[i]);
+			}
+			internal("failed");
+		}
+		count++;
+		if (!(count % 10000)) {
+			printf("%llu\r", count);
+			fflush(stdout);
+		}
+	}
+#endif
 
 	select_loop(init);
 	terminate_all_subsystems();
