@@ -7,6 +7,30 @@
 #include <values.h>
 #endif
 
+#ifndef MAXINT
+#ifdef INT_MAX
+#define MAXINT INT_MAX
+#else
+#define MAXINT ((int)((unsigned int)-1 >> 1))
+#endif
+#endif
+
+#ifndef MAXLONG
+#ifdef LONG_MAX
+#define MAXLONG LONG_MAX
+#else
+#define MAXLONG ((long)((unsigned long)-1L >> 1))
+#endif
+#endif
+
+#if defined(HAVE_STRTOLL) && !defined(MAXLLONG)
+#ifdef LLONG_MAX
+#define MAXLLONG LLONG_MAX
+#else
+#define MAXLLONG ((long long)((unsigned long long)-1L >> 1))
+#endif
+#endif
+
 #if !defined(HAVE_ERRNO) && !defined(errno)
 extern int errno;
 #endif
@@ -166,12 +190,10 @@ int bounced_write(int fd, const void *buf, size_t size);
 #define OS_SETRAW
 int vms_read(int fd, void *buf, size_t size);
 int vms_write(int fd, const void *buf, size_t size);
-int vms_pipe(int fd[2]);
-int vms_close(int fd);
-int vms_select(int n, fd_set *rs, fd_set *ws, fd_set *es, struct timeval *t);
 #define read vms_read
 #define write vms_write
-#define pipe vms_pipe
+int vms_close(int fd);
+int vms_select(int n, fd_set *rs, fd_set *ws, fd_set *es, struct timeval *t);
 #define close vms_close
 #define select vms_select
 #endif
@@ -183,12 +205,10 @@ int vms_select(int n, fd_set *rs, fd_set *ws, fd_set *es, struct timeval *t);
 #endif
 int dos_read(int fd, void *buf, size_t size);
 int dos_write(int fd, const void *buf, size_t size);
-int dos_pipe(int fd[2]);
 int dos_close(int fd);
 int dos_select(int n, fd_set *rs, fd_set *ws, fd_set *es, struct timeval *t, int from_main_loop);
 #define read dos_read
 #define write dos_write
-#define pipe dos_pipe
 #define close dos_close
 #define select(a, b, c, d, e) dos_select(a, b, c, d, e, 0)
 #endif
@@ -219,6 +239,11 @@ int dos_select(int n, fd_set *rs, fd_set *ws, fd_set *es, struct timeval *t, int
 #undef HAVE_HERROR
 #endif
 
+#ifdef HAVE_MAXINT_CONVERSION_BUG
+#undef MAXINT
+#define MAXINT 0x7FFFF000
+#endif
+
 #ifndef HAVE_SOCKLEN_T
 #ifdef OPENVMS
 #define socklen_t unsigned
@@ -242,23 +267,6 @@ int dos_select(int n, fd_set *rs, fd_set *ws, fd_set *es, struct timeval *t, int
 #define getsockopt hp_getsockopt
 #endif
 
-#ifndef MAXINT
-#ifdef INT_MAX
-#define MAXINT INT_MAX
-#else
-#define MAXINT ((int)((unsigned int)-1 >> 1))
-#endif
-#endif
-
-#define MAX_SIZE_T	((size_t)-1)
-
-#ifdef HAVE_MAXINT_CONVERSION_BUG
-#undef MAXINT
-#define MAXINT		0x7FFFF000
-#undef MAX_SIZE_T
-#define MAX_SIZE_T	((size_t)-0x1000)
-#endif
-
 #if defined(OPENVMS_64BIT)
 #define my_intptr_t long long
 #define my_uintptr_t unsigned long long
@@ -298,10 +306,6 @@ int dos_select(int n, fd_set *rs, fd_set *ws, fd_set *es, struct timeval *t, int
 
 #if defined(HAVE_GETADDRINFO) && defined(HAVE_FREEADDRINFO) && !defined(ATHEOS)
 #define USE_GETADDRINFO
-#else
-#ifdef SUPPORT_IPV6
-#undef SUPPORT_IPV6
-#endif
 #endif
 
 #ifdef SUPPORT_IPV6
@@ -353,12 +357,6 @@ int dos_select(int n, fd_set *rs, fd_set *ws, fd_set *es, struct timeval *t, int
 #define sun_volatile
 #endif
 
-#if defined(__TINYC__) && defined(__aarch64__)
-#define tcc_hack	+ 0
-#else
-#define tcc_hack
-#endif
-
 typedef const char *const_char_ptr;
 
 #if defined(BEOS) || (defined(HAVE_BEGINTHREAD) && defined(OS2)) || defined(HAVE_PTHREADS) || (defined(HAVE_ATHEOS_THREADS_H) && defined(HAVE_SPAWN_THREAD) && defined(HAVE_RESUME_THREAD))
