@@ -29,19 +29,21 @@
 #if DEBUGLEVEL < 0
 #define OOPS
 #define LEAK_DEBUG
+#define NO_IE
 #endif
 
-#define LINKS_MANUAL_URL		(!strcmp(cast_const_char language_name(current_language),"Czech") || !strcmp(cast_const_char language_name(current_language),"Slovak")?\
+#define LINKS_MANUAL_URL		(!strcmp(cast_const_char language_name(get_current_language()),"Czech") || !strcmp(cast_const_char language_name(get_current_language()),"Slovak")?\
 					"http://links.twibright.com/user.html":\
 					"http://links.twibright.com/user_en.html")
-#define LINKS_HOMEPAGE_URL		(!strcmp(cast_const_char language_name(current_language),"Czech") || !strcmp(cast_const_char language_name(current_language),"Slovak")?\
+#define LINKS_HOMEPAGE_URL		(!strcmp(cast_const_char language_name(get_current_language()),"Czech") || !strcmp(cast_const_char language_name(get_current_language()),"Slovak")?\
 					"http://links.twibright.com/index_cz.php":\
 					"http://links.twibright.com/")
-#define LINKS_CALIBRATION_URL		(!strcmp(cast_const_char language_name(current_language),"Czech") || !strcmp(cast_const_char language_name(current_language),"Slovak")?\
+#define LINKS_CALIBRATION_URL		(!strcmp(cast_const_char language_name(get_current_language()),"Czech") || !strcmp(cast_const_char language_name(get_current_language()),"Slovak")?\
 					"http://links.twibright.com/kalibrace.html":\
 					"http://links.twibright.com/calibration.html")
 
 #define LINKS_SOCK_NAME			"socket"
+#define LINKS_ANONYMOUS_SOCK_SUFFIX	"-a"
 #define LINKS_PORT			23755
 #define LINKS_G_PORT_START		32768
 #define LINKS_G_PORT_LEN		16384
@@ -50,11 +52,14 @@
 
 #define FG_POLL_TIME			500
 
+#define OPENMP_MAX_THREADS		8
+
 #define HEAPMIN_FACTOR			100
 #define HEAPMIN_INTERVAL		10000
 #define MALLOC_TRIM_INTERVAL		300000
 
 #define DNS_TIMEOUT			3600000UL
+#define SESSION_TIMEOUT			14400000UL
 
 #define HTTP_KEEPALIVE_TIMEOUT		300000
 #define FTP_KEEPALIVE_TIMEOUT		600000
