@@ -74,19 +74,7 @@ void add_bytes_to_str(unsigned char **s, int *l, unsigned char *a, size_t ll)
 	if (x >= old_length) {
 		/* Need to realloc */
 #ifdef HAVE___BUILTIN_CLZ
-#if !(									\
-	defined(__tune_i386__) ||					\
-	defined(__tune_i486__) ||					\
-	defined(__tune_i586__) ||					\
-	defined(__tune_k6__) ||						\
-	defined(__tune_lakemont__) ||					\
-	(defined(__alpha__) && !defined(__alpha_cix__)) ||		\
-	(defined(__mips) && __mips < 32) ||				\
-	(defined(__ARM_ARCH) && __ARM_ARCH < 5) ||			\
-	(defined(__sparc__) && (!defined(__VIS__) || __VIS__ < 0x300)) ||\
-	defined(__hppa) ||						\
-	defined(__riscv) ||						\
-	defined(__sh__))
+#if !(defined(__tune_i386__) || defined(__tune_i486__) || defined(__tune_i586__) || defined(__tune_k6__) || defined(__tune_lakemont__) || (defined(__alpha__) && !defined(__alpha_cix__)) || (defined(__mips) && __mips < 32) || (defined(__ARM_ARCH) && __ARM_ARCH < 5) || (defined(__sparc__) && (!defined(__VIS__) || __VIS__ < 0x300)) || defined(__hppa) || defined(__sh__))
 		if (!(sizeof(unsigned) & (sizeof(unsigned) - 1))) {
 			new_length = 2U << ((sizeof(unsigned) * 8 - 1)
 #ifdef __ICC
@@ -133,25 +121,8 @@ void add_unsigned_long_num_to_str(unsigned char **s, int *l, my_uintptr_t n)
 void add_num_to_str(unsigned char **s, int *l, off_t n)
 {
 	unsigned char a[64];
-	if (n >= 0 && n < 1000) {
-		unsigned sn = (unsigned)n;
-		unsigned char *p = a;
-		if (sn >= 100) {
-			*p++ = '0' + sn / 100;
-			sn %= 100;
-			goto d10;
-		}
-		if (sn >= 10) {
-			d10:
-			*p++ = '0' + sn / 10;
-			sn %= 10;
-		}
-		*p++ = '0' + sn;
-		add_bytes_to_str(s, l, a, p - a);
-	} else {
-		snzprint(a, 64, n);
-		add_to_str(s, l, a);
-	}
+	snzprint(a, 64, n);
+	add_to_str(s, l, a);
 }
 
 void add_knum_to_str(unsigned char **s, int *l, off_t n)
@@ -165,11 +136,10 @@ void add_knum_to_str(unsigned char **s, int *l, off_t n)
 
 long strtolx(unsigned char *c, unsigned char **end)
 {
-	char *end_c;
 	long l;
-	if (c[0] == '0' && upcase(c[1]) == 'X' && c[2]) l = strtol(cast_const_char(c + 2), &end_c, 16);
-	else l = strtol(cast_const_char c, &end_c, 10);
-	*end = cast_uchar end_c;
+	if (c[0] == '0' && upcase(c[1]) == 'X' && c[2]) l = strtol(cast_const_char(c + 2), (char **)(void *)end, 16);
+	else l = strtol(cast_const_char c, (char **)(void *)end, 10);
+	if (!*end) return l;
 	if (upcase(**end) == 'K') {
 		(*end)++;
 		if (l < -MAXINT / 1024) return -MAXINT;
@@ -187,29 +157,31 @@ long strtolx(unsigned char *c, unsigned char **end)
 
 my_strtoll_t my_strtoll(unsigned char *string, unsigned char **end)
 {
-	char *end_c;
 	my_strtoll_t f;
 	errno = 0;
 #if defined(HAVE_STRTOLL)
-	f = strtoll(cast_const_char string, &end_c, 10);
+	f = strtoll(cast_const_char string, (char **)(void *)end, 10);
 #elif defined(HAVE_STRTOQ)
-	f = strtoq(cast_const_char string, &end_c, 10);
+	f = strtoq(cast_const_char string, (char **)(void *)end, 10);
 #elif defined(HAVE_STRTOIMAX)
-	f = strtoimax(cast_const_char string, &end_c, 10);
+	f = strtoimax(cast_const_char string, (char **)(void *)end, 10);
 #else
-	f = strtol(cast_const_char string, &end_c, 10);
+	f = strtol(cast_const_char string, (char **)(void *)end, 10);
 #endif
-	if (end)
-		*end = cast_uchar end_c;
 	if (f < 0 || errno) return -1;
 	return f;
 }
 
 /* Copies at most dst_size chars into dst. Ensures null termination of dst. */
-void safe_strncpy(unsigned char *dst, const unsigned char *src, size_t dst_size)
+unsigned char *safe_strncpy(unsigned char *dst, const unsigned char *src, size_t dst_size)
 {
-	dst[dst_size - 1] = 0;
-	strncpy(cast_char dst, cast_const_char src, dst_size - 1);
+	size_t to_copy;
+	if (!dst_size) return dst;
+	to_copy = strlen(cast_const_char src);
+	if (to_copy >= dst_size) to_copy = dst_size - 1;
+	memcpy(dst, src, to_copy);
+	memset(dst + to_copy, 0, dst_size - to_copy);
+	return dst;
 }
 
 #ifdef JS
@@ -315,10 +287,4 @@ int casestrstr(const unsigned char *h, const unsigned char *n)
 	return 0;
 }
 
-unsigned hash_string(unsigned char *u)
-{
-	unsigned hash = 0;
-	for (; *u; u++)
-		hash = hash * 0x11 + *u;
-	return hash;
-}
+
