@@ -684,7 +684,7 @@ static void svga_unregister_bitmap(struct bitmap *bmp)
 	unsigned char *data=hndl->data;\
 	int mouse_hidden;\
 \
- 	TEST_INACTIVITY\
+	TEST_INACTIVITY\
 	if (x>=dev->clip.x2||x+xs<=dev->clip.x1) return;\
 	if (y>=dev->clip.y2||y+ys<=dev->clip.y1) return;\
 	if (x+xs>dev->clip.x2) xs=dev->clip.x2-x;\
@@ -801,7 +801,7 @@ static inline void draw_bitmap_linear(struct graphics_device *dev,struct bitmap*
  * clipping rectangle set up by svga_driver->set_clip_area. If empty region
  * results, return from current function occurs. */
 #define FILL_CLIP_PREFACE \
- 	int mouse_hidden;\
+	int mouse_hidden;\
 	TEST_INACTIVITY\
 	if (left>=right||top>=bottom) return;\
 	if (left>=dev->clip.x2||right<=dev->clip.x1||top>=dev->clip.y2||bottom<=dev->clip.y1) return;\
@@ -1704,7 +1704,7 @@ void dump_mode_info_into_file(vga_modeinfo* i)
 }
 #endif
 
-static void svgalib_key_in(void *p, unsigned char *ev_, int size)
+static void svgalib_key_in(struct itrm *p, unsigned char *ev_, int size)
 {
 	struct links_event *ev = (struct links_event *)(void *)ev_;
 	if (size != sizeof(struct links_event)) return;
@@ -1749,7 +1749,7 @@ static void mouse_event_handler(int button, int dx, int dy, int dz, int drx, int
 	void (*mh)(struct graphics_device *, int, int, int);
 	struct graphics_device *cd=current_virtual_device;
 
-	mh=cd?cd->mouse_handler:NULL;
+	mh=cd?cd->mouse_handler:(void(*)(struct graphics_device *dev, int x, int y, int buttons))NULL;
 	old_mouse_x=mouse_x;
 	old_mouse_y=mouse_y;
 
@@ -2086,7 +2086,7 @@ static inline void place_mouse_composite(void)
 			if (background_length>0)
 				mouse_drawscansegment(
 					background_ptr +l2,
-				       	mouse_right,mouse_top ,l3);
+					mouse_right,mouse_top ,l3);
 			mouse_ptr+=skip;
 			background_ptr+=skip;
 		}
@@ -2258,14 +2258,14 @@ static void vtswitch_handler(void *nothing)
 }
 #endif
 
-static void svga_ctrl_c(struct itrm *i)
+static void svga_ctrl_c(void *i_)
 {
 	kbd_ctrl_c();
 }
 
 /* Param: one string which is to be compared with one from modes.
  * Copies the svga_driver into gr_driver.
- * Returns: 	0 OK
+ * Returns:	0 OK
  *		1 Passed mode unknown by svga_driver
  *		2 Passed mode unknown by svgalib
  * mikulas: Change: Returns:	NULL: OK
@@ -2286,7 +2286,7 @@ static unsigned char *svga_init_driver(unsigned char *param, unsigned char *disp
 	if (!param || !*param) goto not_found;
 	svga_driver_param=stracpy(param);
 	for (j=0;(size_t)j<sizeof(modes)/sizeof(*modes);j++)
-		if (!strcasecmp(cast_const_char modes[j].name,cast_const_char param)) goto found;
+		if (!casestrcmp(cast_uchar modes[j].name, param)) goto found;
 	j = 1;
 	not_found:
 	{
@@ -2372,7 +2372,7 @@ static unsigned char *svga_init_driver(unsigned char *param, unsigned char *disp
 	ignore_signals();
 	errno = 0;
 	while (signal(SIGTSTP, SIG_IGN) == SIG_ERR && errno == EINTR) errno = 0;
-	install_signal_handler(SIGINT, (void (*)(void *))svga_ctrl_c, svgalib_kbd, 0);
+	install_signal_handler(SIGINT, svga_ctrl_c, svgalib_kbd, 0);
 	return NULL;
 }
 
@@ -2544,6 +2544,7 @@ struct graphics_driver svga_driver={
 	NULL, /* hscroll */
 	NULL, /* vscroll */
 	generic_set_clip_area,
+	NULL,
 	vga_block, /* block */
 	vga_unblock, /* unblock */
 	NULL, /* set_title */
@@ -2552,7 +2553,11 @@ struct graphics_driver svga_driver={
 	NULL, /* get_clipboard_text */
 	0,				/* depth */
 	0, 0,				/* size */
-	GD_NO_LIBEVENT,			/* flags */
+	GD_NO_LIBEVENT
+#ifndef SPAD
+		| GD_NOAUTO
+#endif
+		,			/* flags */
 	0,				/* codepage */
 	NULL,				/* shell */
 };
