@@ -129,7 +129,7 @@ static unsigned char *x_display_string = NULL;
 static int x_screen;   /* screen */
 static int x_display_height,x_display_width;   /* screen dimensions */
 static unsigned long x_black_pixel;  /* black pixel */
-static int x_depth,x_bitmap_bpp;   /* bits per pixel and bytes per pixel */
+static int x_depth, x_bitmap_bpp;   /* bits per pixel and bytes per pixel */
 static int x_bitmap_scanline_pad; /* bitmap scanline_padding in bytes */
 static int x_colors;  /* colors in the palette (undefined when there's no palette) */
 static int x_have_palette;
@@ -212,7 +212,7 @@ static inline void X_SCHEDULE_PROCESS_EVENTS(void)
 {
 	if (!process_events_in_progress)
 	{
-		register_bottom_half(x_process_events,NULL);
+		register_bottom_half(x_process_events, NULL);
 		process_events_in_progress = 1;
 	}
 }
@@ -239,7 +239,7 @@ static inline void X_FLUSH(void)
 #else
 	if (!flush_in_progress)
 	{
-		register_bottom_half(x_do_flush,NULL);
+		register_bottom_half(x_do_flush, NULL);
 		flush_in_progress=1;
 	}
 #endif
@@ -264,7 +264,7 @@ static void x_prepare_for_failure(void)
 
 static int x_test_for_failure(void)
 {
-	XSync(x_display,False);
+	XSync(x_display, False);
 	X_SCHEDULE_PROCESS_EVENTS();
 	XSetErrorHandler(old_error_handler);
 	old_error_handler = NULL;
@@ -279,7 +279,7 @@ static void x_clip_number(int *n,int l,int h)
 }
 
 
-static unsigned char * x_set_palette(void)
+static unsigned char *x_set_palette(void)
 {
 	XColor color;
 	unsigned a;
@@ -303,8 +303,7 @@ static unsigned char * x_set_palette(void)
 	return NULL;
 }
 
-static unsigned char static_color_table[256];
-static int use_static_color_table;
+static unsigned char *static_color_table = NULL;
 
 static unsigned char *x_query_palette(void)
 {
@@ -312,6 +311,8 @@ static unsigned char *x_query_palette(void)
 	XColor color[256];
 	unsigned char valid[256];
 	int some_valid = 0;
+	if (x_depth > 8)
+		return stracpy(cast_uchar "Static color supported for up to 8-bit depth\n");
 	memset(color, 0, sizeof color);
 	memset(valid, 0, sizeof valid);
 	for (i = 0; i < 1 << x_depth; i++) {
@@ -325,6 +326,7 @@ static unsigned char *x_query_palette(void)
 	}
 	if (!some_valid)
 		return stracpy(cast_uchar "Could not query static colormap\n");
+	static_color_table = mem_calloc(256);
 	for (i = 0; i < 1 << x_depth; i++) {
 		double best_distance = 0;
 		int best = -1;
@@ -341,7 +343,6 @@ static unsigned char *x_query_palette(void)
 		}
 		static_color_table[i] = (unsigned char)best;
 	}
-	use_static_color_table = 1;
 	return NULL;
 }
 
@@ -542,8 +543,8 @@ static void x_free_hash_table(void)
 {
 	int a,b;
 
-	unregister_bottom_half(x_process_events,NULL);
-	unregister_bottom_half(x_do_flush,NULL);
+	unregister_bottom_half(x_process_events, NULL);
+	unregister_bottom_half(x_do_flush, NULL);
 
 	for (a=0;a<X_HASH_TABLE_SIZE;a++)
 	{
@@ -555,6 +556,8 @@ static void x_free_hash_table(void)
 
 	x_clear_clipboard();
 
+	if (static_color_table) mem_free(static_color_table), static_color_table = NULL;
+
 	if (x_display) {
 		if (x_icon) XFreePixmap(x_display, x_icon), x_icon = 0;
 		if (fake_window_initialized) XDestroyWindow(x_display,fake_window), fake_window_initialized = 0;
@@ -810,6 +813,7 @@ static void x_process_events(void *data)
 				event.xclient.message_type!=x_wm_protocols_atom||
 				(Atom)event.xclient.data.l[0]!=x_delete_window_atom
 			)break;
+			/*-fallthrough*/
 
 			/* This event is destroy window event from window manager */
 
@@ -927,7 +931,7 @@ static void x_process_events(void *data)
 				}
 				x_clip_number(&(event.xmotion.x),gd->size.x1,gd->size.x2);
 				x_clip_number(&(event.xmotion.y),gd->size.y1,gd->size.y2);
-				gd->mouse_handler(gd,event.xbutton.x,event.xbutton.y,a|(a != B_WHEELDOWN && a != B_WHEELUP && a != B_WHEELLEFT && a != B_WHEELRIGHT ? B_DOWN : B_MOVE));
+				gd->mouse_handler(gd,event.xbutton.x, event.xbutton.y, a | (!BM_IS_WHEEL(a) ? B_DOWN : B_MOVE));
 				p_xx:;
 			}
 			break;
@@ -945,6 +949,8 @@ static void x_process_events(void *data)
 				/* just sign, that this was mouse event */
 				last_was_mouse=1;
 				last_event=event;
+				/* fix lag when using remote X and dragging over some text */
+				XSync(x_display, False);
 			}
 			break;
 
@@ -1049,7 +1055,7 @@ static XIC x_open_xic(Window w);
 #endif
 
 /* initiate connection with X server */
-static unsigned char * x_init_driver(unsigned char *param, unsigned char *display)
+static unsigned char *x_init_driver(unsigned char *param, unsigned char *display)
 {
 	XGCValues gcv;
 	XSetWindowAttributes win_attr;
@@ -1066,7 +1072,7 @@ static unsigned char * x_init_driver(unsigned char *param, unsigned char *displa
 #ifdef X_DEBUG
 	{
 		unsigned char txt[256];
-		sprintf(txt,"x_init_driver(%s, %s)\n",param, display);
+		sprintf(txt,"x_init_driver(%s, %s)\n", param, display);
 		MESSAGE(txt);
 	}
 #endif
@@ -1113,36 +1119,40 @@ static unsigned char * x_init_driver(unsigned char *param, unsigned char *displa
 		return err;
 	}
 
-	x_bitmap_bit_order=BitmapBitOrder(x_display);
-	x_screen=DefaultScreen(x_display);
-	x_display_height=DisplayHeight(x_display,x_screen);
-	x_display_width=DisplayWidth(x_display,x_screen);
-	x_root_window=RootWindow(x_display,x_screen);
+	x_bitmap_bit_order = BitmapBitOrder(x_display);
+	x_screen = DefaultScreen(x_display);
+	x_display_height = DisplayHeight(x_display, x_screen);
+	x_display_width = DisplayWidth(x_display, x_screen);
+	x_root_window = RootWindow(x_display, x_screen);
 
-	x_default_window_width=x_display_width-50;
-	x_default_window_height=x_display_height-50;
+	x_default_window_width = x_display_width;
+	if (x_default_window_width >= 100)
+		x_default_window_width -= 50;
+	x_default_window_height = x_display_height;
+	if (x_default_window_height >= 150)
+		x_default_window_height -= 75;
 
-	x_driver_param=NULL;
+	x_driver_param = NULL;
 
 	if (param && *param)
 	{
 		unsigned char *e;
-		unsigned long w,h;
+		unsigned long w, h;
 
-		x_driver_param=stracpy(param);
+		x_driver_param = stracpy(param);
 
 		if (*x_driver_param < '0' || *x_driver_param > '9') {
 			invalid_param:
 			x_free_hash_table();
 			return stracpy(cast_uchar "Invalid parameter\n");
 		}
-		w=strtoul(cast_const_char x_driver_param,(char **)(void *)&e,10);
+		w=strtoul(cast_const_char x_driver_param, (char **)(void *)&e, 10);
 		if (upcase(*e) != 'X') goto invalid_param;
 		e++;
 		if (*e < '0' || *e > '9') goto invalid_param;
-		h=strtoul(cast_const_char e,(char **)(void *)&e,10);
+		h=strtoul(cast_const_char e, (char **)(void *)&e, 10);
 		if (*e) goto invalid_param;
-		if (w && h && w <= MAXINT && h <= MAXINT) {
+		if (w && h && w <= 30000 && h <= 30000) {
 			x_default_window_width=(int)w;
 			x_default_window_height=(int)h;
 		}
@@ -1260,7 +1270,7 @@ visual_found:;
 		case 386:
 		case 452:
 		case 708:
-/* 			printf("depth=%d visualid=%x\n",x_driver.depth, vinfo.visualid); */
+/*			printf("depth=%d visualid=%x\n",x_driver.depth, vinfo.visualid); */
 		break;
 
 		default:
@@ -1288,7 +1298,6 @@ visual_found:;
 
 	x_colors=1<<x_depth;
 	x_have_palette=0;
-	use_static_color_table = 0;
 	if (vinfo.class==DirectColor||vinfo.class==PseudoColor)
 	{
 		unsigned char *t;
@@ -1360,8 +1369,13 @@ visual_found:;
 		 * So, try to set locale to utf8 for the input method.
 		 */
 		unsigned char *l, *m, *d;
+		int len;
 		l = cast_uchar setlocale(LC_CTYPE, "");
-		if (l) {
+		len = l ? (int)strlen(cast_const_char l) : 0;
+		if (l &&
+		    !(len >= 5 && !casestrcmp(l + len - 5, cast_uchar ".utf8")) &&
+		    !(len >= 6 && !casestrcmp(l + len - 6, cast_uchar ".utf-8"))
+		    ) {
 			m = stracpy(l);
 			d = cast_uchar strchr(cast_const_char m, '.');
 			if (d) *d = 0;
@@ -1374,6 +1388,12 @@ visual_found:;
 		}
 #endif
 		xim = XOpenIM(x_display, NULL, NULL, NULL);
+#if defined(HAVE_SETLOCALE) && defined(LC_CTYPE)
+		if (!xim) {
+			l = cast_uchar setlocale(LC_CTYPE, "en_US.UTF-8");
+			xim = XOpenIM(x_display, NULL, NULL, NULL);
+		}
+#endif
 		if (xim) {
 			XIC xic = x_open_xic(fake_window);
 			if (xic) {
@@ -1497,6 +1517,7 @@ nic_nebude_bobankove:;
 	XSetWMName(x_display, wi->window, &windowName);
 	XStoreName(x_display,wi->window,cast_char links_name);
 	XSetWMIconName(x_display, wi->window, &windowName);
+	XFree(windowName.value);
 
 	XMapWindow(x_display,wi->window);
 
@@ -1530,7 +1551,7 @@ nic_nebude_bobankove:;
 	}
 #endif
 
-	XSync(x_display,False);
+	XSync(x_display, False);
 	X_SCHEDULE_PROCESS_EVENTS();
 	n_wins++;
 	return gd;
@@ -1544,16 +1565,16 @@ static void x_shutdown_device(struct graphics_device *gd)
 #ifdef X_DEBUG
 	MESSAGE("x_shutdown_device\n");
 #endif
-	if (!gd)return;
+	if (!gd) return;
 
 	n_wins--;
-	XDestroyWindow(x_display,wi->window);
+	XDestroyWindow(x_display, wi->window);
 #ifdef X_INPUT_METHOD
 	if (wi->xic) {
 		XDestroyIC(wi->xic);
 	}
 #endif
-	XSync(x_display,False);
+	XSync(x_display, False);
 	X_SCHEDULE_PROCESS_EVENTS();
 
 	x_remove_from_table(&wi->window);
@@ -1564,7 +1585,7 @@ static void x_shutdown_device(struct graphics_device *gd)
 static void x_translate_colors(unsigned char *data, int x, int y, int skip)
 {
 	int i, j;
-	if (!use_static_color_table)
+	if (!static_color_table)
 		return;
 	for (j = 0; j < y; j++) {
 		for (i = 0; i < x; i++)
@@ -1712,7 +1733,7 @@ static long x_get_color(int rgb)
 	b = (unsigned char *)&block;
 	/*fprintf(stderr, "bitmap bpp %d\n", x_bitmap_bpp);*/
 	switch (x_bitmap_bpp) {
-		case 1:		if (use_static_color_table)
+		case 1:		if (static_color_table)
 					return static_color_table[b[0]];
 				return b[0];
 		case 2:		if (x_bitmap_bit_order == LSBFirst)
@@ -1912,7 +1933,7 @@ static int x_hscroll(struct graphics_device *gd, struct rect_set **set, int sc)
 		gd->clip.x2-gd->clip.x1,gd->clip.y2-gd->clip.y1,
 		gd->clip.x1+sc,gd->clip.y1
 	);
-	XSync(x_display,False);
+	XSync(x_display, False);
 	/* ten sync tady musi byt, protoze potrebuju zarucit, aby vsechny
 	 * graphics-expose vyvolane timto scrollem byly vraceny v rect-set */
 
@@ -1993,7 +2014,7 @@ static int x_vscroll(struct graphics_device *gd, struct rect_set **set, int sc)
 		gd->clip.x2-gd->clip.x1,gd->clip.y2-gd->clip.y1,
 		gd->clip.x1,gd->clip.y1+sc
 	);
-	XSync(x_display,False);
+	XSync(x_display, False);
 	/* ten sync tady musi byt, protoze potrebuju zarucit, aby vsechny
 	 * graphics-expose vyvolane timto scrollem byly vraceny v rect-set */
 
@@ -2132,9 +2153,15 @@ static void x_commit_strip(struct bitmap *bmp, int top, int lines)
 }
 
 
+static void x_flush(struct graphics_device *gd)
+{
+	unregister_bottom_half(x_do_flush, NULL);
+	x_do_flush(NULL);
+}
+
+
 static void x_set_window_title(struct graphics_device *gd, unsigned char *title)
 {
-	struct conv_table *ct;
 	unsigned char *t;
 	XTextProperty windowName;
 	int output_encoding;
@@ -2150,10 +2177,8 @@ retry_encode_ascii:
 		output_encoding = 0;
 	}
 
-	ct = get_translation_table(utf8_table,output_encoding);
-
 	if (!gd)internal("x_set_window_title called with NULL graphics_device pointer.\n");
-	t = convert_string(ct, title, (int)strlen(cast_const_char title), NULL);
+	t = convert(utf8_table, output_encoding, title, NULL);
 	clr_white(t);
 	/*XStoreName(x_display,get_window_info(gd)->window,"blabla");*/
 
@@ -2232,12 +2257,8 @@ static void selection_request(XEvent *event)
 #endif
 	if (req->target == XA_STRING) {
 		unsigned char *str, *p;
-		struct conv_table *ct = NULL;
-		int iso1 = get_cp_index(cast_uchar "iso-8859-1");
-		if (iso1 >= 0) ct = get_translation_table(utf8_table, iso1);
 		if (!x_my_clipboard) str = stracpy(cast_uchar "");
-		else if (!ct) str = stracpy(x_my_clipboard);
-		else str = convert_string(ct, x_my_clipboard, (int)strlen(cast_const_char x_my_clipboard), NULL);
+		else str = convert(utf8_table, get_cp_index(cast_uchar "iso-8859-1"), x_my_clipboard, NULL);
 		for (p = cast_uchar strchr(cast_const_char str, 1); p; p = cast_uchar strchr(cast_const_char(str + 1), 1)) *p = 0xa0;
 		l = strlen(cast_const_char str);
 		if (l > X_MAX_CLIPBOARD_SIZE) l = X_MAX_CLIPBOARD_SIZE;
@@ -2358,14 +2379,7 @@ static unsigned char *x_get_clipboard_text(void)
 		if (type_atom == x_utf8_string_atom) {
 			x_my_clipboard = stracpy(buffer);
 		} else {
-			struct conv_table *ct = NULL;
-			int iso1 = get_cp_index(cast_uchar "iso-8859-1");
-			if (iso1 >= 0) ct = get_translation_table(iso1, utf8_table);
-			if (!ct) {
-				x_my_clipboard = stracpy(buffer);
-			} else {
-				x_my_clipboard = convert_string(ct, buffer, (int)strlen(cast_const_char buffer), NULL);
-			}
+			x_my_clipboard = convert(get_cp_index(cast_uchar "iso-8859-1"), utf8_table, buffer, NULL);
 		}
 		XFree(buffer);
 	} else {
@@ -2468,6 +2482,7 @@ struct graphics_driver x_driver={
 	x_hscroll,
 	x_vscroll,
 	x_set_clip_area,
+	x_flush,
 	dummy_block,
 	dummy_unblock,
 	x_set_window_title,
@@ -2476,7 +2491,7 @@ struct graphics_driver x_driver={
 	x_get_clipboard_text,
 	0,				/* depth (filled in x_init_driver function) */
 	0, 0,				/* size (in X is empty) */
-	0,				/* flags */
+	GD_UNICODE_KEYS,		/* flags */
 	0,				/* codepage */
 	NULL,				/* shell */
 };
